/*
 * ComputeThread.cpp
 *
 *  Revised on: Jun 23, 2017
 *      Author: AMdR
 */

#include "ComputeThread.h"
#include "integrator/ebttooldefs.h"
#ifdef DEBUG
#include <QtDebug>
#include <QElapsedTimer>
#endif

// Signal handling; tzt afhankelijk van HAS_SIGNALS:
#include <signal.h>

#if (defined(Q_OS_LINUX) || defined(Q_OS_MACOS) || (defined(Q_OS_WIN) && defined(__MINGW32__)))

// FPU error control
#include <fenv.h>
#define testFPU()                 fetestexcept(FE_DIVBYZERO | FE_INVALID | FE_OVERFLOW)
#define clearFPU()                feclearexcept(FE_ALL_EXCEPT)
#define testExcepts(x)            (x & (FE_DIVBYZERO | FE_INVALID | FE_OVERFLOW))

#elif defined(Q_OS_WIN)

// FPU error control
#include <float.h>
#define testFPU()                 _status87(/*EM_INVALID | EM_ZERODIVIDE | EM_OVERFLOW*/)
#define clearFPU()                _clear87()
#define testExcepts(x)            (x & (EM_ZERODIVIDE | EM_INVALID | EM_OVERFLOW))

#endif

ComputeThread::ComputeThread(QObject *parent) : QThread(parent)
{
  pause          = false;
  quit           = false;
  restart        = false;
  abort          = false;
  readyForSignal = true;
  signal(SIGFPE, SIG_IGN);
}

ComputeThread::~ComputeThread()
{
  mutex.lock();
  abort = true;
  condition.wakeOne();
  mutex.unlock();

  wait();
}

void ComputeThread::newRun(EBTModel *model2Run, EBTDataMap *outDataMap)
{
  QMutexLocker locker(&mutex);

  this->model      = model2Run;
  this->outputData = outDataMap;

  if (!isRunning())
    {
      start(LowPriority);
    }
  else
    {
      restart = true;
      condition.wakeOne();
    }
}

void ComputeThread::pauseRun()
{
  if (isRunning())
    {
      mutex.lock();
      pause = true;
      condition.wakeOne();
      mutex.unlock();
    }
}

void ComputeThread::resumeRun()
{
  if (isRunning())
    {
      mutex.lock();
      pause = false;
      condition.wakeOne();
      mutex.unlock();
    }
}

void ComputeThread::quitRun()
{
  if (isRunning())
    {
      mutex.lock();
      quit = true;
      condition.wakeOne();
      mutex.unlock();
    }
}

void ComputeThread::run()
{
  int        setExcepts = 0;
  static int stepVal    = 1;
  int        done       = 0;
#ifdef DEBUG
  QElapsedTimer runTime;
#endif
  while (1)
    {
      setExcepts = 0;
      stepVal    = 1;
      done       = 0;

      mutex.lock();
      pause          = false;
      quit           = false;
      readyForSignal = true;
      mutex.unlock();

#ifdef DEBUG
      qDebug() << "ComputeThread::run: Starting run";
      runTime.start();
#endif

      while (!done)
        {
          if (abort) return;
          if (pause) condition.wait(&mutex);

          if (stepVal)
            {
              clearFPU();
              done       = model->cycleInit();
              setExcepts = testFPU();

              if (testExcepts(setExcepts))
                {
#ifdef DEBUG
                  qDebug() << "ComputeThread::run: calling processFPEerrors ";
#endif
                  emit processFPEerrors("model->cycleInit()", setExcepts);
                  break;
                }
              else if (done & (WARNING | FATAL_ERROR | NON_FATAL_ERROR | LIB_NOT_LOADED))
                {
#ifdef DEBUG
                  qDebug() << "ComputeThread::run: calling processErrors ";
#endif
                  emit processErrors("model->cycleInit()", done);
                }

              if (done & FATAL_ERROR) break;

              // Clear the WARNING and NON_FATAL_ERROR flags: they should not
              // stop the integration
              done = done & (~WARNING) & (~NON_FATAL_ERROR);
            }

          clearFPU();
          stepVal    = model->cycleStep();
          setExcepts = testFPU();
          clearFPU();

          if (stepVal & NORMAL_OUT)
            {
              for (uint colNr = 0; colNr < model->outputVarNum() + 1; colNr++) (*outputData)[colNr].appendValue(model->outputData()[colNr]);
            }

          if (stepVal & STATE_OUT)
            {
#ifdef DEBUG
              qDebug() << "ComputeThread::run: signaling state and normal output ";
#endif
              emit updateStateGraphs();
              emit updateOutputImmediate();
            }
          else if (stepVal & NORMAL_OUT)
            {
              if (readyForSignal)
                {
                  readyForSignal = false;
#ifdef DEBUG
                  qDebug() << "ComputeThread::run: signaling delayed normal output ";
#endif
                  emit updateOutputDelayed();
                }
            }

          if (testExcepts(setExcepts))
            {
#ifdef DEBUG
              qDebug() << "ComputeThread::run: calling processFPEerrors ";
#endif
              emit processFPEerrors("model->cycleStep()", setExcepts);
              break;
            }
          else if (stepVal & (WARNING | FATAL_ERROR | NON_FATAL_ERROR | LIB_NOT_LOADED))
            {
#ifdef DEBUG
              qDebug() << "ComputeThread::run: calling processErrors ";
#endif
              emit processErrors("model->cycleStep()", stepVal);
              if (stepVal & FATAL_ERROR) break;
            }

          if (quit) break;
        }
#ifdef DEBUG
      qDebug() << "ComputeThread::run: Ending run";
      qDebug("Time elapsed: %lld ms", runTime.elapsed());
#endif

      emit runEnded();

      mutex.lock();
      if (!restart) condition.wait(&mutex);
      restart = false;
      mutex.unlock();
    }
}
