#include "compileform.h"

#include <qvariant.h>
#include <qimage.h>
#include <qpixmap.h>

#include "ebtdefines.h"
#include <QFileDialog>
#include <QTextStream>
#include "qmessagebox.h"
#include "qregularexpression.h"
#include <QSettings>
#include "qapplication.h"

#ifdef DEBUG
#include <QtDebug>
#endif

/*
 *  Constructs a CompileForm as a child of 'parent', with the
 *  name 'name' and widget flags set to 'f'.
 *
 *  The dialog will by default be modeless, unless you set 'modal' to
 *  true to construct a modal dialog.
 */

CompileForm::CompileForm(QWidget* parent, bool modal, Qt::WindowFlags fl)
    : QDialog(parent, fl)
{
    setupUi(this);
    setModal(modal);

    init();
}

/*
 *  Destroys the object and frees any allocated resources
 */

CompileForm::~CompileForm()
{
    // no need to delete child widgets, Qt does it all for us
  if (procList)
    {
      while (procList->size() > 0)
	{
	  delete procList->first();
	  procList->removeAt(0);
	}
      delete procList;
    }
}

/*
 *  Sets the strings of the subwidgets using the current
 *  language.
 */
void CompileForm::languageChange()
{
    retranslateUi(this);
}

void CompileForm::init()
{
    enableConfigure(FALSE);
    enableCompile(TRUE);
    compilationSuccessful = FALSE;
    connect(this, SIGNAL(unloadLib()), parentWidget(), SLOT(modelUnload()));
    comboCompiler->insertItems(0, compilerList);
    procList = NULL;
}

void CompileForm::changeConfiguration()
{
    enableConfigure(TRUE);
    enableCompile(FALSE);
}

void CompileForm::saveSettings()
{
    EBTDir =  editEBTDir->text();
    integratorPath =  EBTDir + "/integrator";
    compiler = comboCompiler->currentText();
    CCFlags = editCCFlags->text().split(" ", Qt::SkipEmptyParts);
    LDFlags = editLDFlags->text().split(" ", Qt::SkipEmptyParts);
    tmpPath = editTmpPath->text();
    
    QSettings settings;

    settings.setValue("EBTDir", EBTDir);

    settings.beginGroup("Compiler");
    settings.setValue("compiler", compiler);
    if(programType == CompileForm::model)
      settings.setValue("modelCCFlags", CCFlags);
    else
      settings.setValue("otherCCFlags", CCFlags);
    settings.setValue("LDFlags", LDFlags);
    settings.setValue("tmpPath", tmpPath);
    settings.endGroup();

    enableConfigure(FALSE);
    enableCompile(TRUE);
}

void CompileForm::cancelConfigure()
{
    EBTDir = oldEBTDir;
    compiler = oldCompiler;
    CCFlags = oldCCFlags;
    LDFlags = oldLDFlags;
    tmpPath = oldTmpPath;

    showSettings();
    enableConfigure(FALSE);
    enableCompile(TRUE);
}

void CompileForm::changeEBTDir()
{
  QString pName = QFileDialog::getExistingDirectory(
            this,
            "Choose the path to where EBT is installed",
	    EBTDir,
	    QFileDialog::ShowDirsOnly | QFileDialog::DontResolveSymlinks);
    
    if (!pName.isEmpty())
	editEBTDir->setText(pName);
}


void CompileForm::changeTmpPath()
{
  QString tName = QFileDialog::getExistingDirectory(
            this,
	    "Choose the directory where to save\ntemporary files (error log, object files)",
            tmpPath,
	    QFileDialog::ShowDirsOnly | QFileDialog::DontResolveSymlinks);
    
  if (!tName.isEmpty())
    editTmpPath->setText(tName);
}

void CompileForm::enableConfigure(bool enabled)
{
    oldEBTDir = EBTDir;
    oldCompiler = compiler;
    oldCCFlags = CCFlags;
    oldLDFlags = LDFlags;
    oldTmpPath = tmpPath;
    
    editEBTDir->setEnabled(enabled);
    buttonChangeEBTDir->setEnabled(enabled);
    comboCompiler->setEnabled(enabled);
    editCCFlags->setEnabled(enabled);
    editLDFlags->setEnabled(enabled);
    editTmpPath->setEnabled(enabled);
    buttonChangeTmpPath->setEnabled(enabled);
    buttonConfigure->setEnabled(!enabled);
    buttonDefaults->setEnabled(enabled);
    buttonSave->setEnabled(enabled);
    buttonCancel->setEnabled(enabled);
}

void CompileForm::enableCompile(bool enabled)
{
    buttonCompile->setEnabled(enabled);
    buttonHelp->setEnabled(enabled);
    buttonClean->setEnabled(enabled);
    buttonClose->setEnabled(enabled);
}

void CompileForm::compile()
{
  compilationSuccessful = FALSE;
  objectIterator        = 0;

  if (procList)
    {
      while (procList->size() > 0)
        {
          delete procList->first();
          procList->removeAt(0);
        }
      delete procList;
    }
  procList = new QList<QProcess *>;

  viewOutput->clear();
  enableCompile(FALSE);

  if (!QDir(integratorPath).exists() || integratorPath.isEmpty())
    {
      QMessageBox::critical(this, QString(EBTCAPTION + "EBT dir"),
                            QString("EBTtool can't find '%1'.\nPlease check if EBT dir is set correctly.").arg(integratorPath));
      tabWidget->setCurrentIndex(1);
      enableCompile(TRUE);
      return;
    }
  if (!QDir(tmpPath).exists() || tmpPath.isEmpty())
    {
      QMessageBox::critical(this, QString(EBTCAPTION + "Tmp path"),
                            QString("Tmp path '%1' does not point to an existing directory.\nPlease enter a directory where EBTtool can store temporary files.").arg(tmpPath));
      tabWidget->setCurrentIndex(1);
      enableCompile(TRUE);
      return;
    }

  tabWidget->setCurrentIndex(0);
  allFound = TRUE;
  QStringList::Iterator it;

  switch (programType)
    {
      case CompileForm::model:
        for (it = hModules.begin(); it != hModules.end(); ++it) checkFile(integratorPath + "/" + *it + ".h");
        checkFile(integratorPath + "/" + "escbox.h");
        for (it = cModules.begin(); it != cModules.end(); ++it) checkFile(integratorPath + "/" + *it + ".c");
        break;
      case CompileForm::defidcard:
        checkFile(integratorPath + "/" + defidcardH);
        break;
      case CompileForm::defoutput:
        checkFile(integratorPath + "/" + defoutputH);
        break;
    }
  if (!allFound)
    {
      viewOutput->append(QString("\nTerminating.\n"));
      enableCompile(TRUE);
      return;
    }

  if (!QDir(programDir).exists() || programDir.isEmpty())
    {
      viewOutput->append(QString("\nModel directory %1 not found.\n").arg(programDir));
      allFound = FALSE;
    }

  if (!allFound)
    {
      viewOutput->append(QString("\nTerminating.\n"));
      enableCompile(TRUE);
      return;
    }

  oldDir = QDir::currentPath();
  QDir::setCurrent(programDir);

  if (programType == CompileForm::model) checkFile(programH);
  checkFile(programC);
  if (!allFound)
    {
      viewOutput->append(QString("\nTerminating.\n"));
      QDir::setCurrent(oldDir);
      enableCompile(TRUE);
      return;
    }

  if (QFile(programO).exists())
    if (QFileInfo(programO).lastModified() < QFileInfo(programH).lastModified() ||
        QFileInfo(programO).lastModified() < QFileInfo(programC).lastModified())
      QFile(programO).remove();

  if (QFile(programSO).exists())
    {
      if (QFileInfo(programSO).lastModified() < QFileInfo(programH).lastModified() ||
          QFileInfo(programSO).lastModified() < QFileInfo(programC).lastModified())
        {
          emit unloadLib();
          QFile(programSO).remove();
        }
    }

  if (QFile(programSO).exists())
    {
      if (QFile(programO).exists())
        {
          if (QFileInfo(programSO).lastModified() < QFileInfo(programO).lastModified())
            {
              emit unloadLib();
              QFile(programSO).remove();
            }
        }
      else
        {
          emit unloadLib();
          QFile(programSO).remove();
        }
    }

  if (QFile(programSO).exists())
    {
      viewOutput->append(QString("\nThe file %1 is up-to-date.\n").arg(programSO));
      compilationSuccessful = TRUE;
      QDir::setCurrent(oldDir);
      enableCompile(TRUE);
      return;
    }

  if (QFile(programO).exists())
    {
      viewOutput->append(QString("\nThe file %1 is up-to-date.\n").arg(programO));
      if (programType == CompileForm::model)
        compileObject(objectIterator);
      else
        compileSO();
    }
  else
    compileModelO();
}

void CompileForm::cleanModel()
{
  QMessageBox::StandardButton ans = QMessageBox::question(this, QString(EBTCAPTION + "Cleanup program files"),
                                                          QString("Are you sure you want to delete\nthe compiled files of program %1?").arg(program),
                                                          QMessageBox::Yes | QMessageBox::No, QMessageBox::No);
  if (ans == QMessageBox::No) return;

  emit unloadLib();
  viewOutput->clear();
  tabWidget->setCurrentIndex(0);

  oldDir = QDir::currentPath();
  if (!QDir(programDir).exists() || programDir.isEmpty())
    {
      viewOutput->append(QString("\nProgram directory %1 not found.\n").arg(programDir));
      return;
    }
  QDir::setCurrent(programDir);
  if (programType == CompileForm::model) deleteFile(program);
  deleteFile(programO);
  deleteFile(programSO);
#ifdef Q_OS_WIN
  deleteFile(program + "module.exp");
  deleteFile(program + "module.lib");
  deleteFile(program + "module.tds");
  deleteFile(".obj.obj");
#endif
  QDir::setCurrent(oldDir);
  viewOutput->append("\nDone.\n");
}

void CompileForm::setModel(QString programName)
{
    programType = CompileForm::model;
    loadSettings();
    initCommon(programName);
    initModel();
}

void CompileForm::checkFile(QString fName)
{
    if(!QFile(fName).exists())
    {
	viewOutput->append(QString("\nError: file %1 not found.\n")
			   .arg(fName));
	allFound = FALSE;
    }
}

void CompileForm::deleteFile(QString fName)
{
    if(QFile(fName).exists())
    {
	if(QFile(fName).remove())
	    viewOutput->append(QString("\nRemoving file %1.").arg(fName));
	else
	    viewOutput->append(QString("\nRemoving file %1 failed.").arg(fName));
	
    }
}

void CompileForm::compileModelO()
{
  errOutput  = outOutput = QString();
  viewOutput->append(QString("\nBuilding %1:\tCompiling %2..........")
		     .arg(programO)
		     .arg(programC));

  QProcess *proc = new QProcess(this);
  QStringList args;

  procList->append(proc);
  args = CCFlags;

  args.append("-DMODULE");
  if(programType == CompileForm::model)
    {
      args.append("-D__USE_FIXED_PROTOTYPES__");
      args.append("-DPROBLEMFILE=\"" + programH + "\"");
    }
  args.append("-I.");
  args.append("-I" + integratorPath);
#if (defined(Q_OS_WIN) && defined(_MSC_VER))
  args.append("/Fo" + programO);
#else
    args.append("-o" + programO);
#endif
  args.append("-c");
  if ((compiler == clangCC) || (compiler == gnuCC) || (compiler == MingwCC))
    {
      args.append("-x");
      args.append("c++");
    }
  if(programType == CompileForm::model)
    args.append(programC);
  else
    args.append(tmpPath + "/" + programC);	

#ifdef Q_OS_WIN
  connect(proc, SIGNAL(readyReadStandardOutput()), this, SLOT(readFromStdout()));
#endif    
  connect(proc, SIGNAL(readyReadStandardError()), this, SLOT(readFromStderr()));
  connect(proc, SIGNAL(finished(int)), this, SLOT(compileModelOExited(int)));

  proc->start(compiler, args);
  if(proc->state() == QProcess::NotRunning)
    {
      QMessageBox::critical(this, QString(EBTCAPTION + "error"), "Could not start the building process.");
      quitCompilation();
      return;
    }
}

void CompileForm::compileModelOExited(int exitCode)
{
  if(exitCode)
    {
      viewOutput->append("\nErrors!\n" + errOutput);
#ifdef Q_OS_WIN
      viewOutput->append("\n" + outOutput); 
#endif
      quitCompilation();
      return;
    }
  if((errOutput.isEmpty()) && (outOutput.isEmpty()))
    viewOutput->append("Done.\n");
  else
    {
      if(!errOutput.isEmpty())
	viewOutput->append("\nWarnings!\n" + errOutput);
      if(!outOutput.isEmpty())
	viewOutput->append("\n" + outOutput);
     }

  errOutput = outOutput = QString();

  if(programType == CompileForm::model)
    compileMethTest();
  else
    compileSO();
}

void CompileForm::compileMethTest()
{
  QProcess *proc = new QProcess(this);
  QStringList args;

  procList->append(proc);
  args = CCFlags;

  args.append("-DMODULE");
  args.append("-D__USE_FIXED_PROTOTYPES__");
  args.append("-DPROBLEMFILE=\"" + programH + "\"");
  args.append("-I.");
  if(compiler == MSVC)
    args.append("/Fo");
  else
    args.append("-o");
#ifdef Q_OS_WIN
  args.append("NUL");
#else
  args.append("/dev/null");
  if ((compiler == clangCC) || (compiler == gnuCC) || (compiler == MingwCC))
    {
      args.append("-x");
      args.append("c++");
    }
#endif
  args.append(integratorPath + "/methtest.c");

  connect(proc, SIGNAL(readyReadStandardError()), this, SLOT(readFromStderr()));
  connect(proc, SIGNAL(finished(int)), this, SLOT(compileMethTestExited()));

  proc->start(compiler, args);
  if(proc->state() == QProcess::NotRunning)
    {
      QMessageBox::critical(this, QString(EBTCAPTION + "error"), "Could not start the building process.");
      quitCompilation();
      return;
    }
}

void CompileForm::compileMethTestExited()
{
  if ((compiler == gnuCC) || (compiler == MingwCC))
    {
      errOutput.replace(QRegularExpression(".*methtest.*error: #error ==> "), "");
      errOutput.replace(QRegularExpression("#error .*"), "");
      viewOutput->append(errOutput);
    }
  else if (compiler == clangCC)
    {
      errOutput.replace(QRegularExpression(".*methtest.*error: ==> "), "");
      errOutput.replace(QRegularExpression("#error .*"), "");
      viewOutput->append(errOutput);
    }
  else
    viewOutput->append("\n" + errOutput.replace(QRegularExpression(".* #error "), ""));

  errOutput = outOutput = QString();
  compileObject(objectIterator);
}

void CompileForm::compileObject(int objIt)
{
  viewOutput->append(QString("\nCompiling %1..........").arg(oModules[objIt] + ".c"));

  QProcess *proc = new QProcess(this);
  QStringList args;

  procList->append(proc);
  args = CCFlags;

  args.append("-DMODULE");
  args.append("-D__USE_FIXED_PROTOTYPES__");
  args.append("-DPROBLEMFILE=\"" + programH + "\"");
  args.append("-I.");
  args.append("-I" + integratorPath);
  args.append("-I" + integratorPath + intMethDir);
#if (defined(Q_OS_WIN) && defined(_MSC_VER))
  args.append("/Fo" + tmpPath + "/" + oModules[objIt] + ".obj");
#else
    args.append("-o" + tmpPath + "/" + oModules[objIt] + ".o");
#endif
  args.append("-c");
  if ((compiler == clangCC) || (compiler == gnuCC) || (compiler == MingwCC))
    {
      args.append("-x");
      args.append("c++");
    }
  args.append(integratorPath + "/" + oModules[objIt] + ".c"); 

#ifdef Q_OS_WIN
  connect(proc, SIGNAL(readyReadStandardOutput()), this, SLOT(readFromStdout()));
#endif    
  connect(proc, SIGNAL(readyReadStandardError()), this, SLOT(readFromStderr()));
  connect(proc, SIGNAL(finished(int)), this, SLOT(compileObjectExited(int)));

  proc->start(compiler, args);
  if(proc->state() == QProcess::NotRunning)
    {
      QMessageBox::critical(this, QString(EBTCAPTION + "error"), "Could not start the building process.");
      quitCompilation();
      return;
    }
}

void CompileForm::compileObjectExited(int exitCode)
{
  if(exitCode)
    {
      viewOutput->append("\nErrors!\n" + errOutput);
#ifdef Q_OS_WIN
      viewOutput->append("\n" + outOutput); 
#endif
      for(QStringList::Iterator it = objects.begin(); it != objects.end(); ++it)
	if(QFile(tmpPath + "/" + *it).exists())
	  QFile(tmpPath + "/" + *it).remove();
      quitCompilation();
      return;
    }
  if((errOutput.isEmpty()) && (outOutput.isEmpty()))
    viewOutput->append("Done.\n");
  else
    {
      if(!errOutput.isEmpty())
	viewOutput->append("\nWarnings!\n" + errOutput);
      if(!outOutput.isEmpty())
	viewOutput->append("\n" + outOutput);
     }

  errOutput = outOutput = QString();

  ++objectIterator;
  if(objectIterator < oModules.size())
    compileObject(objectIterator);
  else
    compileSO();
}

void CompileForm::compileSO()
{
  errOutput = outOutput = QString();
  viewOutput->append(QString("\nLinking %1..........").arg(programSO));

  QProcess *proc = new QProcess(this);
  QStringList args;

  procList->append(proc);
#ifdef Q_OS_WIN
  if(compiler == MingwCC)
    {
      args.append("-shared");
      args.append("-Wl,-soname," + programSO);
      args.append("-o");
      args.append(programSO);
    }
  else	// Microsoft cl
    {
      args.append("/LD");		
      args.append("/Fe" + programSO);		
    }
#elif defined(Q_OS_MACOS)
  // Main changes was to use .dylib file instead of .so file,
  // this needed some new flags in the compiler call "-Wl,-intall_name:" +programSO
  // I replaced with -dynamiclib and -Wl,-soname -shared. Then module.so is replaced with module.dylib.
  args.append("-dynamiclib");
  args.append("-Wl,-install_name," + programSO);
  args.append("-o");
  args.append(programSO);
#else
  args.append("-shared");
  args.append("-Wl,-soname," + programSO);
  args.append("-o");
  args.append(programSO);  
#endif
  if(programType == CompileForm::model)
    for(QStringList::Iterator it = objects.begin(); it != objects.end(); ++it ) 
      args.append(tmpPath + "/" + *it);
  args.append(programO);
  if ((compiler != MSVC) && (compiler != MingwCC))
    args.append("-lc");
  for(QStringList::Iterator it = LDFlags.begin(); it != LDFlags.end(); ++it ) 
    args.append(*it);

#ifdef Q_OS_WIN
  connect(proc, SIGNAL(readyReadStandardOutput()), this, SLOT(readFromStdout()));
#endif    
  connect(proc, SIGNAL(readyReadStandardError()), this, SLOT(readFromStderr()));
  connect(proc, SIGNAL(finished(int)), this, SLOT(compileSOExited(int)));

  proc->start(compiler, args);
  if(proc->state() == QProcess::NotRunning)
    {
      QMessageBox::critical(this, QString(EBTCAPTION + "error"), "Could not start the building process.");
      quitCompilation();
      return;
    }
}

void CompileForm::compileSOExited(int exitCode)
{
  if(!exitCode)
    {
      if(errOutput.isEmpty())
     	viewOutput->append("Done.\n");
      else
     	{
	  viewOutput->append("\nWarnings!\n" + errOutput);
	}
      compilationSuccessful = TRUE;
    }
  else
    {
      viewOutput->append("\nErrors!\n" + errOutput);
#ifdef Q_OS_WIN
      viewOutput->append("\n" + outOutput); 
#endif
    }

  if(programType == CompileForm::model)	
    for(QStringList::Iterator it = objects.begin(); it != objects.end(); ++it ) 
      if(QFile(tmpPath + "/" + *it).exists())
	QFile(tmpPath + "/" + *it).remove();

  errOutput = outOutput = QString();
  QDir::setCurrent(oldDir);
  enableCompile(TRUE);
}

void CompileForm::readFromStderr()
{
  QByteArray msg = ((QProcess *)sender())->readAllStandardError();
  QString msg2 = QString::fromLocal8Bit(msg.constData());

  errOutput.append(msg2);
}

void CompileForm::readFromStdout()
{
  QByteArray msg = ((QProcess *)sender())->readAllStandardOutput();
  QString msg2 = QString::fromLocal8Bit(msg.constData());

  outOutput.append(msg2);
}

void CompileForm::closeForm()
{
  if (procList)
    {
      while (procList->size() > 0)
	{
	  delete procList->first();
	  procList->removeAt(0);
	}
      delete procList;
    }
  procList = NULL;

  if(compilationSuccessful)
    this->done(QDialog::Accepted);
  else
    this->done(QDialog::Rejected);
}

void CompileForm::quitCompilation()
{
    errOutput = outOutput = QString();
    QDir::setCurrent(oldDir);
    enableCompile(TRUE);
}


void CompileForm::loadSettings()
{
    bool ok;
    QSettings settings;

#ifndef DEBUG
    ok = settings.contains("EBTDir");
    if (ok) EBTDir = settings.value("EBTDir").toString();
    else
#endif
      EBTDir = qApp->applicationDirPath();

    settings.beginGroup("Compiler");
    ok = settings.contains("compiler");
    if (ok) compiler = settings.value("compiler").toString();
    if (!ok || compilerList.indexOf(compiler) < 0) compiler = defCompiler;

    if(programType == CompileForm::model)
      {
        ok = settings.contains("modelCCFlags");
	if (ok) CCFlags = settings.value("modelCCFlags").toStringList();
      }
    else
      {
        ok = settings.contains("otherCCFlags");
	if (ok) CCFlags = settings.value("otherCCFlags").toStringList();
      }

    if(!ok)
      {
	CCFlags = defCCFlags().split(" ", Qt::SkipEmptyParts);
      }

    ok = settings.contains("LDFlags");
    if (ok) LDFlags = settings.value("LDFlags").toStringList();
    else LDFlags = defLDFlags().split(" ", Qt::SkipEmptyParts);

    ok = settings.contains("tmpPath");
    if (ok) tmpPath = settings.value("tmpPath").toString();
    else
      {
#ifdef Q_OS_WIN
	tmpPath = getenv("TMP");
	if (tmpPath.isEmpty())
	    tmpPath = getenv("TEMP");
	if (tmpPath.isEmpty())
#endif	    
	  tmpPath = defTmpPath;
      }
    integratorPath = EBTDir + "/integrator";

    settings.endGroup();

    showSettings();
}


void CompileForm::defaultSettings()
{
    comboCompiler->setCurrentIndex(compilerList.indexOf(defCompiler));
    compilerChanged(defCompiler);
    editCCFlags->setText(defCCFlags());
#ifdef Q_OS_WIN
    tmpPath = getenv("TMP");
    if (tmpPath.isEmpty())
      tmpPath = getenv("TEMP");
    if (tmpPath.isEmpty())
#endif
      editTmpPath->setText(defTmpPath);
    editEBTDir->setText(qApp->applicationDirPath());
    editLDFlags->setText(defLDFlags());
}


void CompileForm::showSettings()
{
    editEBTDir->setText(EBTDir);
    comboCompiler->setCurrentIndex(compilerList.indexOf(compiler));
    editCCFlags->setText(CCFlags.join(" "));
    editLDFlags->setText(LDFlags.join(" "));
    editTmpPath->setText(tmpPath);
}


void CompileForm::setDefIDcard(QString programName)
{
    programType = CompileForm::defidcard;
    loadSettings();
    initCommon(programName);
    initDefIDcard();
}


void CompileForm::initModel()
{
    intMethDir = "/Odesolvers";
    
    hModules <<  "ebtmain" << "ebtinit" << "ebtcohrt" << "ebttint" << "ebtutils" << "ebtstop";
    cModules << "ebtmain" << "ebtinit" << "ebtcohrt" << "ebttint" 
	    << (intMethDir + "/ebtrk4") << (intMethDir + "/ebtrkf45") << (intMethDir + "/ebtrkck") 
	    << (intMethDir + "/ebtdopri5") << (intMethDir  + "/ebtdopri8") << (intMethDir + "/ebtradau")
	    << "ebtutils" << "ebtstop" << "methtest";
    // oModules << "ebtmain" << "ebtinit" << "ebtcohrt" << "ebttint" << "ebtutils" << "ebtstop";
    oModules << "ebtcohrt" << "ebtmain" << "ebtinit" << "ebttint" << "ebtutils" << "ebtstop";
    
#if (defined(Q_OS_WIN) && defined(_MSC_VER))
    objects << "ebtmain.obj" << "ebtinit.obj" << "ebtcohrt.obj" << "ebttint.obj" << "ebtutils.obj" 
	    << "ebtstop.obj";
#else
    objects << "ebtmain.o" << "ebtinit.o" << "ebtcohrt.o" << "ebttint.o" << "ebtutils.o" << "ebtstop.o";
#endif
    //    hTemplate = "/templates/ebttmpl.h";
    //    cTemplate = "/templates/ebttmpl.c";
}

void CompileForm::initCommon(QString programName)
{
    program = QFileInfo(programName).completeBaseName();
    programDir = QFileInfo(programName).absolutePath();
    
    setWindowTitle(EBTCAPTION + "compile " + program);
    
    programC = program + ".c";
    programH = program + ".h";	// only functional with models
    
#if (defined(Q_OS_WIN) && defined(_MSC_VER))
    programO = program + "module.obj";
    programSO = program + "module.dll";
#elif (defined(Q_OS_WIN) && defined(__MINGW32__))
    programO = program + "module.o";
    programSO = program + "module.dll";
#elif defined(Q_OS_MACOS)
    programO = program + "module.o";
    programSO = program + "module.dylib";
#else
    programO = program + "module.o";
    programSO = program + "module.so";
#endif
}


void CompileForm::initDefIDcard()
{
  QString     tmpSrc = QString();
  QStringList lines;
  QFile       srcFile(programDir + "/" + programC);
  defidcardH = "DefineIDcard.h";

  if (srcFile.open(QIODevice::ReadOnly | QIODevice::Text))
    {
      QTextStream stream(&srcFile);
      while (!stream.atEnd()) lines.append(stream.readLine());
      srcFile.close();
    }
  else
    {
      QMessageBox::critical(this, QString(EBTCAPTION + "No such program file"),
                            QString("The program file %1 could not be found.").arg(programDir + "/" + programC));
      return;
    }
  QRegularExpression regExp("define[ \t]*I_STATE_DIM");
  tmpSrc.append(lines.filter(regExp).join("\n") + "\n");
  regExp.setPattern("define[ \t]*NEW_CONST_DIM");
  tmpSrc.append(lines.filter(regExp).join("\n") + "\n");
  tmpSrc.append("#include <" + defidcardH + ">\n");
  tmpSrc.append("#line 1\n");
  regExp.setPattern("#define[ \t]*I_STATE_DIM.*$");
  lines.replaceInStrings(regExp, "");
  regExp.setPattern("#define[ \t]*NEW_CONST_DIM.*$");
  lines.replaceInStrings(regExp, "");
  lines.replaceInStrings("popIDcard", "pop");
  regExp.setPattern("pop\\[([^]]*)\\]\\[([^]]*)\\]\\[([^]]*)\\]");
  lines.replaceInStrings(regExp, "(*(pop[\\1]+(\\3)*cohort_no[\\1]+\\2))");
  regExp.setPattern("newIDcard\\[([^]]*)\\]\\[([^]]*)\\]\\[([^]]*)\\]");
  lines.replaceInStrings(regExp, "(*(newIDcard[\\1]+(\\3)*cohort_no[\\1]+\\2))");

  tmpSrc.append(lines.join("\n"));
  tmpSrc.append("\n");

  QFile tmpFile(tmpPath + "/" + programC);
  if (tmpFile.open(QIODevice::WriteOnly | QIODevice::Truncate))
    {
      QTextStream stream(&tmpFile);
      stream << tmpSrc;
      tmpFile.close();
    }
  else
    QMessageBox::critical(this, QString(EBTCAPTION + "couldn't create temp file"),
                          QString("Failed to create temporary file '%1' in %2").arg(programC).arg(tmpPath));
}


void CompileForm::setDefOutput(QString programName)
{
    programType = CompileForm::defoutput;
    loadSettings();
    initCommon(programName);
    initDefOutput();
}


void CompileForm::initDefOutput()
{
  QString     tmpSrc = QString();
  QStringList lines;
  QFile       srcFile(programDir + "/" + programC);
  defoutputH = "DefineOutput.h";

  if (srcFile.open(QIODevice::ReadOnly | QIODevice::Text))
    {
      QTextStream stream(&srcFile);
      while (!stream.atEnd()) lines.append(stream.readLine());
      srcFile.close();
    }
  else
    {
      QMessageBox::critical(this, QString(EBTCAPTION + "No such program file"),
                            QString("The program file %1 could not be found.").arg(programDir + "/" + programC));
      return;
    }
  QRegularExpression regExp("define[ \t]*I_STATE_DIM");
  tmpSrc.append(lines.filter(regExp).join("\n") + "\n");
  regExp.setPattern("define[ \t]*OUTPUT_VAR_NR");
  tmpSrc.append(lines.filter(regExp).join("\n") + "\n");
  tmpSrc.append("#include <" + defoutputH + ">\n");
  tmpSrc.append("#line 1\n");
  regExp.setPattern("#define[ \t]*I_STATE_DIM.*$");
  lines.replaceInStrings(regExp, "");
  regExp.setPattern("#define[ \t]*OUTPUT_VAR_NR.*$");
  lines.replaceInStrings(regExp, "");
  lines.replaceInStrings("popIDcard", "pop");
  regExp.setPattern("pop\\[([^]]*)\\]\\[([^]]*)\\]\\[([^]]*)\\]");
  lines.replaceInStrings(regExp, "(*(pop[\\1]+(\\3)*cohort_no[\\1]+\\2))");

  tmpSrc.append(lines.join("\n"));
  tmpSrc.append("\n");

  QFile tmpFile(tmpPath + "/" + programC);
  if (tmpFile.open(QIODevice::WriteOnly | QIODevice::Truncate))
    {
      QTextStream stream(&tmpFile);
      stream << tmpSrc;
      tmpFile.close();
    }
  else
    QMessageBox::critical(this, QString(EBTCAPTION + "couldn't create temp file"),
                          QString("Failed to create temporary file '%1' in %2").arg(programC).arg(tmpPath));
}


QString CompileForm::defCCFlags()
{
    if(programType == CompileForm::model)
    {
	if      (compiler == clangCC)
	  return defModelCCFlagsClangCC;
	else if (compiler == intelCC)
	  return defModelCCFlagsIntelCC;
	else if (compiler == gnuCC)
#if (defined(Q_OS_WIN) && defined(__MINGW32__))
	  return defModelCCFlagsMingwCC;
#else
	  return defModelCCFlagsGnuCC;
#endif
	else if (compiler == MingwCC)
	  return defModelCCFlagsMingwCC;
	else 						// if (compiler == MSVC)
	  return defModelCCFlagsMSVC;
    }
    else
    {
	if      (compiler == clangCC)
	  return defOtherCCFlagsClangCC;
	else if (compiler == intelCC)
	  return defOtherCCFlagsIntelCC;
	else if (compiler == gnuCC)
#if (defined(Q_OS_WIN) && defined(__MINGW32__))
	  return defOtherCCFlagsMingwCC;
#else
	  return defOtherCCFlagsGnuCC;
#endif
	else if (compiler == MingwCC)
	  return defOtherCCFlagsMingwCC;
	else 						// if (compiler == MSVC)
	  return defOtherCCFlagsMSVC;
   }
}

QString CompileForm::defLDFlags()
{
  if      (compiler == clangCC)
    return defLDFlagsClangCC;
  else if (compiler == intelCC)
    return defLDFlagsIntelCC;
  else if (compiler == gnuCC)
    return defLDFlagsGnuCC;
  else if (compiler == MingwCC)
    return defLDFlagsMingwCC;
  else 						// if (compiler == MSVC)
    return defLDFlagsMSVC;
}

void CompileForm::compilerChanged( const QString &str )
{
    compiler = str;
    CCFlags = defCCFlags().split(" ", Qt::SkipEmptyParts);
    LDFlags = defLDFlags().split(" ", Qt::SkipEmptyParts);
    showSettings();
}

