#include <qmessagebox.h>
#include "ebtcsbfile.h"
#include "integrator/ebttooldefs.h"
#include "integrator/ebtcsbdefs.h"

#ifdef DEBUG
#include <QtDebug>
#endif

EBTCSBFile::EBTCSBFile()
{
    defidcard = new EBTDefIDcard;
    defoutput = new EBTDefOut;

    setTemplateFile(QString());
    setFileExtension("out");
    setDescriptionText("Output values");
    setSaveText(descriptionText() + " (*." + fileExtension() + ")");
    setOpenText(saveText());

    file = NULL;
    csbPar = NULL;
    columnData = NULL;
    newConstDim = 0;
    popIDcardSize = 0;
    currentPopIDcard = -1;

    outputVarNr = 0;
    compOutputSize = -1;
}

EBTCSBFile::~EBTCSBFile()
{
    clearColumnData();
    clearIDcard();
    clearCompOutput();

    if(csbPar) delete [] csbPar;

//    if (defidcard) delete defidcard;		Generates error message on application close
//    if (defoutput) delete defoutput;		Id.

    if (file) delete file;			// shouldn't be necessary
}

bool EBTCSBFile::open()
{
    bool openOK = EBTFile::open();
    if(openOK)
	scanOK = FALSE;
    return openOK;
}

void EBTCSBFile::scan()
{
  if(!checkFile()) return;

  if (!file->open(QIODevice::ReadOnly))
    {
      QMessageBox::critical(0, EBTCAPTION + "Can't open",
			    QString("Cannot open file %1")
			    .arg(fileName()));
      return;
    }
  if(file->size() == 0)
    {
      QMessageBox::critical(0, EBTCAPTION + "No data",
			    QString("File %1 does not contain any data")
			    .arg(fileName()));
      file->close();
      return;
    }

  uint32_t magic;
  char data[sizeof(Envdim)];

  if(file->read(data, sizeof(magic)) > 0)
    magic = *(uint32_t*)(data);
  else
    {
      fileCorrupt();
      return;
    }

  uint32_t filePosition = 0;
  int numPars;
  bool noErrors =TRUE;

  switch (magic)
    {
      case CSB_MAGIC_KEY1:
	// CSB file version dd. May 09, 2003 with array of parameter values at
	// the start of the file
	if(file->read(data, sizeof(numPars)) > 0)
	  numPars = *(int*)(data);
	else
	  noErrors = FALSE;

	if (noErrors)
	  {
	    if(csbPar)
	      {
		delete [] csbPar;
		csbPar = NULL;
	      }
	    csbPar = new double[numPars];
	    double parameter;
	    for(int ii = 0; ii < numPars; ii++)
	      {
		if(file->read(data, sizeof(parameter)) > 0)
		  {
		    parameter = *(double*)(data);
		    csbPar[ii] = parameter;
		  }
		else
		  {
		    noErrors = FALSE;
		    break;
		  }
	      }
	  }
	if(!noErrors)
	  {
	    if(csbPar)
	      {
		delete [] csbPar;
		csbPar = NULL;
	      }
	    fileCorrupt();
	    return;
	  }
	filePosition = file->pos();	//beginning of states after parameters
	break;
      default:
	file->seek(filePosition);	//beginning of states at beginning of file
	break;
    }

  times = 0;
  Envdim currentEnv;
  stateRecordMap.clear();

  while (!file->atEnd() && noErrors)
    {
      if(file->read(data, sizeof(Envdim)) > 0)
	      currentEnv = *(Envdim*)(data);
      else
	      noErrors = FALSE;

      if (noErrors)
        {
          stateRecordMap[times].timeVal = currentEnv.timeval;
          stateRecordMap[times].filePos = filePosition;
          stateRecordMap[times].read    = FALSE;

          file->seek(filePosition + currentEnv.memory_used);
          times++;
          filePosition = file->pos();
        }
    }
  if (!noErrors)
    {
      if(csbPar)
	{
	  delete [] csbPar;
	  csbPar = NULL;
	}
      fileCorrupt();
      return;
    }

  QFileInfo fi = QFileInfo(*file);
  lastModified.setDate(fi.lastModified().date());
  lastModified.setTime(fi.lastModified().time());
  file->close();
  scanOK = TRUE;
  return;
}  // scan

void EBTCSBFile::fileCorrupt()
{
  QMessageBox::critical(0, EBTCAPTION + "Corrupt file",
			QString("File %1 appears to be invalid or corrupt")
			.arg(fileName()));
  stateRecordMap.clear();
  file->close();
}

void EBTCSBFile::unload()
{
    if(!checkFile()) return;

    clearColumnData();
    clearIDcard();
    // clearCompOutput();
    stateRecordMap.clear();
    // outputVarNr = 0;
    file->close();
    file = NULL;
}

bool EBTCSBFile::readState(uint stateNr)
{
  if(stateRecordMap[stateNr].read) return TRUE;

  if(!checkFile()) return FALSE;
  if(!file->isOpen())
    {
      if (!file->open(QIODevice::ReadOnly))
	{
	  QMessageBox::critical(0, EBTCAPTION + "Can't open",
				QString("Cannot open file %1")
				.arg(fileName()));
	  return FALSE;
	}
    }
  if(fileChanged()) return FALSE;

  file->seek(stateRecordMap[stateNr].filePos);

  char *data = new char[sizeof(Envdim)];
  Envdim currentEnv;

  if(file->read(data, sizeof(Envdim)) > 0)
    currentEnv = *(Envdim*)(data);
  else
    {
      fileCorrupt();
      delete [] data;
      return FALSE;
    }
  delete [] data;

  file->seek(stateRecordMap[stateNr].filePos);

  data = new char[currentEnv.memory_used];
  Popdim currentPop;
  double *basePoint;

  if(file->read(data, currentEnv.memory_used) > 0)
    {
      basePoint =  (double*)(data) + currentEnv.data_offset + currentEnv.columns;
      currentPop = *(Popdim*)basePoint;
    }
  else
    {
      fileCorrupt();
      delete [] data;
      return FALSE;
    }

  if(currentPop.data_offset > (int)(sizeof(Popdim)/sizeof(double) + 2))
    stateRecordMap[stateNr].description[currentPop.population] =
	QString((char *) (basePoint + (sizeof(Popdim)/sizeof(double) + 1)));	// ok? should be a '\0'-terminated char*..
  else
    stateRecordMap[stateNr].description[currentPop.population] = QString();
  stateRecordMap[stateNr].popCohorts[currentPop.population] = currentPop.cohorts;
  stateRecordMap[stateNr].popColumns[currentPop.population] = currentPop.columns;
  stateRecordMap[stateNr].numPopulations = 1;
  stateRecordMap[stateNr].maxCohorts = currentPop.cohorts;

  while(!currentPop.lastpopdim)
    {
      basePoint += currentPop.data_offset + (currentPop.columns * currentPop.cohorts);
      currentPop = *(Popdim*)basePoint;
      if(currentPop.data_offset > (int)(sizeof(Popdim)/sizeof(double) + 2))
	stateRecordMap[stateNr].description[currentPop.population] =
	    QString((char *) (basePoint + (sizeof(Popdim)/sizeof(double) + 1)));
      else
	stateRecordMap[stateNr].description[currentPop.population] = QString();
      stateRecordMap[stateNr].popCohorts[currentPop.population] = currentPop.cohorts;
      stateRecordMap[stateNr].popColumns[currentPop.population] = currentPop.columns;
      if(stateRecordMap[stateNr].maxCohorts < currentPop.cohorts)
	stateRecordMap[stateNr].maxCohorts = currentPop.cohorts;
      stateRecordMap[stateNr].numPopulations++;
    }

  newConstDim = computeIDcard(stateNr);	// called here in order to make newConstDim available for numColumns()
  if(outputVarNr > 0) computeOutput(stateNr);

  stateRecordMap[stateNr].read = TRUE;
  file->close();
  delete [] data;

  return TRUE;
}  // readState

double EBTCSBFile::timeVal(uint stateNr)
{
    return stateRecordMap[stateNr].timeVal;
}

uint EBTCSBFile::numPopulations(uint stateNr)
{
    if(!readState(stateNr))
	return 0;
    return stateRecordMap[stateNr].numPopulations;
}

uint EBTCSBFile::numColumns(uint stateNr, uint popNr)
{
    if(!readState(stateNr))
	return 0;
    if(popNr > (uint)stateRecordMap[stateNr].popColumns.size())
	return 0;

    return stateRecordMap[stateNr].popColumns[popNr] + newConstDim;
}

uint EBTCSBFile::numCohorts(uint stateNr, uint popNr)
{
    if(!readState(stateNr))
	return 0;
    if(popNr > (uint)stateRecordMap[stateNr].popCohorts.size())
	return 0;
    return stateRecordMap[stateNr].popCohorts[popNr];
}

QString EBTCSBFile::description(uint stateNr, uint popNr)
{
    if(!readState(stateNr))
	return QString();
    return stateRecordMap[stateNr].description[popNr];
}

void EBTCSBFile::column(uint stateNr, uint popNr, uint xCol, uint yCol, double* colPtr[2])
{
  colPtr[0] = NULL;
  colPtr[1] = NULL;

  if(!checkFile()) return;
  if(!file->isOpen())
    {
      if (!file->open(QIODevice::ReadOnly))
	{
	  QMessageBox::critical(0, EBTCAPTION + "Can't open",
				QString("Cannot open file %1")
				.arg(fileName()));
	  return;
	}
    }
  if(fileChanged()) return;

  newConstDim = computeIDcard(stateNr);

  file->seek(stateRecordMap[stateNr].filePos);

  char data[sizeof(Envdim)];
  Envdim currentEnv;

  if(file->read(data, sizeof(Envdim)) > 0)
    currentEnv = *(Envdim*)(data);
  else
    {
      fileCorrupt();
      return;
    }

  file->seek(stateRecordMap[stateNr].filePos);

  clearColumnData();
  columnData = new char[currentEnv.memory_used];
  Popdim currentPop;
  double *basePoint;

  if(file->read(columnData, currentEnv.memory_used) > 0)
    {
      basePoint =  (double*)(columnData) + currentEnv.data_offset + currentEnv.columns;
      currentPop = *(Popdim*)basePoint;
    }
  else
    {
      fileCorrupt();
      return;
    }

  while((currentPop.population != (int)popNr) && !currentPop.lastpopdim)
    {
      basePoint += currentPop.data_offset + (currentPop.columns * currentPop.cohorts);
      currentPop = *(Popdim*)basePoint;
    }
  if(currentPop.population == (int)popNr)
    {
      colPtr[0] = colPtr[1] = basePoint;
      if((int)xCol < currentPop.columns)
	colPtr[0] += currentPop.data_offset + (xCol * currentPop.cohorts);
      else if(newConstDim && (((int)xCol - currentPop.columns) < newConstDim))
	colPtr[0] = popIDcard[popNr] + (xCol - currentPop.columns) * currentPop.cohorts;
      if((int)yCol < currentPop.columns)
	colPtr[1] += currentPop.data_offset + (yCol * currentPop.cohorts);
      else if(newConstDim && (((int)yCol - currentPop.columns) < newConstDim))
	colPtr[1] = popIDcard[popNr] + (yCol - currentPop.columns) * currentPop.cohorts;
    }
  file->close();
}

bool EBTCSBFile::checkFile()
{
  if(file) delete file;
  if(fileName().isEmpty()) return FALSE;

  file = new QFile(absFilePath());
  return TRUE;
}

bool EBTCSBFile::fileChanged()
{
    // this function should only be called where checkFile() has already been called
    QFileInfo fi = QFileInfo(*file);
    if(fi.lastModified() != lastModified)
	return TRUE;
    return FALSE;
}

uint EBTCSBFile::numStates()
{
    return stateRecordMap.size();
}

QString EBTCSBFile::exportState(uint stateNr)
{
  QString returnValue;

  if(!checkFile()) return returnValue;
  if(!file->isOpen())
    {
      if (!file->open(QIODevice::ReadOnly))
	{
	  QMessageBox::critical(0, EBTCAPTION + "Can't open",
				QString("Cannot open file %1")
				.arg(fileName()));
	  return returnValue;
	}
    }
  if(fileChanged()) return returnValue;

  newConstDim = computeIDcard(stateNr);

  file->seek(stateRecordMap[stateNr].filePos);

  char *data = new char[sizeof(Envdim)];
  Envdim currentEnv;

  if(file->read(data, sizeof(Envdim)) > 0)
    currentEnv = *(Envdim*)(data);
  else
    {
      fileCorrupt();
      delete [] data;
      return returnValue;
    }
  delete [] data;

  file->seek(stateRecordMap[stateNr].filePos);

  data = new char[currentEnv.memory_used];
  double *basePoint, *popPoint;
  Popdim currentPop;
  if(file->read(data, currentEnv.memory_used) > 0)
    {
      basePoint =  (double*)(data) + currentEnv.data_offset;
      popPoint =  basePoint + currentEnv.columns;
      currentPop = *(Popdim*)popPoint;
    }
  else
    {
      fileCorrupt();
      delete [] data;
      return returnValue;
    }

  double output;
  for(int ii = 0; ii < currentEnv.columns; ii++)			// write environment
    {
      if(ii == 0)
	output = 0;
      else
	{
	  returnValue+="\t";
	  output = basePoint[ii];
	}
      returnValue += formatDouble(output);
    }
  returnValue += "\n\n";

  for(int popNr = 0; popNr < (int)stateRecordMap[stateNr].numPopulations; popNr++)	// write population states
    {
      basePoint = popPoint + currentPop.data_offset;
      for(int cohortNr = 0; cohortNr < currentPop.cohorts; cohortNr++)
	{
	  for(int colNr = 0; colNr < (currentPop.columns + newConstDim); colNr++)
	    {
	      if(colNr>0)
		returnValue += "\t";
	      if(colNr < currentPop.columns)
		output = *(double*)(basePoint + (colNr*currentPop.cohorts) + cohortNr);
	      else
		output = *(double*)(popIDcard[popNr] + ((colNr - currentPop.columns) * currentPop.cohorts) + cohortNr);
	      returnValue += formatDouble(output);
	    }
	  returnValue += "\n";
	}
      returnValue += "\n";
      popPoint += currentPop.data_offset + (currentPop.cohorts * currentPop.columns);
      currentPop = *(Popdim*)popPoint;
    }

  returnValue += "\n";
  file->close();
  delete [] data;

  return returnValue;
}

int EBTCSBFile::computeIDcard(uint stateNr)
{
  uint ii;

  if(defidcard->fileName().isEmpty() || !defidcard->libLoaded()) return 0;

  for(ii = 0; ii < stateRecordMap[stateNr].numPopulations; ii++)
    {
      if(defidcard->iStateDim() > stateRecordMap[stateNr].popColumns[ii] - 1)
	{
	  QMessageBox::warning(0, EBTCAPTION + "incompatibility",
			       QString("The state output file has %1 column(s),\nbut the defidcard, %2,\nexpects %3.\nThe defidcard will not be used.")
			       .arg(stateRecordMap[stateNr].popColumns[ii] - 1)
			       .arg(defidcard->fileName())
			       .arg(defidcard->iStateDim()));
	  defidcard->libUnload();
	  return 0;
	}
    }

  if(currentPopIDcard == (int)stateNr) return newConstDim;

  double *environment;
  double **population;
  int *cohortNumbers;
  char *dataToo = new char[sizeof(Envdim)];

  // first clear memory used by a previously calculated popIDcard:
  clearIDcard();

  file->seek(stateRecordMap[stateNr].filePos);

  Envdim currentEnv;

  if(file->read(dataToo, sizeof(Envdim)) > 0)
    currentEnv = *(Envdim*)(dataToo);
  else
    {
      fileCorrupt();
      delete [] dataToo;
      return 0;
    }
  delete [] dataToo;

  file->seek(stateRecordMap[stateNr].filePos);

  Popdim currentPop;
  double *basePoint;

  dataToo = new char[currentEnv.memory_used];
  if(file->read(dataToo, currentEnv.memory_used) > 0)
    {
      basePoint =  (double*)(dataToo) + currentEnv.data_offset;
      environment = (double *)basePoint;
      basePoint += currentEnv.columns;
    }
  else
    {
      fileCorrupt();
      delete [] dataToo;
      return 0;
    }

  cohortNumbers = new int[stateRecordMap[stateNr].numPopulations];
  population = new doublePtr[stateRecordMap[stateNr].numPopulations];

  for(ii = 0; ii < stateRecordMap[stateNr].numPopulations; ii++)
    {
      currentPop = *(Popdim*)basePoint;
      cohortNumbers[ii] = currentPop.cohorts;
      basePoint += currentPop.data_offset;
      population[ii] = (double*)basePoint;
      basePoint += currentPop.columns * currentPop.cohorts;
    }

  int nCols = defidcard->defIDcard(environment, population, NULL, csbPar, cohortNumbers);

  if(nCols > 0)
    {
      popIDcardSize = stateRecordMap[stateNr].numPopulations;

      // Existing IDcards have been cleared above
      popIDcard = new doublePtr[popIDcardSize];

      for (ii = 0; ii < popIDcardSize; ii++)
	{
	  popIDcard[ii] = new double[cohortNumbers[ii] * nCols];
	  // initialize these doubles, since the popIDcard-lib may not fill all of them!
	  for(int jj = 0; jj < (cohortNumbers[ii] * nCols); jj++)
	    popIDcard[ii][jj] = 0;
	}
      defidcard->defIDcard(environment, population, popIDcard, csbPar, cohortNumbers);
    }
  else popIDcardSize = 0;

  currentPopIDcard = stateNr;

  delete [] population;
  delete [] cohortNumbers;
  delete [] dataToo;

  return nCols;
}

void EBTCSBFile::computeOutput(uint stateNr)
{
    uint ii;

    if(defoutput->fileName().isEmpty() || !defoutput->libLoaded())
    {
#ifdef DEBUG
	  qDebug() <<  "computeOutput: File empty or not loaded" ;
#endif
	outputVarNr = 0;
	return;
    }

    for(ii = 0; ii < stateRecordMap[stateNr].numPopulations; ii++)
    {
	if(defoutput->iStateDim() > stateRecordMap[stateNr].popColumns[ii] - 1)
	{
	    QMessageBox::warning(0, EBTCAPTION + "incompatibility",
				 QString("The state output file has %1 column(s),\nbut the defoutput, %2,\nexpects %3.\nThe defoutput will not be used.")
				 .arg(stateRecordMap[stateNr].popColumns[ii] - 1)
				 .arg(defoutput->fileName())
				 .arg(defoutput->iStateDim()));
	    defoutput->libUnload();
	    clearCompOutput(); //!!
	    return;
	}
    }

    double *environment;
    double **population;
    int *cohortNumbers;
    char *dataToo;

    file->seek(stateRecordMap[stateNr].filePos);

    dataToo = new char[sizeof(Envdim)];
    Envdim currentEnv;

    if(file->read(dataToo, sizeof(Envdim)) > 0)
	currentEnv = *(Envdim*)(dataToo);
    else
      {
	fileCorrupt();
	clearCompOutput(); //!!
	delete [] dataToo;
	return;
      }
    delete [] dataToo;

    file->seek(stateRecordMap[stateNr].filePos);

    dataToo = new char[currentEnv.memory_used];
    Popdim currentPop;
    double *basePoint;

    if(file->read(dataToo, currentEnv.memory_used) > 0)
    {
	basePoint =  (double*)(dataToo) + currentEnv.data_offset;
	environment = (double *)basePoint;
	basePoint += currentEnv.columns;
    }
    else
      {
	fileCorrupt();
	clearCompOutput(); //!!
	delete [] dataToo;
	return;
      }

    cohortNumbers = new int[stateRecordMap[stateNr].numPopulations];
    population = new doublePtr[stateRecordMap[stateNr].numPopulations];

    for(ii = 0; ii < stateRecordMap[stateNr].numPopulations; ii++)
    {
	currentPop = *(Popdim*)basePoint;
	cohortNumbers[ii] = currentPop.cohorts;
	basePoint += currentPop.data_offset;
	population[ii] = (double*)basePoint;
	basePoint += currentPop.columns * currentPop.cohorts;
    }

    defoutput->defOut(environment, population, computedOutput[stateNr], csbPar, cohortNumbers);

    delete [] population;
    delete [] cohortNumbers;
    delete [] dataToo;
}

void EBTCSBFile::setDefID(EBTDefIDcard *defid)
{
    defidcard = defid;
    for(uint ii = 0; ii < numStates(); ii++)
	stateRecordMap[ii].read = FALSE;
    currentPopIDcard = -1;
    newConstDim = 0;
}

void EBTCSBFile::setDefOut(EBTDefOut *defout)
{
    defoutput = defout;
    outputVarNr = initComputeOutput();
    if(outputVarNr > 0)
    {
	for(uint stateNr = 0; stateNr < numStates(); stateNr++)
	{
	    // als in DefOutput weleens iets gedaan wordt met states van andere
	    // tijdstappen, dan moeten de volgende statements in aparte loops!!
	    stateRecordMap[stateNr].read = FALSE;
	    readState(stateNr);
	    // if computeOutput is unsuccessful the outputVarNr will be put to zero
	    // and this loop can be terminated
	    if(outputVarNr == 0)
		break;
	}
    }
}

void EBTCSBFile::clearIDcard()
{
  if(popIDcardSize > 0)
    {
      for (uint ii = 0; ii < popIDcardSize; ++ii)
	delete [] popIDcard[ii];
      delete [] popIDcard;
      popIDcardSize = 0;
    }
}

void EBTCSBFile::clearCompOutput()
{
  if(compOutputSize > 0)
    {
      for(int ii = 0; ii < compOutputSize; ++ii)
	delete [] computedOutput[ii];
    }
  if(compOutputSize > -1)
    {
      delete [] computedOutput;
      computedOutput = NULL;
    }
  compOutputSize = -1;
}

int EBTCSBFile::initComputeOutput()
{
    if(defoutput->fileName().isEmpty() || !defoutput->libLoaded())
	return 0;
    int retVal = defoutput->defOut(NULL, NULL, NULL, NULL, NULL);
    if(retVal == 0)
	return 0;
    clearCompOutput();
    computedOutput = new doublePtr[numStates()];
    compOutputSize = numStates();
    for(int ii = 0; ii < compOutputSize; ii++)
    {
	computedOutput[ii] = new double[retVal];
	for(int jj = 0; jj < retVal; jj++)
	    computedOutput[ii][jj] = 0;
    }
    return retVal;
}

void EBTCSBFile::clearColumnData()
{
  if (columnData)
    {
      delete [] columnData;
      columnData = NULL;
    }
}
