#include "ebtdefout.h"
#include "integrator/ebttooldefs.h"
#include <QDateTime>
#include <qmessagebox.h>

#ifdef DEBUG
#include <QtDebug>
#endif

EBTDefOut::EBTDefOut()
{
  libHandle     = NULL;
  defoutwrapper = NULL;

  setDescriptionText("DefineOutput");
#if defined(Q_OS_LINUX)
  soFile.setFileExtension("so");
#elif defined(Q_OS_MACOS)
  soFile.setFileExtension("dylib");
#elif defined(Q_OS_WIN)
  soFile.setFileExtension("dll");
#endif
  cFile.setTemplateFile("templates/DefineOutput.c");
  cFile.setFileExtension("c");
  cFile.setDescriptionText("DefineOutput source file");
  cFile.setSaveText(cFile.descriptionText() + " (*." + cFile.fileExtension() + ")");
  cFile.setOpenText(cFile.saveText());
  setOpenText(descriptionText() + " (*." + cFile.fileExtension() + " *." + soFile.fileExtension() + ")");
  setLibChosen(FALSE);
}

EBTDefOut::~EBTDefOut() { libUnload(); }

void EBTDefOut::setFileName(const QString &str)
{
  libUnload();

  if (str.isEmpty())
    {
      EBTFile::setFileName(QString());
      cFile.setFileName(QString());
      soFile.setFileName(QString());
    }
  else
    {
      QFileInfo fi(str);
      if (fi.suffix() == soFile.fileExtension())
        setLibChosen(TRUE);
      else
        setLibChosen(FALSE);
      QString basename = fi.absolutePath() + "/" + fi.completeBaseName();
      if (basename.endsWith("module")) basename = basename.left(basename.length() - 6);                            // ok?
      EBTFile::setFileName(basename);
      cFile.setFileName(basename + QString(".") + cFile.fileExtension());
      soFile.setFileName(basename + QString("module.") + soFile.fileExtension());
    }
}

void EBTDefOut::libUnload()
{
  if (libLoaded()) libHandle->unload();
  if (libHandle)
    {
      delete libHandle;
      libHandle = NULL;
    }
}

bool EBTDefOut::initHandles()
{
  if (fileName().isEmpty())
    {
      QMessageBox::warning(0, QString(EBTCAPTION + "No file selected"), QString("No defoutput file has been selected yet."));
      return FALSE;
    }

  libUnload();

  // lib is loaded using the basename of the model files (without extension: .so, .dll, dylib, ...)
  libHandle = new QLibrary(absFilePath() + "module");

  if (!libHandle->load())
    {
      QMessageBox::warning(0, QString(EBTCAPTION + "No such lib file"), QString("The lib file %1 could not be found.").arg(libHandle->fileName()));
      return FALSE;
    }

  defoutwrapper = (functionType1)libHandle->resolve("DefOutWrapper");
  if (!defoutwrapper) defoutwrapper = (functionType1)libHandle->resolve("_DefOutWrapper");

  if (!defoutwrapper)
    {
      messageLoadError("DefOut");
      libUnload();
      return FALSE;
    }

  istatedim = (functionType2)libHandle->resolve("iStateDim");
  if (!istatedim) istatedim = (functionType2)libHandle->resolve("_iStateDim");

  if (!istatedim)
    {
      messageLoadError("iStateDim");
      libUnload();
      return FALSE;
    }

  return TRUE;
}

void EBTDefOut::messageLoadError(QString handle)
{
  QMessageBox::warning(0, QString(EBTCAPTION + "DefOut load problem"), QString("Loading of %1 unsuccessful.").arg(handle));
}

int EBTDefOut::defOut(double *env, double **pop, double *output, double *pars, int *cohortNum)
{
  //int		ok = LIB_NOT_LOADED;
  int ok = -1;

  if (libLoaded()) ok = (*defoutwrapper)(env, pop, output, pars, cohortNum);
#ifdef DEBUG
  if (output) qDebug() << "defOut: " << output[0] << output[1];
#endif

  return ok;
}

bool EBTDefOut::checkLib()
{
  if (QFile(soFile.absFilePath()).exists())
    if (QFileInfo(soFile.absFilePath()).lastModified() > QFileInfo(cFile.absFilePath()).lastModified()) return TRUE;
  return FALSE;
}

void EBTDefOut::cleanDefOut()
{
  QString                     fn  = fileName();
  QMessageBox::StandardButton ans = QMessageBox::question(NULL, QString(EBTCAPTION + "Cleanup compiled DefOut files"),
                                                          QString("Are you sure you want to delete\nthe compiled file of %1?").arg(fn),
                                                          QMessageBox::Yes | QMessageBox::No, QMessageBox::No);
  if (ans == QMessageBox::No) return;

  libUnload();

  fn = QString(absFilePath() + "module.o");
  if (QFile(fn).exists()) QFile(fn).remove();
  fn = soFile.absFilePath();
  if (QFile(fn).exists()) QFile(fn).remove();

#ifdef Q_OS_WIN
  fn = QString(absFilePath() + "module.exp");
  if (QFile(fn).exists()) QFile(fn).remove();
  fn = QString(absFilePath() + "module.lib");
  if (QFile(fn).exists()) QFile(fn).remove();
  fn = QString(absFilePath() + "module.tds");
  if (QFile(fn).exists()) QFile(fn).remove();
  fn = QString(absFilePath() + "module.obj");
  if (QFile(fn).exists()) QFile(fn).remove();
  fn = QString(".obj.obj");
  if (QFile(fn).exists()) QFile(fn).remove();
#endif

  EBTFile::setFileName(QString());
  cFile.setFileName(QString());
  soFile.setFileName(QString());
}
