#include "ebtgraph.h"

#include <qvariant.h>
#include <qimage.h>
#include <qpixmap.h>

#include "qstatusbar.h"
#include "qapplication.h"
#include "qwt_plot.h"
#include "qmessagebox.h"
#include "QFileDialog"
#include "QDir"
#include "qfile.h"
#include "qprinter.h"
#include "ebtmoviedialog.h"
#include "ebtplotcurve.h"
#include "qwt_plot_canvas.h"
#include "qwt_dyngrid_layout.h"
#include "qwt_plot_layout.h"
#include <qwt_plot_renderer.h>
#include "qwt_scale_engine.h"
#include "qwt_legend.h"
#include "qwt_plot_legenditem.h"
#include "qwt_scale_map.h"
#include "float.h"
#include "math.h"

#include <QPixmap>
#include <QCloseEvent>
#include <QVBoxLayout>
#include <QtGlobal>
#include <QPrintDialog>

#ifdef DEBUG
#include <QtDebug>
#endif

class LegendItem : public QwtPlotLegendItem
{
  QPoint m_pos;

public:
  LegendItem()
  {
    setRenderHint(QwtPlotItem::RenderAntialiased);
    setAlignmentInCanvas(Qt::AlignLeft | Qt::AlignBottom);
    setPosition(QPoint(0, 0));
  }
  virtual QRect geometry(const QRectF &canvasRect) const
  {
    QRect rect = QwtPlotLegendItem::geometry(canvasRect);
    rect.moveTopLeft(m_pos + canvasRect.topLeft().toPoint());
    return rect;
  }
  void moveTo(double xpos, double ypos, const QRectF &canvasRect)
  {
    int lx, ly;

    lx = (int)(xpos*canvasRect.width());
    ly = (int)((1.0 - ypos)*canvasRect.height());

    QRect legendSize                                         = geometry(canvasRect);
    if (lx > (canvasRect.width() - legendSize.width())) lx   = canvasRect.width() - legendSize.width();
    if (ly > (canvasRect.height() - legendSize.height())) ly = canvasRect.height() - legendSize.height();
    setPosition(QPoint(lx, ly));
  }
  void setPosition(const QPoint &pos)
  {
    m_pos = pos;
    itemChanged();
  }
};

/*
 *  Constructs a EBTGraphForm as a child of 'parent', with the
 *  name 'name' and widget flags set to 'f'.
 *
 */

EBTGraphForm::EBTGraphForm(QWidget *parent, Qt::WindowFlags fl) : QMainWindow(parent, fl)
{
  hide();
  setupUi(this);

  (void)statusBar();
  init();
}


/*
 *  Destroys the object and frees any allocated resources
 */
EBTGraphForm::~EBTGraphForm()
{
  thisPlot->detachItems(QwtPlotItem::Rtti_PlotMarker, true);                        // Detach all markers and delete them automatically
  thisPlot->detachItems(QwtPlotItem::Rtti_PlotCurve, true);                         // Detach all curves and delete them automatically

  // no need to delete child widgets, Qt does it all for us
  if (floatingLegend)                                                                // remove existing painter
    {
      floatingLegend->detach();
      delete floatingLegend;
    }
  if (plotDirectPainter)                                                             // remove existing painter
    delete plotDirectPainter;
  if (plotGrid)                                                                      // remove existing grid
    delete plotGrid;
  if (zoomer)                                                                        // remove existing zoomer
    delete zoomer;
}


/*
 *  Sets the strings of the subwidgets using the current
 *  language.
 */
void EBTGraphForm::languageChange() { retranslateUi(this); }


void EBTGraphForm::init()
{
  zoomOn          = FALSE;
  hiddenFlag      = FALSE;
  dirtyFlag       = TRUE;
  recordingMovie  = FALSE;
  isOutputGraph   = TRUE;                                                             // has to be initialized
  currentFile     = NULL;
  currentGIFDelay = defaultGIFDelay;

  thisPlot       = this->ebtPlot;
  floatingLegend = new LegendItem();
  floatingLegend->attach(thisPlot);

  // Using a backing store might improve the performance significantly, when working with widget
  // overlays ( like rubberbands ). Disabling the cache might improve the performance for
  // incremental paints (using QwtPlotDirectPainter ).

  QwtPlotCanvas *plotCanvas = new QwtPlotCanvas();
  plotCanvas->setPaintAttribute(QwtPlotCanvas::BackingStore, true);
  plotCanvas->setFrameStyle(QFrame::NoFrame);

#if defined(Q_OS_LINUX)
  // Even if not recommended by TrollTech, Qt::WA_PaintOutsidePaintEvent
  // works on X11. This has a nice effect on the performance.

  plotCanvas->setAttribute(Qt::WA_PaintOutsidePaintEvent, true);

  // Disabling the backing store of Qt improves the performance
  // for the direct painter even more, but the canvas becomes
  // a native window of the window system, receiving paint events
  // for resize and expose operations. Those might be expensive
  // when there are many points and the backing store of
  // the canvas is disabled. So in this application
  // we better don't both backing stores.

  if (plotCanvas->testPaintAttribute(QwtPlotCanvas::BackingStore))
    {
      plotCanvas->setAttribute(Qt::WA_PaintOnScreen, true);
      plotCanvas->setAttribute(Qt::WA_NoSystemBackground, true);
    }

// Mac OS X always double buffers the screen so the Qt::WA_PaintOnScreen attribute has no effect.
// Also it is impossible to paint outside of a paint event so Qt::WA_PaintOutsidePaintEvent has
// no effect either.
#endif

  thisPlot->setCanvas(plotCanvas);

  thisPlot->setContentsMargins(10, 10, 10, 5);

  // Create the direct painter
  plotDirectPainter = new QwtPlotDirectPainter();
  plotDirectPainter->setAttribute(QwtPlotDirectPainter::CopyBackingStore, true);

  // Create and attach no legend
  thisPlot->insertLegend(NULL);

  // Create and attach the grid grid
  plotGrid = new QwtPlotGrid;
  plotGrid->attach(thisPlot);

  // Create the zoomer
  zoomer = new QwtPlotZoomer(QwtPlot::xBottom, QwtPlot::yLeft, plotCanvas);
  zoomer->setRubberBandPen(QPen(Qt::red));

  thisPlot->setAutoDelete(true);
  thisPlot->setAutoReplot(false);

  // Here I need the main window as a parent to set up these connections
  connect(this, SIGNAL(newGrFile(EBTGraphForm *, bool)), parentWidget(), SLOT(updateGraphSlot(EBTGraphForm *, bool)));
  connect(this, SIGNAL(closing(EBTGraphForm *)), parentWidget(), SLOT(graphCloses(EBTGraphForm *)));
  connect(this, SIGNAL(minimizing(EBTGraphForm *)), parentWidget(), SLOT(graphHides(EBTGraphForm *)), Qt::QueuedConnection);
  connect(this, SIGNAL(openNewGraph(bool)), parentWidget(), SLOT(graphNewGraphSlot(bool)));

  zoomEnable(FALSE);
  cleanUp();
  showText(QString());

  // From here on I do not need the parent-child relationship any longer
  setParent(NULL);
  graphMenuItem = (QAction *)NULL;

  show();
  showNormal();
  move(100, 100);                                                                    // Move it to a visible inital position
}


void EBTGraphForm::hideEvent(QHideEvent *e)
{
#ifdef DEBUG
  qDebug("EBTGraphForm::hideEvent()");
#endif
  currentSettings.setXPos((this->pos()).x());
  currentSettings.setYPos((this->pos()).y());

  if (e->spontaneous())
    {
#ifdef DEBUG
      qDebug("EBTGraphForm::hideEvent(): emitting signal");
#endif
      emit minimizing(this);
    }
}

void EBTGraphForm::showEvent(QShowEvent *e)
{
  if (e->spontaneous())                                                              // Empty statement to silence an unused argument warning
    {
      ;
    }

#ifdef DEBUG
  qDebug("EBTGraphForm::showEvent()");
#endif

  show();
  showNormal();
  raise();
  setHideFlag(FALSE);
  if (graphMenuItem) graphMenuItem->setIcon(QIcon(":/images/check_checked.png"));
  this->move(currentSettings.xPos(), currentSettings.yPos());
}

void EBTGraphForm::fileNew()
{
  if (okToClear())
    {
      currentFile->createNew();
      cleanUp();
    }
}


void EBTGraphForm::fileOpen()
{
  // open an existing settings file
  if (!okToClear()) return;

  statusBar()->showMessage("Opening file...");
  if (!currentFile->open())
    statusBar()->showMessage("File open abandoned", 2000);
  else
    setFile(*currentFile);
}


void EBTGraphForm::fileSaveAs()
{
  currentSettings.setXSize((this->size()).width());
  currentSettings.setYSize((this->size()).height());
  currentSettings.setXPos((this->pos()).x());
  currentSettings.setYPos((this->pos()).y());

  currentFile->settings = currentSettings;
  if (currentFile->saveAs())
    {
      statusBar()->showMessage(QString("Saved graph settings to '%1'").arg(currentFile->fileName()), 3000);
      currentFile->setChanged(FALSE);
    }
  else
    QMessageBox::critical(this, "Warning", QString("Failed to save '%1'").arg(currentFile->fileName()));
  setWindowTitle(currentFile->captionText());
}


void EBTGraphForm::fileSave()
{
  currentSettings.setXSize((this->size()).width());
  currentSettings.setYSize((this->size()).height());
  currentSettings.setXPos((this->pos()).x());
  currentSettings.setYPos((this->pos()).y());

  currentFile->settings = currentSettings;
  if (currentFile->save())
    {
      statusBar()->showMessage(QString("Saved graph settings to '%1'").arg(currentFile->fileName()), 3000);
      currentFile->setChanged(FALSE);
    }
  else
    QMessageBox::critical(this, "Warning", QString("Failed to save '%1'").arg(currentFile->fileName()));

  setWindowTitle(currentFile->captionText());
}


void EBTGraphForm::filePrint()
{
  QPrinter printer;
  QString  docName = (thisPlot->title()).text();

  if (!docName.isEmpty())
    // docName.replace (QRegExp (QString::fromLatin1 ("\n")), tr (" -- "));
    printer.setDocName(docName);

  if (isOutputGraph)
    printer.setCreator(EBTCAPTION + OUTPUTDESC);
  else
    printer.setCreator(EBTCAPTION + STATEDESC);

  printer.setPageOrientation(QPageLayout::Landscape);

  QPrintDialog dialog(&printer);
  if (dialog.exec())
    {
      QwtPlotRenderer renderer;

      if (printer.colorMode() == QPrinter::GrayScale)
        {
          renderer.setDiscardFlag(QwtPlotRenderer::DiscardBackground);
          renderer.setDiscardFlag(QwtPlotRenderer::DiscardCanvasBackground);
          renderer.setDiscardFlag(QwtPlotRenderer::DiscardCanvasFrame);
          renderer.setLayoutFlag(QwtPlotRenderer::FrameWithScales);
        }

      renderer.renderTo(thisPlot, printer);
    }
}

void EBTGraphForm::closeEvent(QCloseEvent *) { fileExit(); }

void EBTGraphForm::fileExit()
{
  if (okToClear())
    {
      hide();
      emit closing(this);
    }
}


void EBTGraphForm::helpIndex() {}


void EBTGraphForm::helpContents() {}


void EBTGraphForm::helpAbout() {}


void EBTGraphForm::layoutEdit()
{
  GraphSettingsForm *settingsForm = new GraphSettingsForm(this, TRUE);
  settingsForm->setCurrentSettings(currentSettings, isOutputGraph);
  settingsForm->exec();
  delete settingsForm;
}


void EBTGraphForm::cleanUp()
{
  if (isOutputGraph)
    setWindowTitle(EBTCAPTION + OUTPUTDESC);
  else
    setWindowTitle(EBTCAPTION + STATEDESC);
}


bool EBTGraphForm::okToClear()
{
  QMessageBox::StandardButton ans;

  if (recordingMovie)
    {
      ans = QMessageBox::question(this, EBTCAPTION + "Unfinished movie",
                                  QString("The movie %1\nwas being recorded.\nDo you want to save it?").arg(movieName), QMessageBox::Save | QMessageBox::Cancel | QMessageBox::Discard, QMessageBox::Save);
      if (ans == QMessageBox::Save)
        plotFinishMovie();
      else if (ans == QMessageBox::Cancel)
        return FALSE;
      else
        plotCleanupMovie(TRUE);
    }

  if (((this->size()).width() != currentSettings.xSize()) || ((this->size()).height() != currentSettings.ySize()))
    {
      currentFile->setChanged(TRUE);
      currentSettings.setXSize((this->size()).width());
      currentSettings.setYSize((this->size()).height());
    }

  if (((this->pos()).x() != currentSettings.xPos()) || ((this->pos()).y() != currentSettings.yPos()))
    {
      currentSettings.setXPos((this->pos()).x());
      currentSettings.setYPos((this->pos()).y());
    }
  if (currentFile->isChanged())
    {
      QString msg = "The graph settings";
      if (!currentFile->fileName().isEmpty()) msg += QString(" in '%1'\n").arg(currentFile->fileName());
      msg += QString(" have changed.\n Do you want to save the changes?");
      ans = QMessageBox::question(this, EBTCAPTION + "Unsaved changes", msg, QMessageBox::Save | QMessageBox::Cancel | QMessageBox::Discard, QMessageBox::Save);
      if (ans == QMessageBox::Save)
        fileSave();
      else if (ans == QMessageBox::Cancel)
        return FALSE;
      else
        currentFile->setChanged(FALSE);
    }
  return TRUE;
}


void EBTGraphForm::incrementalOutputPlot()
{
  uint jj;

  thisPlot->setAutoReplot(false);
  for (uint ii = 0; ii < currentSettings.dataSetMapSize(); ii++)
    {
      if (currentSettings.allSetsSameAs() < 0)
        jj = ii;
      else
        jj = uint(currentSettings.allSetsSameAs());

      if (currentSettings.dataSetCurveStyle(jj) == QwtPlotCurve::UserCurve) continue;

      if (currentSettings.dataSetCurveID(ii) && currentSettings.dataSetSize(ii))
        {
          if (!currentSettings.dataSetExists(ii)) return;

          // During the integration we avoid SetData to minimize copy actions of the data
#ifdef DEBUG
          qDebug() << "incrementalOutputPlot: Plotting additional points set " << ii;
#endif

          currentSettings.dataSetCurveID(ii)->setRawSamples(currentSettings.dataSetXDataPnt(ii), currentSettings.dataSetYDataPnt(ii),
                                                            currentSettings.dataSetSize(ii));

          const QwtScaleMap xMap = thisPlot->canvasMap(currentSettings.dataSetCurveID(ii)->xAxis());
          const QwtScaleMap yMap = thisPlot->canvasMap(currentSettings.dataSetCurveID(ii)->yAxis());
          double            xMin = currentSettings.dataSetXDataPnt(ii)[currentSettings.dataSetOldSize(ii) - 1];
          double            xMax = xMin;
          double            yMin = currentSettings.dataSetYDataPnt(ii)[currentSettings.dataSetOldSize(ii) - 1];
          double            yMax = yMin;
          for (int jj = currentSettings.dataSetOldSize(ii); jj < currentSettings.dataSetSize(ii); jj++)
            {
              xMin = qMin(xMin, currentSettings.dataSetXDataPnt(ii)[jj]);
              xMax = qMax(xMax, currentSettings.dataSetXDataPnt(ii)[jj]);
              yMin = qMin(yMin, currentSettings.dataSetYDataPnt(ii)[jj]);
              yMax = qMax(yMax, currentSettings.dataSetYDataPnt(ii)[jj]);
            }
          QRectF clipRange(xMin, yMin, (xMax - xMin), (yMax - yMin));
          QRect  clipRect = QwtScaleMap::transform(xMap, yMap, clipRange).toRect();

          if (currentSettings.dataSetCurveID(ii)->symbol()) 
            {
              const QSize symbolSize = currentSettings.dataSetCurveID(ii)->symbol()->size();
              clipRect.adjust(-symbolSize.width()/2 - 1, -symbolSize.height()/2 - 1, symbolSize.width() + 2, symbolSize.height() + 2);
            }

          plotDirectPainter->setClipRegion(clipRect);

          plotDirectPainter->drawSeries(currentSettings.dataSetCurveID(ii), currentSettings.dataSetOldSize(ii) - 1,
                                        currentSettings.dataSetSize(ii) - 1);

          currentSettings.setDataSetOldSize(ii, currentSettings.dataSetSize(ii));
        }
    }
  saveMovieFrame();
}

void EBTGraphForm::setFile(EBTGrFile &ef)
{
  cleanUp();

  currentFile = &ef;
  statusBar()->showMessage(QString("Loading '%1'...").arg(currentFile->fileName()));
#ifdef DEBUG
  qDebug() << "Reading graph file" << currentFile->fileName();
#endif
  currentFile->load();

  currentSettings = currentFile->settings;
  isOutputGraph   = currentFile->isOutputGraph;
  setWindowTitle(currentFile->captionText());
  currentFile->setChanged(FALSE);

  if (currentSettings.xSize() > 0 && currentSettings.ySize() > 0) this->resize(QSize(currentSettings.xSize(), currentSettings.ySize()));
  if (currentSettings.xPos() > 0 && currentSettings.yPos() > 0) this->move(QPoint(currentSettings.xPos(), currentSettings.yPos()));

  emit newGrFile(this, isOutputGraph);

  statusBar()->showMessage("Ready");
}


void EBTGraphForm::redrawData()
{
  uint ii, jj;

  if (!currentSettings.plotTitle().isEmpty())
    {
      if (currentText.isEmpty())
        thisPlot->setTitle(currentSettings.plotTitle());
      else
        thisPlot->setTitle(currentSettings.plotTitle() + " - " + currentText);
    }
  else
    {
      if (!currentText.isEmpty())
        thisPlot->setTitle(currentText);
      else
        thisPlot->setTitle("");
    }

  // Load the data to the different sets without plotting them
  for (ii = 0; ii < currentSettings.dataSetMapSize(); ii++)
    {
      // Skip if curve does not exist for some unlikely reason
      if (!currentSettings.dataSetCurveID(ii)) continue;

      if (currentSettings.allSetsSameAs() < 0)
        jj = ii;
      else
        jj = uint(currentSettings.allSetsSameAs());

      if (currentSettings.dataSetSize(ii) > 0)
        {
          /*
          double xMin, xMax, yMin, yMax;

          if (currentSettings.dataSetCurveStyle(jj) == QwtPlotCurve::UserCurve)
            {
              if (!makeHistogram(ii)) continue;
              xMin = axisMinX(ii);
              xMax = axisMaxX(ii);
              yMin = axisMinY(ii);
              yMax = histYMax;
            }
          else
            {
              xMin = currentSettings.dataSetXMinVal(ii);
              xMax = currentSettings.dataSetXMaxVal(ii);
              yMin = currentSettings.dataSetYMinVal(ii);
              yMax = currentSettings.dataSetYMaxVal(ii);
            }
          */
          // Skip if curve does not exist for some unlikely reason
          if (!currentSettings.dataSetCurveID(ii)) continue;

          // Bounding logarithmic values to positive
          if (currentSettings.axisLogarithmic(currentSettings.dataSetXAxis(ii)))
            {
#ifdef DEBUG
              qDebug() << "redrawData: Bounding X-values";
#endif
              for (uint kk = 0; kk < (uint)currentSettings.dataSetSize(ii); kk++)
                currentSettings.dataSetXDataPnt(ii)[kk] = qMax(currentSettings.dataSetXDataPnt(ii)[kk], axisMinX(ii));
            }
          if (currentSettings.axisLogarithmic(currentSettings.dataSetYAxis(ii)))
            {
#ifdef DEBUG
              qDebug() << "redrawData: Bounding Y-values";
#endif
              for (uint kk = 0; kk < (uint)currentSettings.dataSetSize(ii); kk++)
                currentSettings.dataSetYDataPnt(ii)[kk] = qMax(currentSettings.dataSetYDataPnt(ii)[kk], axisMinY(ii));
            }

#ifdef DEBUG
          qDebug() << "redrawData: Size:" << currentSettings.dataSetSize(ii) << currentSettings.dataSetOldSize(ii);
          qDebug() << "redrawData: Size:" << currentSettings.dataSetXDataPnt(ii)[0] << currentSettings.dataSetYDataPnt(ii)[0];
#endif

          currentSettings.dataSetCurveID(ii)->setRawSamples(currentSettings.dataSetXDataPnt(ii), currentSettings.dataSetYDataPnt(ii),
                                                            currentSettings.dataSetSize(ii));

          currentSettings.dataSetCurveID(ii)->setBaseline(axisMinY(ii));
          currentSettings.dataSetCurveID(ii)->setPen(currentSettings.dataSetLinePen(jj));

          if (currentSettings.dataSetCurveStyle(jj) == QwtPlotCurve::UserCurve)
            {
              currentSettings.dataSetCurveID(ii)->setStyle(QwtPlotCurve::Lines);
              currentSettings.dataSetCurveID(ii)->setSymbol(new QwtSymbol(QwtSymbol::NoSymbol));
              currentSettings.dataSetCurveID(ii)->setBrush(currentSettings.dataSetCurveSymbol(jj)->brush());
            }
          else
            {
              currentSettings.dataSetCurveID(ii)->setSymbol(
                  new QwtSymbol(currentSettings.dataSetCurveSymbol(jj)->style(), currentSettings.dataSetCurveSymbol(jj)->brush(),
                                currentSettings.dataSetCurveSymbol(jj)->pen(), currentSettings.dataSetCurveSymbol(jj)->size()));
              currentSettings.dataSetCurveID(ii)->setStyle(currentSettings.dataSetCurveStyle(jj));
            }

          if (currentSettings.showLegend())
            {
              currentSettings.dataSetCurveID(ii)->setItemAttribute(QwtPlotItem::Legend, true);
              currentSettings.dataSetCurveID(ii)->setLegendAttribute(QwtPlotCurve::LegendShowLine, true);
              currentSettings.dataSetCurveID(ii)->setLegendAttribute(QwtPlotCurve::LegendShowSymbol, true);
              currentSettings.dataSetCurveID(ii)->setLegendAttribute(QwtPlotCurve::LegendShowBrush, true);
            }
          else
            currentSettings.dataSetCurveID(ii)->setItemAttribute(QwtPlotItem::Legend, false);

          currentSettings.setDataSetOldSize(ii, currentSettings.dataSetSize(ii));
        }
    }
  plotDirectPainter->reset();
  plotDirectPainter->setAttribute(QwtPlotDirectPainter::CopyBackingStore, true);
  thisPlot->replot();

  saveMovieFrame();
}                                                                                    // redrawData()

void EBTGraphForm::applySettings()
{
  uint ii, jj;
  QMap<QwtPlot::Axis, double> axisMin, axisMax;
  double axisDiv;
  QMap<QwtPlot::Axis, uint> setsAtAxis;

  thisPlot->setAutoReplot(false);

  // Refresh the grid settings
  plotGrid->enableX(currentSettings.showGrid());
  plotGrid->enableY(currentSettings.showGrid());
  plotGrid->enableXMin(currentSettings.showFineGrid());
  plotGrid->enableYMin(currentSettings.showFineGrid());

  thisPlot->detachItems(QwtPlotItem::Rtti_PlotMarker, true);                        // Detach all markers and delete them automatically
  thisPlot->detachItems(QwtPlotItem::Rtti_PlotCurve, false);                        // Detach all curves, but do NOT delete them (being re-used)

  // Initialize the axes titles and the minimum and maximum values for autoscaling
  for (QwtPlot::Axis a = QwtPlot::yLeft; a < QwtPlot::axisCnt; a = QwtPlot::Axis(a + 1))
    {
      axisMax[a]    = -(DBL_MAX);
      axisMin[a]    = DBL_MAX;
      setsAtAxis[a] = 0;

      QwtText curTitle(currentSettings.axisTitle(a));
      curTitle.setFont(currentSettings.axisTitleFont(a));
      thisPlot->setAxisTitle(a, curTitle);
      thisPlot->setAxisFont(a, currentSettings.legendFont());
    }

  // by default, the secondary axes aren't displayed:
  thisPlot->enableAxis(QwtPlot::xTop, FALSE);
  thisPlot->enableAxis(QwtPlot::yRight, FALSE);

  bool smallestLog = (currentSettings.axisLogarithmic(QwtPlot::xBottom) && (currentSettings.axisMinVal(QwtPlot::xBottom) <= MIN_LOG_VAL));
  bool largestLog  = (currentSettings.axisLogarithmic(QwtPlot::xBottom) && (currentSettings.axisMinVal(QwtPlot::xBottom) >= MAX_LOG_VAL));
  axisToLeftAction->setEnabled(!currentSettings.axisAutoscale(QwtPlot::xBottom) && !smallestLog);
  axisToRightAction->setEnabled(!currentSettings.axisAutoscale(QwtPlot::xBottom) && !largestLog);

  QwtText newtitle;

  // Set the font of the plot title
  newtitle.setFont(currentSettings.plotTitleFont());

  if (!currentSettings.plotTitle().isEmpty())
    {
      if (currentText.isEmpty())
        newtitle.setText(currentSettings.plotTitle());
      else
        newtitle.setText(currentSettings.plotTitle() + " - " + currentText);
    }
  else
    {
      if (!currentText.isEmpty())
        newtitle.setText(currentText);
      else
        newtitle.setText("");
    }

  thisPlot->setTitle(newtitle);

  if (currentSettings.showLegend())
    {
      if (currentSettings.legendPosition() > QwtPlot::TopLegend)
        {
          if (thisPlot->legend())                                                   // remove existing attached legend
            thisPlot->insertLegend(NULL);
          if (floatingLegend)                                                       // remove existing floating legend
            {
              floatingLegend->detach();
              delete floatingLegend;
            }

          // Create the new legend
          floatingLegend = new LegendItem();
          floatingLegend->attach(thisPlot);

          // Set font of the legend
          floatingLegend->setFont(currentSettings.legendFont());

          // Paint the frame of the legend, if required
          if (currentSettings.legendFramed())
            {
              QPen legendPen = QPen(QColor(Qt::black));
              legendPen.setWidth(currentSettings.legendFrameWidth());
              floatingLegend->setBorderPen(legendPen);
              floatingLegend->setBorderRadius(2.0);
            }
          else
            floatingLegend->setBorderPen(QPen(Qt::NoPen));

          floatingLegend->setMaxColumns(1);
          floatingLegend->show();
        }
      else
        {
          if (thisPlot->legend())                                                   // remove existing attached legend
            thisPlot->insertLegend(NULL);
          if (floatingLegend)                                                       // remove existing floating legend
            {
              floatingLegend->detach();
              delete floatingLegend;
              floatingLegend = NULL;
            }

          // Create the new legend
          QwtLegend *plotLegend = new QwtLegend();

          // Set font of the legend
          plotLegend->setFont(currentSettings.legendFont());

          // Paint the frame of the legend, if required
          if (currentSettings.legendFramed())
            {
              plotLegend->setFrameShadow(QFrame::Plain);
              plotLegend->setFrameShape(QFrame::Box);
              plotLegend->setLineWidth(currentSettings.legendFrameWidth());
            }
          else
            plotLegend->setFrameShape(QFrame::NoFrame);

          if ((currentSettings.legendPosition() == QwtPlot::BottomLegend) || (currentSettings.legendPosition() == QwtPlot::RightLegend) ||
              (currentSettings.legendPosition() == QwtPlot::TopLegend))
            thisPlot->insertLegend(plotLegend, QwtPlot::LegendPosition(currentSettings.legendPosition()));
          else
            thisPlot->insertLegend(plotLegend, QwtPlot::BottomLegend);
        }
    }

  // Now attach all curves to this plot
  for (ii = 0; ii < currentSettings.dataSetMapSize(); ii++)
    {
      if (currentSettings.allSetsSameAs() < 0)
        jj = ii;
      else
        jj = uint(currentSettings.allSetsSameAs());

      if (currentSettings.dataSetXAxis(ii) == QwtPlot::xTop) thisPlot->enableAxis(QwtPlot::xTop);
      if (currentSettings.dataSetYAxis(ii) == QwtPlot::yRight) thisPlot->enableAxis(QwtPlot::yRight);

      if (currentSettings.dataSetSize(ii) > 0)
        {
          double xMin, xMax, yMin, yMax;

          if (currentSettings.dataSetCurveStyle(jj) == QwtPlotCurve::UserCurve)
            {
              if (!makeHistogram(ii)) continue;
              xMin = axisMinX(ii);
              xMax = axisMaxX(ii);
              yMin = axisMinY(ii);
              yMax = histYMax;
            }
          else
            {
              xMin = currentSettings.dataSetXMinVal(ii);
              xMax = currentSettings.dataSetXMaxVal(ii);
              yMin = currentSettings.dataSetYMinVal(ii);
              yMax = currentSettings.dataSetYMaxVal(ii);
            }

          if (xMin < axisMin[currentSettings.dataSetXAxis(ii)]) axisMin[currentSettings.dataSetXAxis(ii)] = xMin;
          if (xMax > axisMax[currentSettings.dataSetXAxis(ii)]) axisMax[currentSettings.dataSetXAxis(ii)] = xMax;
          if (yMin < axisMin[currentSettings.dataSetYAxis(ii)]) axisMin[currentSettings.dataSetYAxis(ii)] = yMin;
          if (yMax > axisMax[currentSettings.dataSetYAxis(ii)]) axisMax[currentSettings.dataSetYAxis(ii)] = yMax;

          setsAtAxis[currentSettings.dataSetXAxis(ii)]++;
          setsAtAxis[currentSettings.dataSetYAxis(ii)]++;

          // Skip if curve does not exist for some unlikely reason
          if (!currentSettings.dataSetCurveID(ii)) continue;

          currentSettings.dataSetCurveID(ii)->setTitle(currentSettings.dataSetLegendLabel(ii));

          // Bounding logarithmic values to positive
          if (currentSettings.axisLogarithmic(currentSettings.dataSetXAxis(ii)))
            {
#ifdef DEBUG
              qDebug() << "applySettings: Bounding X-values";
#endif
              for (uint kk = 0; kk < (uint)currentSettings.dataSetSize(ii); kk++)
                currentSettings.dataSetXDataPnt(ii)[kk] = qMax(currentSettings.dataSetXDataPnt(ii)[kk], axisMinX(ii));
            }
          if (currentSettings.axisLogarithmic(currentSettings.dataSetYAxis(ii)))
            {
#ifdef DEBUG
              qDebug() << "applySettings: Bounding Y-values";
#endif
              for (uint kk = 0; kk < (uint)currentSettings.dataSetSize(ii); kk++)
                currentSettings.dataSetYDataPnt(ii)[kk] = qMax(currentSettings.dataSetYDataPnt(ii)[kk], axisMinY(ii));
            }

          currentSettings.dataSetCurveID(ii)->setRawSamples(currentSettings.dataSetXDataPnt(ii), currentSettings.dataSetYDataPnt(ii),
                                                            currentSettings.dataSetSize(ii));

          currentSettings.dataSetCurveID(ii)->setAxes(currentSettings.dataSetXAxis(ii), currentSettings.dataSetYAxis(ii));

          currentSettings.dataSetCurveID(ii)->setPen(currentSettings.dataSetLinePen(jj));
          currentSettings.dataSetCurveID(ii)->setBaseline(axisMinY(jj));

          if (currentSettings.dataSetCurveStyle(jj) == QwtPlotCurve::UserCurve)
            {
              currentSettings.dataSetCurveID(ii)->setStyle(QwtPlotCurve::Lines);
              currentSettings.dataSetCurveID(ii)->setSymbol(new QwtSymbol(QwtSymbol::NoSymbol));
              currentSettings.dataSetCurveID(ii)->setBrush(currentSettings.dataSetCurveSymbol(jj)->brush());
            }
          else
            {
              currentSettings.dataSetCurveID(ii)->setSymbol(
                  new QwtSymbol(currentSettings.dataSetCurveSymbol(jj)->style(), currentSettings.dataSetCurveSymbol(jj)->brush(),
                                currentSettings.dataSetCurveSymbol(jj)->pen(), currentSettings.dataSetCurveSymbol(jj)->size()));
              currentSettings.dataSetCurveID(ii)->setStyle(currentSettings.dataSetCurveStyle(jj));
            }

          currentSettings.dataSetCurveID(ii)->setItemAttribute(QwtPlotItem::AutoScale, true);

          if (currentSettings.showLegend())
            {
              currentSettings.dataSetCurveID(ii)->setItemAttribute(QwtPlotItem::Legend, true);
              currentSettings.dataSetCurveID(ii)->setLegendAttribute(QwtPlotCurve::LegendShowLine, true);
              currentSettings.dataSetCurveID(ii)->setLegendAttribute(QwtPlotCurve::LegendShowSymbol, true);
              currentSettings.dataSetCurveID(ii)->setLegendAttribute(QwtPlotCurve::LegendShowBrush, true);
            }
          else
            currentSettings.dataSetCurveID(ii)->setItemAttribute(QwtPlotItem::Legend, false);

          // Plot curves will be attached to the plot and automatically deleted when closing
          // the plot (see the destructor above).
          currentSettings.dataSetCurveID(ii)->attach(thisPlot);
          currentSettings.setDataSetOldSize(ii, currentSettings.dataSetSize(ii));
        }
    }

  // Now set the axis scaling
  autoScale = FALSE;
  for (QwtPlot::Axis a = QwtPlot::yLeft; a < QwtPlot::axisCnt; a = QwtPlot::Axis(a + 1))
    {
      if (currentSettings.axisExists(a))
        {
          if (currentSettings.axisLogarithmic(a))
            {
              if (currentSettings.axisMinVal(a) < MIN_LOG_VAL) currentSettings.setAxisMinVal(a, MIN_LOG_VAL);
              if (currentSettings.axisMaxVal(a) <= MIN_LOG_VAL) currentSettings.setAxisMaxVal(a, 10*MIN_LOG_VAL);

              thisPlot->setAxisScaleEngine(a, new QwtLogScaleEngine);

              thisPlot->setAxisMaxMinor(a, MINOR_LOG_TICKS);
            }
          else
            {
              thisPlot->setAxisScaleEngine(a, new QwtLinearScaleEngine);
            }

          if (setsAtAxis[a] && currentSettings.axisAutoscale(a))
            {
              autoScale = TRUE;
              thisPlot->axisScaleEngine(a)->autoScale(MAX_AUTO_TICKS, axisMin[a], axisMax[a], axisDiv);
              thisPlot->setAxisScale(a, axisMin[a], axisMax[a], axisDiv);
            }
          else
            thisPlot->setAxisScale(a, currentSettings.axisMinVal(a), currentSettings.axisMaxVal(a));

#ifdef DEBUG
          qDebug() << "applySettings: Setting title: " << a << " to " << currentSettings.axisTitle(a);
#endif
        }
    }

  // Position a floating legend
  if ((currentSettings.showLegend()) && (currentSettings.legendPosition() > QwtPlot::TopLegend))
    {
      floatingLegend->show();
      floatingLegend->moveTo(currentSettings.legendXPosition(), currentSettings.legendYPosition(), thisPlot->canvas()->geometry());
    }

  plotDirectPainter->reset();
  thisPlot->replot();

  saveMovieFrame();
}                                                                                    // applySettings()

void EBTGraphForm::setSettings(EBTGraphSettings newsettings)
{
  currentSettings = newsettings;
  currentFile->setChanged(TRUE);
  zoomer->zoom(0);

  emit newGrFile(this, isOutputGraph);
}


bool EBTGraphForm::parentClose()
{
  if (okToClear())
    {
      hide();
      emit closing(this);
      return TRUE;
    }
  else
    return FALSE;
}


void EBTGraphForm::newGraph() { emit openNewGraph(isOutputGraph); }


QString EBTGraphForm::absFilePath() { return currentFile->absFilePath(); }


void EBTGraphForm::zoomEnable(bool on)
{
  if (on)
    {
      zoomer->setZoomBase();
      zoomOn = TRUE;
      // AMdR: I'm not sure why this connection is made here. Perhaps it is just connected because
      // saveMovieFrame() is not functional if you are not saving a movie
      // I redraw the data to handle properly missing values and non-postive values on logarithmic
      // axes
      // connect(zoomer, SIGNAL(zoomed(const QwtDoubleRect &)), this, SLOT(saveMovieFrame()));

      connect(zoomer, SIGNAL(zoomed(const QRectF &)), this, SLOT(redrawData()), Qt::UniqueConnection);
    }
  else
    disconnect(zoomer, SIGNAL(zoomed(const QRectF &)), this, SLOT(redrawData()));
  //	disconnect(zoomer, SIGNAL(zoomed(const QwtDoubleRect &)), this, SLOT(saveMovieFrame()));
  zoomer->setEnabled(on);
  zoomer->zoom(0);
  if (!on)
    {
      emit newGrFile(this, isOutputGraph);
      zoomOn = FALSE;
    }
}


void EBTGraphForm::axisToLeft()
{
  double newMin, newMax;
  bool   zoomed = (zoomOn && zoomer->zoomRect() != zoomer->zoomBase());

  if (currentSettings.axisLogarithmic(QwtPlot::xBottom))
    {
      double ratio = axisMax(QwtPlot::xBottom)/axisMin(QwtPlot::xBottom);
      newMin       = axisMin(QwtPlot::xBottom)/ratio;
      newMax       = axisMin(QwtPlot::xBottom);
      if (zoomed && (newMin < currentSettings.axisMinVal(QwtPlot::xBottom)))
        {
          newMin = currentSettings.axisMinVal(QwtPlot::xBottom);
          newMax = ratio*newMin;
        }
    }
  else
    {
      newMin = 2*axisMin(QwtPlot::xBottom) - axisMax(QwtPlot::xBottom);
      newMax = axisMin(QwtPlot::xBottom);
      if (zoomed && (newMin < currentSettings.axisMinVal(QwtPlot::xBottom)))
        {
          newMin = currentSettings.axisMinVal(QwtPlot::xBottom);
          newMax = newMin + axisMax(QwtPlot::xBottom) - axisMin(QwtPlot::xBottom);
        }
    }
  thisPlot->setAxisScale(QwtPlot::xBottom, newMin, newMax);

  if (zoomed)
    {
      axisToLeftAction->setEnabled(newMin > currentSettings.axisMinVal(QwtPlot::xBottom));
      axisToRightAction->setEnabled(newMax < currentSettings.axisMaxVal(QwtPlot::xBottom));
      thisPlot->replot();
      saveMovieFrame();
      return;
    }

  currentFile->setChanged(TRUE);
  currentSettings.setAxisMinVal(QwtPlot::xBottom, newMin);
  currentSettings.setAxisMaxVal(QwtPlot::xBottom, newMax);
  emit newGrFile(this, isOutputGraph);
}


void EBTGraphForm::axisToRight()
{
  double newMin, newMax;
  bool   zoomed = (zoomOn && zoomer->zoomRect() != zoomer->zoomBase());

  if (currentSettings.axisLogarithmic(QwtPlot::xBottom))
    {
      double ratio = axisMax(QwtPlot::xBottom)/axisMin(QwtPlot::xBottom);
      newMin       = axisMax(QwtPlot::xBottom);
      newMax       = axisMax(QwtPlot::xBottom)*ratio;
      if (zoomed && (newMax > currentSettings.axisMaxVal(QwtPlot::xBottom)))
        {
          newMin = newMax/ratio;
          newMax = currentSettings.axisMaxVal(QwtPlot::xBottom);
        }
    }
  else
    {
      newMin = axisMax(QwtPlot::xBottom);
      newMax = 2*axisMax(QwtPlot::xBottom) - axisMin(QwtPlot::xBottom);
      if (zoomed && (newMax > currentSettings.axisMaxVal(QwtPlot::xBottom)))
        {
          newMax = currentSettings.axisMaxVal(QwtPlot::xBottom);
          newMin = newMax + axisMin(QwtPlot::xBottom) - axisMax(QwtPlot::xBottom);
        }
    }
  thisPlot->setAxisScale(QwtPlot::xBottom, newMin, newMax);

  if (zoomed)
    {
      axisToLeftAction->setEnabled(newMin > currentSettings.axisMinVal(QwtPlot::xBottom));
      axisToRightAction->setEnabled(newMax < currentSettings.axisMaxVal(QwtPlot::xBottom));
      thisPlot->replot();
      saveMovieFrame();
      return;
    }
  currentFile->setChanged(TRUE);
  currentSettings.setAxisMinVal(QwtPlot::xBottom, newMin);
  currentSettings.setAxisMaxVal(QwtPlot::xBottom, newMax);
  emit newGrFile(this, isOutputGraph);
}


bool EBTGraphForm::hasAutoScale()
{
  // graphs with at least one autoscale axis will be replotted at the end of runs
  return autoScale;
}

bool EBTGraphForm::hasAutoScaleX(uint setNr) { return currentSettings.axisAutoscale(currentSettings.dataSetXAxis(setNr)); }

double EBTGraphForm::axisMin(QwtPlot::Axis axis) { return thisPlot->axisScaleDiv(axis).lowerBound(); }

double EBTGraphForm::axisMax(QwtPlot::Axis axis) { return thisPlot->axisScaleDiv(axis).upperBound(); }

double EBTGraphForm::axisMinX(uint setNr) { return currentSettings.axisMinVal(currentSettings.dataSetXAxis(setNr)); }

double EBTGraphForm::axisMaxX(uint setNr) { return currentSettings.axisMaxVal(currentSettings.dataSetXAxis(setNr)); }

double EBTGraphForm::axisMinY(uint setNr) { return currentSettings.axisMinVal(currentSettings.dataSetYAxis(setNr)); }

double EBTGraphForm::axisMaxY(uint setNr) { return currentSettings.axisMaxVal(currentSettings.dataSetYAxis(setNr)); }

QString EBTGraphForm::fileName() { return currentFile->fileName(); }

void EBTGraphForm::showText(const QString &text)
{
  currentText = text;

  this->toolbarText->setVisible(FALSE);
  // this toolbar only exists to show text; text is now shown in the graph title
  // this->toolbarText->setShown(!text.isEmpty());
  // textInfo->setText(text);
}

void EBTGraphForm::plotSavePNG()
{
  static QString fName(".");

  fName = QFileDialog::getSaveFileName(this, EBTCAPTION + "Save plot", fName, tr("PNG files (*.png)"));

  if (fName.isEmpty()) return;
  if (!fName.endsWith(".png", Qt::CaseInsensitive)) fName.append(QString(".png"));

  QPixmap pixmap = thisPlot->grab();
  QImage  image  = pixmap.toImage().convertToFormat(QImage::Format_Indexed8);
  if (image.save(fName, "PNG"))
    statusBar()->showMessage(QString("Saved file to '%1'").arg(fName), 3000);
  else
    QMessageBox::critical(this, "Warning", QString("Failed to save '%1'").arg(fName));
}

void EBTGraphForm::updateMenu(bool disable)
{
  fileNewAction->setDisabled(disable);
  fileOpenAction->setDisabled(disable);
  fileExitAction->setDisabled(disable);
  layoutEditAction->setDisabled(disable);
}

void EBTGraphForm::deleteFile(QString fName)
{
  if (QFile(fName).exists())
    {
      if (QFile(fName).remove())
        statusBar()->showMessage(QString("\nRemoving file %1.").arg(fName));
      else
        QMessageBox::critical(this, "Warning", QString("\nRemoving file %1 failed.").arg(fName));
    }
}

void EBTGraphForm::plotSaveMovie()
{
  movieName              = QString();
  EBTMovieDialog *dialog = new EBTMovieDialog(this, currentGIFDelay);
  if (dialog->exec() == QDialog::Accepted)
    {
      currentGIFDelay = dialog->delay();

      QStringList files               = dialog->selectedFiles();
      if (!files.isEmpty()) movieName = files[0];
    }
  delete dialog;

  if (movieName.isEmpty()) return;
  if (!movieName.endsWith(".gif", Qt::CaseInsensitive)) movieName.append(QString(".gif"));

  if (QFile::exists(movieName))
    {
      if (QMessageBox::warning(this, EBTCAPTION + "Overwrite File", QString("Overwrite\n'%1'?").arg(movieName), QMessageBox::Yes | QMessageBox::No, QMessageBox::No) == QMessageBox::No)
        return;
    }

  QPixmap pixmap = thisPlot->grab();
  QImage  image  = pixmap.toImage().convertToFormat(QImage::Format_Indexed8);
  movie          = new EBTGIF(movieName);
  movie->initializeGIF(currentGIFDelay, image);

  recordingMovie = TRUE;
  updateMenu(TRUE);

  plotSaveMovieAction->disconnect(SIGNAL(triggered()));
  connect(plotSaveMovieAction, SIGNAL(triggered()), this, SLOT(plotFinishMovie()));
  plotSaveMovieAction->setIcon(QIcon(":/images/moviestop.png"));
  plotSaveMovieAction->setText("Stop saving movie...");
}

void EBTGraphForm::saveMovieFrame()
{
  if (!recordingMovie) return;
  QPixmap pixmap = thisPlot->grab();
  QImage  image  = pixmap.toImage().convertToFormat(QImage::Format_Indexed8);
  movie->writeGIFImage(image);
}

void EBTGraphForm::plotFinishMovie()
{
  movie->finishGIF();
  plotCleanupMovie(FALSE);
}

void EBTGraphForm::plotCleanupMovie(bool discard)
{
  if (discard) deleteFile(movieName);
  recordingMovie = FALSE;
  updateMenu(FALSE);
  if (movie)
    {
      delete movie;
      movie = NULL;
    }
  movieName = QString();

  plotSaveMovieAction->disconnect(SIGNAL(triggered()));
  connect(plotSaveMovieAction, SIGNAL(triggered()), this, SLOT(plotSaveMovie()));
  plotSaveMovieAction->setIcon(QIcon(":/images/movie.png"));
  plotSaveMovieAction->setText("Start saving movie...");
}


bool EBTGraphForm::makeHistogram(uint setNr)
{
  uint    histDataLength, barNr, ii;
  double  xMin     = axisMinX(setNr);
  double  xMax     = axisMaxX(setNr);
  double  yMin     = axisMinY(setNr);
  int     numBars  = currentSettings.dataSetHistNumBars(setNr);
  double  barWidth = 0;
  double *xHistData, *yHistData;

  if (currentSettings.axisLogarithmic(currentSettings.dataSetXAxis(setNr)))
    barWidth = (myLog10(xMax) - myLog10(xMin))/numBars;

  else
    barWidth     = (xMax - xMin)/numBars;
  histDataLength = 1 + 3*numBars;

  xHistData = new double[histDataLength];
  yHistData = new double[histDataLength];

  for (ii = 0; ii < histDataLength; ii++)
    {
      xHistData[ii] = xMin;
      yHistData[ii] = 0.0;
    }
  for (ii = 0; ii < (uint)currentSettings.dataSetSize(setNr); ii++)
    {
      if ((currentSettings.dataSetXData(setNr)[ii] > xMin) && (currentSettings.dataSetXData(setNr)[ii] < xMax))
        {
          if (currentSettings.axisLogarithmic(currentSettings.dataSetXAxis(setNr)))
            barNr = (uint)floor((myLog10(currentSettings.dataSetXData(setNr)[ii]) - myLog10(xMin))/barWidth);
          else
            barNr = (uint)floor((currentSettings.dataSetXData(setNr)[ii] - xMin)/barWidth);
          yHistData[barNr] += currentSettings.dataSetYData(setNr)[ii];
        }
    }
  xHistData[histDataLength - 1] = xMax;
  yHistData[histDataLength - 1] = yMin;
  for (int jj = numBars - 1; jj >= 0; jj--)
    {
      if (currentSettings.axisLogarithmic(currentSettings.dataSetXAxis(setNr)))
        {
          xHistData[3*jj + 2] = qMin(xMax, pow(10.0, myLog10(xMin) + (jj + 1)*barWidth));
          xHistData[3*jj + 1] = xHistData[3*jj + 2]/pow(10.0, barWidth);
        }
      else
        {
          xHistData[3*jj + 2] = qMin(xMax, xMin + (jj + 1)*barWidth);
          xHistData[3*jj + 1] = xHistData[3*jj + 2] - barWidth;
        }
      xHistData[3*jj] = xHistData[3*jj + 1];

      yHistData[3*jj + 2] = (yHistData[jj] > yMin) ? yHistData[jj] : yMin;
      yHistData[3*jj + 1] = yHistData[3*jj + 2];
      yHistData[3*jj]     = yMin;
    }
  currentSettings.clearDataSet(setNr);
  histYMax = yMin;
  for (ii = 0; ii < histDataLength; ii++)
    {
      currentSettings.dataSetXAppend(setNr, xHistData[ii]);
      currentSettings.dataSetYAppend(setNr, yHistData[ii]);
      if (yHistData[ii] > histYMax) histYMax = yHistData[ii];
    }

  delete[] xHistData;
  delete[] yHistData;

  return TRUE;
}

void EBTGraphForm::graphShowAndRaiseSlot()
{
  show();
#ifdef Q_OS_WIN
  showNormal();
#endif
  raise();
  setHideFlag(FALSE);
  if (graphMenuItem) graphMenuItem->setIcon(QIcon(":/images/check_checked.png"));
}
