#include <QFileDialog>
#include <qmessagebox.h>
#include "ebtmodel.h"
#include "integrator/ebttooldefs.h"

EBTModel::EBTModel()
{
    libHandle = NULL;
    startup = NULL;
    cycleinit = NULL;
    cyclestep = NULL;
    shutdown = NULL;
    
    setDescriptionText("Model");
    setOpenText("Model (*.c)") ;
    
    cFile.setTemplateFile("templates/ebttmpl.c");
    cFile.setFileExtension("c");
    cFile.setDescriptionText("Model program");
    cFile.setSaveText(cFile.descriptionText() + " (*." + cFile.fileExtension() + ")") ;
    cFile.setOpenText(cFile.saveText());
    
    hFile.setTemplateFile("templates/ebttmpl.h");
    hFile.setFileExtension("h");
    hFile.setDescriptionText("Model header");
    hFile.setSaveText(hFile.descriptionText() + " (*." + hFile.fileExtension() + ")") ;
    hFile.setOpenText(hFile.saveText());
}


EBTModel::~EBTModel()
{
    libUnload();
}


bool EBTModel::open()
{
    QString fName = QFileDialog::getOpenFileName(NULL, "EBTtool -- file open", dirPath(), openText());
    if (!fName.isEmpty())
    {
	QFileInfo fi(fName);
	QString basename = fi.absolutePath() + "/" + fi.completeBaseName();
	
	//waarschuwing als  .h ontbreekt
	QString hFilename = basename + QString(".") + hFile.fileExtension();
	if(!QFile::exists(hFilename))
	{
	    QMessageBox::warning(0, EBTCAPTION + "File missing",
				 QString("The file %1 does not exist.")
				 .arg(hFilename));
	    return FALSE;
	}
	setFileName(basename);
	return TRUE;
    }
    else
	return FALSE;
    
}

bool EBTModel::save()
{
    if(!cFile.save())
	return FALSE;
    if(!hFile.save())
	return FALSE;
    return TRUE;
}

void EBTModel::setFileName(const QString& str)
{
    EBTFile::setFileName(str);  
    hFile.setFileName(str + QString(".") + hFile.fileExtension());
    cFile.setFileName(str + QString(".") + cFile.fileExtension());
}

bool EBTModel::initHandles()
{
    if(fileName().isEmpty())
    {
	QMessageBox::warning(0, EBTCAPTION + "No model selected",
			     QString("No model has been selected yet."));
	return FALSE;
    }
    
    libUnload();
    
    // lib is loaded using the basename of the model files (without extension: .so, .dll, dylib, ...)
    libHandle = new QLibrary(absFilePath() + "module");
    
#if defined(Q_OS_LINUX)
    if(!QFile::exists(libHandle->fileName() + ".so"))
#elif defined(Q_OS_MACOS)
    if(!QFile::exists(libHandle->fileName() + ".dylib"))
#elif defined(Q_OS_WIN)
    if(!QFile::exists(libHandle->fileName() + ".dll"))
#endif
    {
	QMessageBox::warning(0, EBTCAPTION + "No such lib file",
			     QString("The lib file %1 could not be found.")
			     .arg(libHandle->fileName()));
	return FALSE;
    }
    
    startup   = (functionType1) libHandle->resolve("StartUp");
    if (!startup)
	startup   = (functionType1) libHandle->resolve("_StartUp");
    if (!startup)
    {
	messageLoadError("StartUp");
	return FALSE;
    }
    
    cycleinit = (functionType2) libHandle->resolve("CycleInit");
    if (!cycleinit)
	cycleinit = (functionType2) libHandle->resolve("_CycleInit");
    if (!cycleinit)
    {
	messageLoadError("CycleInit");
	return FALSE;
    }
    
    cyclestep = (functionType2) libHandle->resolve("CycleStep");
    if (!cyclestep)
	cyclestep = (functionType2) libHandle->resolve("_CycleStep");
    if (!cyclestep)
    {
	messageLoadError("CycleStep");
	return FALSE;
    }
    
    shutdown = (functionType3) libHandle->resolve("ShutDown");
    if (!shutdown)
	shutdown = (functionType3) libHandle->resolve("_ShutDown");
    if (!shutdown)
    {
	messageLoadError("ShutDown");
	return FALSE;
    }
    
    handlesignal = (functionType3) libHandle->resolve("HandleSignal");
    if (!handlesignal)
	handlesignal = (functionType3) libHandle->resolve("_HandleSignal");
    if (!handlesignal)
    {
	messageLoadError("HandleSignal");
	return FALSE;
    }
    
    errormsg = (functionType4) libHandle->resolve("exp_error_msg");
    if(!errormsg)
	errormsg = (functionType4) libHandle->resolve("_exp_error_msg");
    if(!errormsg)
    {
	messageLoadError("errormsg");
	return FALSE;
    }
    
    outputvarnum = (functionType2) libHandle->resolve("exp_output_var_nr");
    if (!outputvarnum )
	outputvarnum = (functionType2) libHandle->resolve("_exp_output_var_nr");
    if (!outputvarnum )
    {
	messageLoadError("output_var_nr");
	return FALSE;
    }
    
    environdim = (functionType2) libHandle->resolve("exp_environ_dim");
    if (!environdim )
	environdim = (functionType2) libHandle->resolve("_exp_environ_dim");
    if (!environdim )
    {
	messageLoadError("environ_dim");
	return FALSE;
    }
    
    populationnum = (functionType2) libHandle->resolve("exp_population_nr");
    if (!populationnum )
	populationnum = (functionType2) libHandle->resolve("_exp_population_nr");
    if (!populationnum )
    {
	messageLoadError("population_nr");
	return FALSE;
    }
    
    istatedim = (functionType2) libHandle->resolve("exp_i_state_dim");
    if (!istatedim )
	istatedim = (functionType2) libHandle->resolve("_exp_i_state_dim");
    if (!istatedim )
    {
	messageLoadError("i_state_dim");
	return FALSE;
    }
    
    iconstdim = (functionType2) libHandle->resolve("exp_i_const_dim");
    if (!iconstdim )
	iconstdim = (functionType2) libHandle->resolve("_exp_i_const_dim");
    if (!iconstdim )
    {
	messageLoadError("i_const_dim");
	return FALSE;
    }
    
    parameternum = (functionType2) libHandle->resolve("exp_parameter_nr");
    if (!parameternum )
	parameternum = (functionType2) libHandle->resolve("_exp_parameter_nr");
    if (!parameternum )
    {
	messageLoadError("parameter_nr");
	return FALSE;
    }
    
    output = (functionType5) libHandle->resolve("exp_output");
    if (!output)
	output = (functionType5) libHandle->resolve("_exp_output");
    if (!output)
    {
	messageLoadError("output");
	return FALSE;
    }
    
    env = (functionType5) libHandle->resolve("exp_env");
    if (!env)
	env = (functionType5) libHandle->resolve("_exp_env");
    if (!env)
    {
	messageLoadError("env");
	return FALSE;
    }
    
    pop = (functionType6) libHandle->resolve("exp_pop");
    if (!pop)
	pop = (functionType6) libHandle->resolve("_exp_pop");
    if (!pop)
    {
	messageLoadError("pop");
	return FALSE;
    }
    
    cohortnum = (functionType7) libHandle->resolve("exp_CohortNo");
    if (!cohortnum)
	cohortnum = (functionType7) libHandle->resolve("_exp_CohortNo");
    if (!cohortnum)
    {
	messageLoadError("CohortNo");
	return FALSE;
    }
    
    popidcard = (functionType6) libHandle->resolve("exp_popIDcard");
    if (!popidcard)
	popidcard = (functionType6) libHandle->resolve("_exp_popIDcard");
    if (!popidcard)
    {
	messageLoadError("popIDcard");
	return FALSE;
    }
    
    statelabel = (functionType8) libHandle->resolve("exp_statelabel");
    if (!statelabel)
	statelabel = (functionType8) libHandle->resolve("_exp_statelabel");
    if (!statelabel)
    {
	messageLoadError("StateLabel");
	return FALSE;
    }
    
    return TRUE; 
}

void EBTModel::messageLoadError(QString handle)
{
    QMessageBox::warning(0, EBTCAPTION + "Model load problem",
			 QString("Loading of %1 unsuccessful.")
			 .arg(handle));
}

int EBTModel::startUp(int numArgs, char **theArgs)
{
    if(libLoaded()) 
	return (*startup)(numArgs, theArgs);
    else
	return LIB_NOT_LOADED;
}
	
int EBTModel::cycleInit()
{
    if(libLoaded()) 
	return (*cycleinit)(); 
    else 	
	return LIB_NOT_LOADED;
}

int EBTModel::cycleStep()
{
    if(libLoaded()) 
	return (*cyclestep)(); 
    else
	return LIB_NOT_LOADED;
}

int EBTModel::shutDown(int flag)
{
    if(libLoaded())
    {
	int retVal = (*shutdown)(flag);
//	libHandle->unload();
//	delete libHandle;
//	libHandle = NULL;
	return retVal;
    }
    else
	return LIB_NOT_LOADED;
}

int EBTModel::handleSignal(int signl)
{
    if(libLoaded())
	return (*handlesignal)(signl);
    else
	return LIB_NOT_LOADED;
}

QString EBTModel::stateLabel(int popnr)
{
    if(libLoaded())
	return QString((*statelabel)(popnr));
    else
	return QString();
}

QString EBTModel::soName()
{
    return QLibrary(absFilePath() + "module").fileName();
}

QString EBTModel::exportState()
{
    QString returnValue = QString();
 
    if(!libLoaded())
    {
	QMessageBox::warning(0, "debug", "lib not loaded");
	return returnValue;
    }
    
    double output;				
    for(uint ii = 0; ii < environDim(); ii++)			// write environment
    {
	if(ii == 0)
	    output = 0;
	else
	{
	    returnValue+="\t";
	    output = environment()[ii];
	}
	returnValue += formatDouble(output);
    }
    returnValue += "\n\n";
    
    for(uint popNr = 0; popNr < populationNum(); popNr++)		// write population states
    {
//	for(int cohortNr = 0; cohortNr <  cohortNum()[popNr]; cohortNr++)
	for(int cohortNr = cohortNum()[popNr] - 1; cohortNr >= 0; cohortNr--)
	{
	    for(uint colNr = 0; colNr < 1 + iStateDim() + iConstDim(); colNr++)
	    {
		if(colNr>0)
		    returnValue += "\t";
		if(colNr <= iStateDim())
		    output = *(((double *)population()[popNr])+cohortNr*(1 + iStateDim())+colNr);	
		else
		    output = *(((double *)popIDcard()[popNr])+cohortNr*iConstDim()+(colNr-iStateDim()-1));
		returnValue += formatDouble(output);
	    }
	    returnValue += "\n";
	}
	returnValue += "\n";
    }
    returnValue += "\n";
    return returnValue;    
}


void EBTModel::libUnload()
{
  if(libLoaded()) libHandle->unload();
  if (libHandle)
    {
      delete libHandle;
      libHandle = NULL;
    }
}
