/*
 * ebtplotcurve.cpp
 *
 *  Created on: Mar 14, 2010
 *      Author: andre
 */
#include <float.h>
#include <math.h>
#include "ebtplotcurve.h"
#include "ebtdefines.h"

#include <qpainter.h>
#include <qbitarray.h>

#include <QtMath>

#include "qwt_scale_map.h"
#include "qwt_painter.h"
#include "qwt_clipper.h"
#include "qwt_curve_fitter.h"
#include "qwt_symbol.h"

#ifdef DEBUG
#include <QtDebug>
#endif

void EBTPlotCurve::drawCurve( QPainter *painter, int style,
    const QwtScaleMap &xMap, const QwtScaleMap &yMap,
    const QRectF &canvasRect, int from, int to ) const
{
  switch (style)
    {
        case Lines:
            drawLines(painter, xMap, yMap, canvasRect, from, to);
            break;
        case Sticks:
            drawSticks(painter, xMap, yMap, canvasRect, from, to);
            break;
        case Steps:
            drawSteps(painter, xMap, yMap, canvasRect, from, to);
            break;
        case Dots:
            drawDots(painter, xMap, yMap, canvasRect, from, to);
            break;
        case NoCurve:
        default:
            break;
    }
}

void EBTPlotCurve::drawLines( QPainter *painter,
    const QwtScaleMap &xMap, const QwtScaleMap &yMap,
    const QRectF &canvasRect, int from, int to ) const
{
  if ( to <= from ) return;
  int		first, last;
  const bool doAlign = QwtPainter::roundingAlignment( painter );

  last = from;
  QPolygonF 	polyline;
  while (last <= to)
    {
      first = last;
      while ((first <= to) && (isnan(sample(first).x()) || isnan(sample(first).y()))) first++;

      last = first;
      while ((last <= to) && (!isnan(sample(last).x())) && (!isnan(sample(last).y()))) last++;
      if (last <= (first+1)) continue;

      polyline.resize( last-first );
      QPointF *points = polyline.data();

      for ( int i = first; i < last; i++ )
	{
	  const QPointF sample = this->data()->sample( i );

	  double x = xMap.transform( sample.x() );
          double y = yMap.transform( sample.y() );
          if ( doAlign )
            {
              x = qRound( x );
              y = qRound( y );
            }

          points[i - first].rx() = x;
          points[i - first].ry() = y;
	}

      QwtPainter::drawPolyline( painter, polyline );

      if ( brush().style() != Qt::NoBrush )
          fillCurve( painter, xMap, yMap, canvasRect, polyline );
    }
}

void EBTPlotCurve::drawSymbols( QPainter *painter, const QwtSymbol &symbol,
    const QwtScaleMap &xMap, const QwtScaleMap &yMap,
    const QRectF &canvasRect, int from, int to ) const
{
  const bool doAlign = QwtPainter::roundingAlignment( painter );

  bool usePixmap = (symbol.cachePolicy() == QwtSymbol::Cache );
  if ( usePixmap && !doAlign )
    {
      // Don't use the pixmap, when the paint device
      // could generate scalable vectors

      usePixmap = false;
    }

  if ( usePixmap )
    {
      QPixmap pm( symbol.boundingRect().size() );
      pm.fill( Qt::transparent );

      const double pw2 = 0.5 * pm.width();
      const double ph2 = 0.5 * pm.height();

      QPainter p( &pm );
      p.setRenderHints( painter->renderHints() );
      symbol.drawSymbol( &p, QPointF( pw2, ph2 ) );
      p.end();

      for ( int i = from; i <= to; i++ )
	{
	  const QPointF sample = this->data()->sample( i );
          if (isnan(sample.x()) || isnan(sample.y())) continue;

          double xi = xMap.transform( sample.x() );
          double yi = yMap.transform( sample.y() );
          if ( doAlign )
            {
              xi = qRound( xi );
              yi = qRound( yi );
            }

          if ( canvasRect.contains( xi, yi ) )
            {
              const int left = qCeil( xi ) - pw2;
              const int top = qCeil( yi ) - ph2;

              painter->drawPixmap( left, top, pm );
            }
	}
    }
  else
    {
      const int chunkSize = 500;

      for ( int i = from; i <= to; i += chunkSize )
	{
	  const int n = qMin( chunkSize, to - i + 1 );

          QPolygonF points;
          for ( int j = 0; j < n; j++ )
            {
              const QPointF sample = this->data()->sample( i + j );
              if (isnan(sample.x()) || isnan(sample.y())) continue;

              const double xi = xMap.transform( sample.x() );
              const double yi = yMap.transform( sample.y() );

              if ( canvasRect.contains( xi, yi ) )
        	points += QPointF( xi, yi );
            }

          if ( points.size() > 0 )
            symbol.drawSymbols( painter, points );
	}
    }
}
