#include "graphsettingsform.h"

#include <qvariant.h>
#include <qimage.h>
#include <qpixmap.h>

#include "qmessagebox.h"
#include "qcolordialog.h"
#include "qfontdialog.h"
#include <QShortcut>

EBTGraphSettings gs;
QMap<int, QwtPlot::Axis> comboAxisIndex;
QMap<int, QwtPlotCurve::CurveStyle> comboCurveStyleIndex;
QMap<int, QwtSymbol::Style> comboSymbolIndex;
QMap<int, QColor> comboSymbolColorIndex;
QMap<int, QColor> comboLineColorIndex;
enum {outputValues = 0, computedValues, fixedValues};

QwtSymbol *setSymbol;
QPen setLinePen;
QFont titleFont;
QFont legendFont;
QFont axisTitleFont;

/*
 *  Constructs a GraphSettingsForm as a child of 'parent', with the
 *  name 'name' and widget flags set to 'f'.
 *
 *  The dialog will by default be modeless, unless you set 'modal' to
 *  true to construct a modal dialog.
 */


GraphSettingsForm::GraphSettingsForm(QWidget* parent, bool modal, Qt::WindowFlags fl)
    : QDialog(parent, fl)
{
    setupUi(this);
    setModal(modal);

    init();
}


/*
 *  Destroys the object and frees any allocated resources
 */
GraphSettingsForm::~GraphSettingsForm()
{
    // no need to delete child widgets, Qt does it all for us
}


/*
 *  Sets the strings of the subwidgets using the current
 *  language.
 */
void GraphSettingsForm::languageChange()
{
    retranslateUi(this);
}


void GraphSettingsForm::init()
{    
    connect(this, SIGNAL(newSettings(EBTGraphSettings)), 
	    parentWidget(), SLOT(setSettings(EBTGraphSettings)));
    prevDisplayedSet = 0;
    prevDisplayedAxis = 0;
    comboSymbolColorIndex.clear();
    comboLineColorIndex.clear();
    fillComboBoxes();
    
    QShortcut *escAct = new QShortcut(Qt::Key_Escape, this);
    connect(escAct, SIGNAL(activated()), this, SLOT(reject()));
    delete escAct;
}


void GraphSettingsForm::acceptChanges()
{
    applyChanges();
    accept();
}


void GraphSettingsForm::applyChanges()
{
    if(spinSetSelect->value() > 0)
	applySetChanges(spinSetSelect->value());
    applyAxisChanges(comboAxisSelect->currentIndex());
    applyGeneralChanges();
    emit newSettings(gs);
}


void GraphSettingsForm::clearSettings()
{
  if (QMessageBox::question(this, EBTCAPTION + "Clear settings", QString("Are you sure you want to delete\nall graph settings?"),
                            QMessageBox::Yes | QMessageBox::No, QMessageBox::No) == QMessageBox::No)
    return;
  prevDisplayedSet        = 0;
  EBTGraphSettings *newgs = new EBTGraphSettings;
  setCurrentSettings(*newgs, outGraph);
  delete newgs;
}


void GraphSettingsForm::setCurrentSettings(EBTGraphSettings graphsettings, bool oGraph)
{
    gs = graphsettings;
    outGraph = oGraph;
    if(outGraph)
	labelGraphType->setText(OUTPUTDESC);
    else
	labelGraphType->setText(STATEDESC);
    
    fillSpinBoxes(gs.dataSetMapSize());
    spinSetSelect->setValue(spinSetSelect->minimum());
    //    spinStyleSetSelect->setValue(firstSet); gebeurt automatisch na bovenstaande regel
    if(spinSetSelect->minimum() == 0)  // in this case valueChanged is not emitted
	showSetSettings(0);  // but the set _should_ be shown
    checkStyleSetAll->setChecked(gs.allSetsSameAs() >= 0);
    
    comboAxisSelect->setCurrentIndex(0);
    showAxisSettings(0);
    
    if(gs.showGrid())
    {
	if(gs.showFineGrid())
	    radioGridFine->setChecked(TRUE);
	else
	    radioGridCoarse->setChecked(TRUE);
    }
    else
	radioGridNone->setChecked(TRUE);
    
    editTitle->setText(gs.plotTitle());
    titleFont = gs.plotTitleFont();

    checkLegendShow->setChecked(gs.showLegend());

    comboFrameType->setCurrentIndex(gs.legendFramed());
    valueLegendFrameLine->setValue(gs.legendFrameWidth());

    if (gs.legendPosition() == QwtPlot::BottomLegend)
      comboLegendPosition->setCurrentIndex(comboLegendPosition->findText(QString("Bottom")));
    else if (gs.legendPosition() == QwtPlot::RightLegend)
      comboLegendPosition->setCurrentIndex(comboLegendPosition->findText(QString("Right")));
    else if (gs.legendPosition() == QwtPlot::TopLegend)
      comboLegendPosition->setCurrentIndex(comboLegendPosition->findText(QString("Top")));
    else
      comboLegendPosition->setCurrentIndex(comboLegendPosition->findText(QString("Floating")));

    valueLegendX->setValue(gs.legendXPosition());
    valueLegendY->setValue(gs.legendYPosition());

    checkTimestampShow->setChecked(gs.showTimestamp());
    checkDescrShow->setChecked(gs.showDescription());
    checkDescrDefault->setChecked(gs.defaultDescription());
    editDescr->setText(gs.plotDescription());
    legendFont = gs.legendFont();
    setLegendShow(checkLegendShow->isChecked());
}


void GraphSettingsForm::showAxisSettings(int ii)
{
    if(ii != prevDisplayedAxis)
	applyAxisChanges(prevDisplayedAxis);
    
    QwtPlot::Axis a = comboAxisIndex[ii];
    if(gs.axisExists(a))
    {
	if(gs.axisLogarithmic(a))
	    radioAxisLog->setChecked(TRUE);
	else
	    radioAxisLin->setChecked(TRUE);
	editAxisLabel->setText(gs.axisTitle(a));
	axisTitleFont = gs.axisTitleFont(a);
	checkAxisAutoscale->setChecked(gs.axisAutoscale(a));
	setAutoscale(gs.axisAutoscale(a));
	prevDisplayedAxis = ii;
    }
}


void GraphSettingsForm::showSetSettings(int setNo)
{
    if(setNo > 0 && prevDisplayedSet > 0 && setNo != prevDisplayedSet)
	applySetChanges(prevDisplayedSet); 
    
    int ii = setNo - 1;
    buttonSetClear->setDisabled(ii < 0);
    
    comboValueType->setDisabled((ii < 0)||(!outGraph));
    spinPopSelect->setDisabled((ii < 0)||(outGraph));
    spinXColumn->setDisabled(ii < 0);
    spinYColumn->setDisabled(ii < 0);
    comboXAxisPrimary->setDisabled(ii < 0);
    comboYAxisPrimary->setDisabled(ii < 0);
    editLegendLabel->setDisabled(ii < 0);
    checkStyleSetAll->setDisabled(ii < 0);
    groupCurve->setDisabled(ii < 0);
    groupSymbol->setDisabled(ii < 0);
    groupLine->setDisabled(ii < 0);
    
    if(ii < 0) 
	return;
    
    if(gs.dataSetExists(ii))
    {
	// first tab page
	spinXColumn->setValue(gs.dataSetXCol(ii));
	spinYColumn->setValue(gs.dataSetYCol(ii));
	if(gs.dataSetDataType(ii) == COMPUTEDVALUES)
	    comboValueType->setCurrentIndex(computedValues);
	else
	    if(gs.dataSetDataType(ii) == FIXEDVALUES)
		comboValueType->setCurrentIndex(fixedValues);
	else
	    comboValueType->setCurrentIndex(outputValues);
	spinPopSelect->setValue(gs.dataSetPopulationNr(ii));
	if(gs.dataSetXAxis(ii) == QwtPlot::xTop)
	    comboXAxisPrimary->setCurrentIndex(0);
	else
	    comboXAxisPrimary->setCurrentIndex(1);
	if(gs.dataSetYAxis(ii) == QwtPlot::yRight)
	    comboYAxisPrimary->setCurrentIndex(0);
	else
	    comboYAxisPrimary->setCurrentIndex(1);
	editLegendLabel->setText(gs.dataSetLegendLabel(ii));
	
	// second tab page
	int jj;
	if(gs.allSetsSameAs() < 0)
	    jj = ii;
	else 
	    jj = gs.allSetsSameAs();
	
	setSymbol = gs.dataSetCurveSymbol(jj);
	comboSymbolStyle->setCurrentIndex(comboSymbolIndex.values().indexOf(setSymbol->style()));
	spinSymbolSize->setValue(setSymbol->size().height());

	setComboSymbolColor(QColor(setSymbol->pen().color()));
	comboSymbolFill->setCurrentIndex(setSymbol->brush().style());
	spinSymbolPenWidth->setValue(setSymbol->pen().width());
	enableGroupSymbol(setSymbol->style());

	int curveStyleKey = comboCurveStyleIndex.values().indexOf(gs.dataSetCurveStyle(jj));
	comboCurveStyle->setCurrentIndex(curveStyleKey);
	setCurveStyle(curveStyleKey);
	
	setLinePen = gs.dataSetLinePen(jj);
	comboLineStyle->setCurrentIndex(setLinePen.style());
	spinLineWidth->setValue(setLinePen.width());
	setComboLineColor(setLinePen.color());
	enableGroupLine(setLinePen.style());
	
	prevDisplayedSet = ii + 1;
    }    
}


void GraphSettingsForm::newDataSet()
{
    gs.addDataSet();
    
    fillSpinBoxes(gs.dataSetMapSize());
    spinSetSelect->setValue(gs.dataSetMapSize());
}


void GraphSettingsForm::removeDataSet()
{
  int ii = spinSetSelect->value() - 1;

  if (ii >= 0)
    {
      if (QMessageBox::question(this, EBTCAPTION + "Clear data set", QString("Are you sure you want to delete\nthe current data set?"),
                                QMessageBox::Yes | QMessageBox::No, QMessageBox::No) == QMessageBox::No)
        return;
      gs.removeDataSet(ii);
      prevDisplayedSet = 0;
      fillSpinBoxes(gs.dataSetMapSize());
      if (ii < spinSetSelect->minimum())
        {
          ii = spinSetSelect->minimum();
          spinSetSelect->setValue(ii);                                               // in this case valueChanged is not emitted
          showSetSettings(ii);                                                       // but another set _should_ be shown
        }
      else
        spinSetSelect->setValue(ii);
    }
}


void GraphSettingsForm::fillComboBoxes()
{
    comboValueType->insertItem(outputValues, "Output values");
    comboValueType->insertItem(computedValues, "Computed values");
    comboValueType->insertItem(fixedValues, "Fixed values");
    
    comboAxisSelect->insertItems(0, QStringList() << "Primary X axis" << "Secondary X axis" << "Primary Y axis" << "Secondary Y axis");
    comboAxisIndex[0] = QwtPlot::xBottom;
    comboAxisIndex[1] = QwtPlot::xTop;
    comboAxisIndex[2] = QwtPlot::yLeft;
    comboAxisIndex[3] = QwtPlot::yRight;
    
    comboCurveStyle->insertItems(0, QStringList() << "None" << "Line" << "Bars" << "Histogram");
    comboCurveStyleIndex[0] = QwtPlotCurve::NoCurve;
    comboCurveStyleIndex[1] = QwtPlotCurve::Lines;
    comboCurveStyleIndex[2] = QwtPlotCurve::Sticks;
    comboCurveStyleIndex[3] = QwtPlotCurve::UserCurve;
    // comboCurveStyle->insertItem("Steps");
    // comboCurveStyleIndex[2] = QwtPlotCurve::Steps;
    // comboCurveStyle->insertItem("Dots");
    // comboCurveStyleIndex[3] = QwtPlotCurve::Dots;
    // comboCurveStyle->insertItem("Spline");
    // comboCurveStyleIndex[4] = QwtCurve::Spline;
    
    comboSymbolStyle->insertItems(0, QStringList() << "None" << "Ellipse" << "Rectangle" << "Diamond" << "Triangle"
	      << "Triangle pointing down" << "Triangle pointing up" << "Triangle pointing left" << "Triangle pointing right"
	      << "Cross" << "X Cross");
    comboSymbolIndex[ 0] = QwtSymbol::NoSymbol;
    comboSymbolIndex[ 1] = QwtSymbol::Ellipse;
    comboSymbolIndex[ 2] = QwtSymbol::Rect;
    comboSymbolIndex[ 3] = QwtSymbol::Diamond;
    comboSymbolIndex[ 4] = QwtSymbol::Triangle;
    comboSymbolIndex[ 5] = QwtSymbol::DTriangle;
    comboSymbolIndex[ 6] = QwtSymbol::UTriangle;
    comboSymbolIndex[ 7] = QwtSymbol::LTriangle;
    comboSymbolIndex[ 8] = QwtSymbol::RTriangle;
    comboSymbolIndex[ 9] = QwtSymbol::Cross;
    comboSymbolIndex[10] = QwtSymbol::XCross;
    
    comboSymbolFill->insertItem(Qt::NoBrush,          "None");
    comboSymbolFill->insertItem(Qt::SolidPattern,     "solid (100%)");
    comboSymbolFill->insertItem(Qt::Dense1Pattern,    "94%");
    comboSymbolFill->insertItem(Qt::Dense2Pattern,    "88%");
    comboSymbolFill->insertItem(Qt::Dense3Pattern,    "63%");
    comboSymbolFill->insertItem(Qt::Dense4Pattern,    "50%");
    comboSymbolFill->insertItem(Qt::Dense5Pattern,    "37%");
    comboSymbolFill->insertItem(Qt::Dense6Pattern,    "12%");
    comboSymbolFill->insertItem(Qt::Dense7Pattern,    "6%");
    comboSymbolFill->insertItem(Qt::HorPattern,       "horizontal lines");
    comboSymbolFill->insertItem(Qt::VerPattern,       "vertical lines");
    comboSymbolFill->insertItem(Qt::CrossPattern,     "crossing lines");
    comboSymbolFill->insertItem(Qt::BDiagPattern,     "diagonal lines (directed /)");
    comboSymbolFill->insertItem(Qt::FDiagPattern,     "diagonal lines (directed \\)");
    comboSymbolFill->insertItem(Qt::DiagCrossPattern, "diagonal crossing lines");

    setComboSymbolColor(QColor(Qt::black));
    setComboSymbolColor(QColor(Qt::red));
    setComboSymbolColor(QColor(Qt::blue));
    setComboSymbolColor(QColor(Qt::green));
    setComboSymbolColor(QColor(Qt::magenta));
    setComboSymbolColor(QColor(Qt::cyan));
    setComboSymbolColor(QColor(Qt::yellow));
    setComboSymbolColor(QColor(Qt::gray));
    
    comboLineStyle->insertItem(Qt::NoPen,          "None");
    comboLineStyle->insertItem(Qt::SolidLine,      "Solid line");
    comboLineStyle->insertItem(Qt::DashLine,       "Dash line");
    comboLineStyle->insertItem(Qt::DotLine,        "Dot line");
    comboLineStyle->insertItem(Qt::DashDotLine,    "Dash dot line");
    comboLineStyle->insertItem(Qt::DashDotDotLine, "Dash dot dot line");
    
    setComboLineColor(Qt::black);
    setComboLineColor(Qt::red);
    setComboLineColor(Qt::blue);
    setComboLineColor(Qt::green);
    setComboLineColor(Qt::magenta);
    setComboLineColor(Qt::cyan);
    setComboLineColor(Qt::yellow);
    setComboLineColor(Qt::gray);
    
    comboFrameType->addItem(QString("No Frame"));
    comboFrameType->addItem(QString("Framed"));

    comboLegendPosition->addItem(QString("Bottom"));
    comboLegendPosition->addItem(QString("Right"));
    comboLegendPosition->addItem(QString("Top"));
    comboLegendPosition->addItem(QString("Floating"));
}


void GraphSettingsForm::newSymbolColor()
{
    QColor symbolColor = QColorDialog::getColor(setSymbol->pen().color() , this );
    if ( !symbolColor.isValid() )
	return;
    setComboSymbolColor(symbolColor);
}

void GraphSettingsForm::newLineColor()
{
    QColor lineColor = QColorDialog::getColor(setLinePen.color() , this );
    if ( !lineColor.isValid() )
	return;
    setComboLineColor(lineColor);
}


void GraphSettingsForm::setComboSymbolColor( QColor color )
{
    if(!comboSymbolColorIndex.values().contains( color) )
    {
        QPixmap pm(QSize(40, 40));
	pm.fill(color);
	comboSymbolColor->addItem(QIcon(pm), QString());
	comboSymbolColorIndex[comboSymbolColor->count() - 1] = color;
    }
    int colorKey = comboSymbolColorIndex.values().indexOf(color);
    comboSymbolColor->setCurrentIndex(colorKey);
}


void GraphSettingsForm::setComboLineColor( QColor color )
{
    if(!comboLineColorIndex.values().contains( color) )
    {
	QPixmap pm(QSize(40, 40));
	pm.fill(color);
	comboLineColor->addItem(QIcon(pm), QString());
	comboLineColorIndex[comboLineColor->count() - 1] = color;
    }
    int colorKey = comboLineColorIndex.values().indexOf(color);
    comboLineColor->setCurrentIndex(colorKey);
}


void GraphSettingsForm::setStyleSame( bool same )
{
    spinStyleSetSelect->setDisabled(same);
    if(same)
    {
	gs.setAllSetsSameAs(spinStyleSetSelect->value() - 1);
    }
    else
    {
	gs.setAllSetsSameAs(-1);
	showSetSettings(spinStyleSetSelect->value());	//DEBUG
    }
}


void GraphSettingsForm::applySetChanges( int setNo )
{
    int ii = (uint) setNo - 1;
//    if(!gs.dataSetExists(ii))
//	QMessageBox::warning(this, EBTCAPTION + "DEBUG",
//			     QString("applySetChanges op niet-bestaande set %1").arg(ii));
    if(gs.dataSetExists(ii))
    {
	gs.setDataSetXCol(ii, uint(spinXColumn->value()));
	gs.setDataSetYCol(ii, uint(spinYColumn->value()));
	if(comboValueType->currentIndex() == computedValues)
	    gs.setDataSetDataType(ii, COMPUTEDVALUES);
	else if(comboValueType->currentIndex() == fixedValues)
	    gs.setDataSetDataType(ii, FIXEDVALUES);
	else
	    gs.setDataSetDataType(ii, OUTPUTVALUES);
	gs.setDataSetPopulationNr(ii, spinPopSelect->value());
	gs.setDataSetLegendLabel(ii, editLegendLabel->text());
	if(comboXAxisPrimary->currentIndex() == 0)
	    gs.setDataSetXAxis(ii, QwtPlot::xTop);
	else
	    gs.setDataSetXAxis(ii, QwtPlot::xBottom);
	if(comboYAxisPrimary->currentIndex() == 0)
	    gs.setDataSetYAxis(ii, QwtPlot::yRight);
	else
	    gs.setDataSetYAxis(ii, QwtPlot::yLeft);
	
	int jj;
	if(gs.allSetsSameAs() < 0)
	    jj = ii;
	else 
	    jj = gs.allSetsSameAs();
	
	gs.setDataSetCurveStyle(jj, (QwtPlotCurve::CurveStyle)comboCurveStyleIndex[comboCurveStyle->currentIndex()]);
	gs.setDataSetHistNumBars(jj, spinHistNumBars->value());
	
	QColor symbolColor = comboSymbolColorIndex[comboSymbolColor->currentIndex()];
	setSymbol->setStyle(comboSymbolIndex[comboSymbolStyle->currentIndex()]);
	setSymbol->setSize(spinSymbolSize->value());
	setSymbol->setBrush(QBrush(symbolColor,
				  Qt::BrushStyle(comboSymbolFill->currentIndex())));
	setSymbol->setPen(QPen(symbolColor,
			      (uint)spinSymbolPenWidth->value(), 
			      Qt::SolidLine));
	gs.setDataSetCurveSymbol(jj, setSymbol);

	setLinePen.setStyle(Qt::PenStyle(comboLineStyle->currentIndex()));
	setLinePen.setWidth(spinLineWidth->value());
	setLinePen.setColor(comboLineColorIndex[comboLineColor->currentIndex()]);
	gs.setDataSetLinePen(jj, setLinePen);
    }
}


void GraphSettingsForm::newTitleFont()
{
    bool ok;
    QFont font = QFontDialog::getFont(&ok, titleFont , this );
    if ( ok)
	titleFont = font;
    else
	return;
}


void GraphSettingsForm::newLegendFont()
{
    bool ok;
    QFont font = QFontDialog::getFont(&ok, legendFont , this );
    if ( ok)
	legendFont = font;
    else
	return;
}


void GraphSettingsForm::newAxisTitleFont()
{
    bool ok;
    QFont font = QFontDialog::getFont(&ok, axisTitleFont , this );
    if ( ok)
	axisTitleFont = font;
    else
	return;
}


void GraphSettingsForm::closeEvent(QCloseEvent *)
{
  switch (QMessageBox::question(this, EBTCAPTION + "Apply changes", "Apply any changed settings?",
                                QMessageBox::Yes | QMessageBox::No | QMessageBox::Cancel, QMessageBox::Cancel))
    {
      case QMessageBox::Yes:
        acceptChanges();
        break;
      case QMessageBox::No:
        reject();
        break;
      case QMessageBox::Cancel:
      default:
        return;
    }
}


void GraphSettingsForm::applyAxisChanges(int ii)
{
    QwtPlot::Axis a = comboAxisIndex[ii];
    if(gs.axisExists(a))
    {
	gs.setAxisAutoscale(a, checkAxisAutoscale->isChecked());
	gs.setAxisLogarithmic(a, radioAxisLog->isChecked());
	gs.setAxisTitle(a, editAxisLabel->text());
	gs.setAxisTitleFont(a, axisTitleFont);
	if(!checkAxisAutoscale->isChecked())
	{
	    if(editAxisMinValue->text().toDouble() > editAxisMaxValue->text().toDouble())
	    {
		QString tmpString = editAxisMinValue->text(); 
		editAxisMinValue->setText(editAxisMaxValue->text());
		editAxisMaxValue->setText(tmpString);
		QMessageBox::warning(this, EBTCAPTION + "invalid values",
				     QString("The minimum value of an axis can't be\ngreater than its maximum value!\nThey will be exchanged"));
	    }
	    if(gs.axisLogarithmic(a) && 
	       ((editAxisMinValue->text().toDouble() < MIN_LOG_VAL) || (editAxisMaxValue->text().toDouble() <= MIN_LOG_VAL)))
	    {
		if(editAxisMinValue->text().toDouble() < MIN_LOG_VAL)
		    editAxisMinValue->setText(formatDouble(MIN_LOG_VAL));
		if(editAxisMaxValue->text().toDouble() <= MIN_LOG_VAL)
		    editAxisMaxValue->setText(formatDouble(10 * MIN_LOG_VAL));
		QMessageBox::warning(this, EBTCAPTION + "invalid values",
				     QString("The minimum and maximum value of\n an axis with a logarithmic scale \nshould be >= %1").arg(MIN_LOG_VAL));
		
	    }
	    gs.setAxisMinVal(a, editAxisMinValue->text().toDouble());
	    gs.setAxisMaxVal(a, editAxisMaxValue->text().toDouble());
	} 
    } 
}

void GraphSettingsForm::setAxesSame( bool same )
{
    comboAxisSelect->setDisabled(same);
    if(same)
    {
	for(int ii = 0; ii < 4; ii++)
	    applyAxisChanges(ii);
    }
}


void GraphSettingsForm::applyGeneralChanges()
{
  gs.setShowFineGrid(radioGridFine->isChecked());
  gs.setShowGrid((radioGridFine->isChecked() || radioGridCoarse->isChecked()));

  gs.setPlotTitle(editTitle->text());
  gs.setPlotTitleFont(titleFont);

  gs.setShowLegend(checkLegendShow->isChecked());
  if (checkLegendShow->isChecked())
    {
      gs.setLegendFramed(comboFrameType->currentText() == "Framed");
      if (comboFrameType->currentText() == "Framed")
	gs.setLegendFrameWidth(valueLegendFrameLine->value());
      if (comboLegendPosition->currentText() == "Bottom")
	gs.setLegendPosition(QwtPlot::BottomLegend);
      else if (comboLegendPosition->currentText() == "Right")
	gs.setLegendPosition(QwtPlot::RightLegend);
      else if (comboLegendPosition->currentText() == "Top")
	gs.setLegendPosition(QwtPlot::TopLegend);
      else // Floating
	{
	  gs.setLegendPosition(QwtPlot::TopLegend + 1);
	  gs.setLegendXPosition(valueLegendX->value());
	  gs.setLegendYPosition(valueLegendY->value());
	}
    }
  gs.setShowTimestamp(checkTimestampShow->isChecked());
  gs.setShowDescription(checkDescrShow->isChecked());
  gs.setDefaultDescription(checkDescrDefault->isChecked());
  gs.setPlotDescription(editDescr->text());
  gs.setLegendFont(legendFont);
}


void GraphSettingsForm::setDescrDefault( bool def )
{
    if(def)
    {
	if(outGraph)
	    editDescr->setText(OUTPUTDESC);
	else
	    editDescr->setText(STATEDESC);
    }
    else
	editDescr->setText(gs.plotDescription());
    
    editDescr->setEnabled(!def && checkDescrShow->isChecked());
}


void GraphSettingsForm::setDescrShow( bool show )
{
    checkDescrDefault->setEnabled(show);
    setDescrDefault(checkDescrDefault->isChecked());	
}


void GraphSettingsForm::setLegendShow( bool show )
{
  groupLegendBox->setEnabled(show);
  labelLegendFrameLine->setEnabled(comboFrameType->currentText() == "Framed");
  valueLegendFrameLine->setEnabled(comboFrameType->currentText() == "Framed");
  labelLegendX->setEnabled(comboLegendPosition->currentText() == "Floating");
  valueLegendX->setEnabled(comboLegendPosition->currentText() == "Floating");
  labelLegendY->setEnabled(comboLegendPosition->currentText() == "Floating");
  valueLegendY->setEnabled(comboLegendPosition->currentText() == "Floating");
}

void GraphSettingsForm::comboFrameTypeToggled( )
{
  labelLegendFrameLine->setEnabled(comboFrameType->currentText() == "Framed");
  valueLegendFrameLine->setEnabled(comboFrameType->currentText() == "Framed");
}

void GraphSettingsForm::comboLegendPositionToggled( )
{
  labelLegendX->setEnabled(comboLegendPosition->currentText() == "Floating");
  valueLegendX->setEnabled(comboLegendPosition->currentText() == "Floating");
  labelLegendY->setEnabled(comboLegendPosition->currentText() == "Floating");
  valueLegendY->setEnabled(comboLegendPosition->currentText() == "Floating");
}



void GraphSettingsForm::setAutoscale( bool autoscale )
{
    editAxisMinValue->setDisabled(autoscale);
    editAxisMaxValue->setDisabled(autoscale);
    
    if(autoscale)
    {
	editAxisMinValue->setText(QString());
	editAxisMaxValue->setText(QString());
    }
    else
    {
	QwtPlot::Axis a = comboAxisIndex[comboAxisSelect->currentIndex()];
	QString valueText;
	editAxisMinValue->setText(valueText.setNum(gs.axisMinVal(a)));
	editAxisMaxValue->setText(valueText.setNum(gs.axisMaxVal(a)));
    }
}

void GraphSettingsForm::setCurveStyle(int curveStyle)
{
    if(comboCurveStyleIndex[curveStyle] == QwtPlotCurve::UserCurve)
    {
	uint setNr = spinSetSelect->value() - 1;
	spinHistNumBars->setValue(gs.dataSetHistNumBars(setNr));
	spinHistNumBars->setHidden(FALSE);
	labelHistNumBars->setHidden(FALSE);
	
	// change groupSymbol so that the symbol fill can be used to customize histogram bars
	groupSymbol->setTitle("Bar fill");
	comboSymbolStyle->setHidden(TRUE);
	labelSymbolSize->setHidden(TRUE);
	spinSymbolSize->setHidden(TRUE);
	labelSymbolPenWidth->setHidden(TRUE);
	spinSymbolPenWidth->setHidden(TRUE);
	comboSymbolColor->setDisabled(FALSE);
	buttonSymbolColor->setDisabled(FALSE);
	labelSymbolFill->setText("Pattern:");
	comboSymbolFill->setDisabled(FALSE);
    }
    else
    {
	spinHistNumBars->setHidden(TRUE);
	labelHistNumBars->setHidden(TRUE);
	// change groupSymbol back (some of which has been done before this function call)
	groupSymbol->setTitle("Symbol");
	comboSymbolStyle->setHidden(FALSE);
	labelSymbolSize->setHidden(FALSE);
	spinSymbolSize->setHidden(FALSE);
	labelSymbolPenWidth->setHidden(FALSE);
	spinSymbolPenWidth->setHidden(FALSE);
	labelSymbolFill->setText("Fill:");
    }
}

void GraphSettingsForm::fillSpinBoxes(uint lastSet)
{
    uint firstSet;
    if(lastSet == 0) 
    {
	firstSet = 0;
	spinSetSelect->setSpecialValueText("None");
	spinStyleSetSelect->setSpecialValueText("None");
    }
    else 
    {
	firstSet = 1;
	spinSetSelect->setSpecialValueText(QString());
	spinStyleSetSelect->setSpecialValueText(QString());
    }
    spinSetSelect->setRange(firstSet, lastSet);    
    spinStyleSetSelect->setRange(firstSet, lastSet);    
    labelSetCount->setNum(int(lastSet));
}


void GraphSettingsForm::enableGroupSymbol(int selectedStyle)
{
    if(groupSymbol->isEnabled())
    {
	spinSymbolSize->setDisabled(selectedStyle == QwtSymbol::NoSymbol);
	comboSymbolColor->setDisabled(selectedStyle == QwtSymbol::NoSymbol);
	buttonSymbolColor->setDisabled(selectedStyle == QwtSymbol::NoSymbol);
	spinSymbolPenWidth->setDisabled(selectedStyle == QwtSymbol::NoSymbol);
	comboSymbolFill->setDisabled(selectedStyle == QwtSymbol::NoSymbol);
    }
}


void GraphSettingsForm::enableGroupLine(int selectedPen)
{
    if(groupLine->isEnabled())
    {
	spinLineWidth->setDisabled(selectedPen == Qt::NoPen);
	comboLineColor->setDisabled(selectedPen == Qt::NoPen);
	buttonLineColor->setDisabled(selectedPen == Qt::NoPen);
    }
}
