#include "texteditor.h"

#include <qvariant.h>
#include <qimage.h>
#include <qpixmap.h>

#include "qfile.h"
#include "qstatusbar.h"
#include "qfileinfo.h"
#include "qclipboard.h"
#include "qmessagebox.h"
#include "qprinter.h"
#include "qpainter.h"

#include <QPrintDialog>

/*
 *  Constructs a TextEditor as a child of 'parent', with the
 *  name 'name' and widget flags set to 'f'.
 *
 */
TextEditor::TextEditor(QWidget* parent, Qt::WindowFlags fl)
    : QMainWindow(parent, fl)
{
  hide();
  setupUi(this);

#if defined (Q_OS_MACOS)
  setUnifiedTitleAndToolBarOnMac (TRUE);
#endif
  (void)statusBar();
  init();
}


/*
 *  Destroys the object and frees any allocated resources
 */
TextEditor::~TextEditor()
{
  // no need to delete child widgets, Qt does it all for us
}


/*
 *  Sets the strings of the subwidgets using the current
 *  language.
 */
void TextEditor::languageChange()
{
    retranslateUi(this);
}


void TextEditor::fileSave()
{
    currentFile->buffer = textField->toPlainText();			// Qt4
    if (currentFile->save())
	statusBar()->showMessage(QString("Saved file to '%1'").
			     arg(currentFile->fileName()), 3000);
    else
      QMessageBox::critical(this, "Warning", QString("Failed to save '%1'").arg(currentFile->fileName()));
    updateView();
}


void TextEditor::fileSaveAs()
{
    currentFile->buffer = textField->toPlainText();			// Qt4
    if (currentFile->saveAs())
	statusBar()->showMessage(QString("Saved file to '%1'").
			     arg(currentFile->fileName()), 3000);
    else
      QMessageBox::critical(this, "Warning", QString("Failed to save '%1'").arg(currentFile->fileName()));
    updateView();
}


void TextEditor::filePrint()
{
#ifndef QT_NO_PRINTER
  QPrinter printer( QPrinter::HighResolution );
  printer.setFullPage(TRUE);

  QPrintDialog *dlg = new QPrintDialog(&printer, this);
  if (dlg->exec() != QDialog::Accepted) return;

  QTextDocument *document = textField->document();
  document->print(&printer);

  delete dlg;
#endif
}


void TextEditor::closeEvent(QCloseEvent* )
{
  fileExit();
}

bool TextEditor::okToClear()
{
  if (currentFile->isChanged())
    {
      QString msg;
      if (currentFile->fileName().isEmpty())
        msg = "Unnamed " + currentFile->descriptionText() + "\n";
      else
        msg = QString("File '%1'\n").arg(currentFile->fileName());
      msg += QString("has been changed. \nDo you want to save the changes?");
      QMessageBox::StandardButton ans = QMessageBox::question(this, EBTCAPTION + "Unsaved Changes", msg, QMessageBox::Save | QMessageBox::Cancel | QMessageBox::Discard,
                                      QMessageBox::Save);
      if (ans == QMessageBox::Save)
        fileSave();
      else if (ans == QMessageBox::Cancel)
        return FALSE;
      else
        currentFile->setChanged(FALSE);
    }
  return TRUE;
}

void TextEditor::fileExit()
{
  if(okToClear())
    {
      currentFile->setEditor(0);
      hide();
      emit closing(this);
    }
}



void TextEditor::editFind()
{
    SearchDialog dialog( this, TRUE );
    connect( &dialog, SIGNAL( searchString( const QString & ) ),
	     this, SLOT( searchString( const QString & ) ), Qt::UniqueConnection);
    dialog.setSearchString(sString);
    dialog.exec();
}

void TextEditor::searchString( const QString &expr )
{    
    sString = expr;
    textField->find(expr);
}


void TextEditor::helpIndex()
{
    
}


void TextEditor::helpContents()
{
    
}


void TextEditor::helpAbout()
{
    
}


void TextEditor::setFile(EBTFile& ef, bool vOnly)
{
    currentFile = &ef;
    statusBar()->showMessage(QString("Loading '%1'...").arg(currentFile->fileName()));
    currentFile->load();
    textField->setText(currentFile->buffer);
    currentFile->setChanged(FALSE);
    setViewOnly(vOnly);
    statusBar()->showMessage("Ready");
    updateView();
}


void TextEditor::updateView()
{
    setWindowTitle(currentFile->captionText());
}


void TextEditor::textChanged()
{
    currentFile->setChanged(TRUE);
}


bool TextEditor::parentClose()
{
    if(okToClear())
    {
	currentFile->setEditor(0);
	hide();
	emit closing(this);
	return TRUE;
    }
    else
	return FALSE;
}


void TextEditor::enableEdit( bool enable )
{
    textField->setReadOnly(!enable);
    fileSaveAction->setEnabled(enable);
    fileSaveAsAction->setEnabled(enable);
    editUndoAction->setEnabled(enable);
    editRedoAction->setEnabled(enable);
    editCutAction->setEnabled(enable);
    editPasteAction->setEnabled(enable);
}


bool TextEditor::isViewOnly()
{
    return viewOnly;
}


void TextEditor::setViewOnly(bool vOnly)
{
    viewOnly = vOnly;
    enableEdit(!viewOnly);
    if(!viewOnly)
	textField->viewport()->setFocus();
}


void TextEditor::init()
{
    connect(this, SIGNAL(closing(TextEditor*)), parentWidget(), SLOT(editorCloses(TextEditor*)));

    // From here on I do not need the parent-child relationship any longer
    setParent(NULL);

    QFont widgetFont("Courier New");
    widgetFont.setStyleHint(QFont::TypeWriter);
    widgetFont.setFixedPitch(TRUE);
    widgetFont.setKerning(FALSE);
#if defined (Q_OS_MACOS)
    widgetFont.setWeight(QFont::DemiBold);
    widgetFont.setPointSize(12);
#else
    widgetFont.setWeight(QFont::Normal);
    widgetFont.setPointSize(10);
#endif
    textField->setFont(widgetFont);
}


QString TextEditor::fileName()
{
    return currentFile->fileName();
}
