/***
   NAME
     ebtrktsi5
   DESCRIPTION
     This file contains the specification of an explicit Runge-Kutta integration 
     solver for systems of ODEs as presented in

     Ch. Tsitouras, 2011. Runge–Kutta pairs of order 5(4) satisfying only the first 
     column simplifying assumption, Computers & Mathematics with Applications 62(2):
     770-775, https://doi.org/10.1016/j.camwa.2011.06.002.
   
     The derivatives f_i(x_i, y_i) are evaluated at the intermediate points
   
             x_i = x_n + c_i * h
             y_i = y_n + sum_{j=1}^{j=i-1} f_j(x_j, y_j) 
   
     the final estimate of the solution y is given by
     
             y_{n+1}    = y_n + sum_{i=1}^{i=s} b_i f_i(x_i, y_i)
   
     and the lower order estimate of the solution y by
     
             yhat_{n+1} = y_n + sum_{i=1}^{i=s} bhat_i f_i(x_i, y_i)

   Last modification: AMdR - May 24, 2023
***/
#ifndef EBTRKTSI5
#define EBTRKTSI5
#endif

/*
 *==================================================================================
 *
 * Defining all constants that are local to this specific file.
 * 
 * DESCRIPTION OF INTEGRATION CONSTANTS
 * 
 * BETA        The "beta" for stabilized step size control (see section IV.2 of the
 *             book). Larger values for beta ( <= 0.1 ) make the step size control 
 *             more stable. rktsi5 needs a larger beta than Higham & Hall. Negative 
 *             initial value provoke beta=0; default beta=0.04.
 * 
 * FAC1     Parameters for step size selection; the new step size is chosen
 * FAC2     subject to the restriction  fac1 <= hnew/hold <= fac2.
 *             Default values are fac1=0.2 and fac2=10.0.
 * 
 * SAFETY   Safety factor in the step size prediction, default 0.9.
 * 
 *==================================================================================
 */

#define BETA                      0.1
#define FAC1                      0.2
#define FAC2                      10.0
#define FACC1                     5.0                                               // 1.0 / FAC1
#define FACC2                     0.1                                               // 1.0 / FAC2
#define SAFETY                    0.9
#define ABS_ERR                   1.0E-13
#define FACOLD                    1.0E-4

/*
 *==================================================================================
 * The error messages that occur in the routines in the present file.
 *==================================================================================
 */

#define MAFO                      "Memory allocation failure in ODE integration routine!"
#define REC                       "Too many recursions in integration to find suitable stepsize!"
#define SSS                       "Step size in integration routine too small!"
#define ZBB                       "Root is not bracketed in ZBRENT!"
#define ZBM                       "Maximum number of iterations exceeded in ZBRENT!"

/*
 *==================================================================================
 * Definitions of static variables, restricted to this file.
 *==================================================================================
 */

static int                        step_failed     = 0;
static int                        recur_no;
static long                       accepted_steps  = 0L;
static long                       ODEAllocated    = 0L;
static long                       SystemSize;
static double                     *y      = NULL;
static double                     *yy1    = NULL;
static double                     *yco    = NULL;
static double                     *yerr   = NULL;
static double                     *k1     = NULL;
static double                     *k2     = NULL;
static double                     *k3     = NULL;
static double                     *k4     = NULL;
static double                     *k5     = NULL;
static double                     *k6     = NULL;
static double                     *k7     = NULL;
#if (POPULATION_NR > 0)
static int                        table_size[POPULATION_NR];
static population                 u_pop[POPULATION_NR];
static population                 c_pop[POPULATION_NR];
static population                 u_ofs[POPULATION_NR];
static population                 c_ofs[POPULATION_NR];
static population                 u_popgrad1[POPULATION_NR];
static population                 u_popgrad2[POPULATION_NR];
static population                 u_popgrad3[POPULATION_NR];
static population                 u_popgrad4[POPULATION_NR];
static population                 u_popgrad5[POPULATION_NR];
static population                 u_popgrad6[POPULATION_NR];
static population                 u_popgrad7[POPULATION_NR];
static population                 u_ofsgrad1[POPULATION_NR];
static population                 u_ofsgrad2[POPULATION_NR];
static population                 u_ofsgrad3[POPULATION_NR];
static population                 u_ofsgrad4[POPULATION_NR];
static population                 u_ofsgrad5[POPULATION_NR];
static population                 u_ofsgrad6[POPULATION_NR];
static population                 u_ofsgrad7[POPULATION_NR];
#else
static population                 *bpoints = NULL;
static population                 *u_pop   = NULL;
static population                 *u_ofs   = NULL;
#if (EVENT_NR > 0)
static population                 *c_pop = NULL;
static population                 *c_ofs = NULL;
#endif
static population                 *u_popgrad1 = NULL;
static population                 *u_popgrad2 = NULL;
static population                 *u_popgrad3 = NULL;
static population                 *u_popgrad4 = NULL;
static population                 *u_popgrad5 = NULL;
static population                 *u_popgrad6 = NULL;
static population                 *u_popgrad7 = NULL;
static population                 *u_ofsgrad1 = NULL;
static population                 *u_ofsgrad2 = NULL;
static population                 *u_ofsgrad3 = NULL;
static population                 *u_ofsgrad4 = NULL;
static population                 *u_ofsgrad5 = NULL;
static population                 *u_ofsgrad6 = NULL;
static population                 *u_ofsgrad7 = NULL;
#endif // (POPULATION_NR > 0)

static double                     facold = FACOLD;
#if (EVENT_NR > 0)
static double                     oldELvalue[EVENT_NR] = {0.0};
static double                     newELvalue[EVENT_NR] = {0.0};
static int                        located[EVENT_NR]    = {0};
#endif


/*
 *==================================================================================
 *
 * Start of function implementations.
 *==================================================================================
 */

void PrepareCycle()

/* 
 * PrepareCycle - This routine is called at the beginning of a cohort cycle, when the 
 *                size of the system of ODEs is fixed and will not change until the 
 *                end of the cohort cycle has been reached. The different memory 
 *                copies of the data to be integrated and the pointers into the data 
 *                heap can hence safely be set up.
 */

{
  SystemSize = ENVIRON_DIM;
#if (POPULATION_NR > 0)
  register int                    ii;

  for(ii = 0; ii < POPULATION_NR; ii++)
    {
      table_size[ii] = CohortNo[ii]   + BpointNo[ii];
      SystemSize    += table_size[ii] * COHORT_SIZE;
    }
#endif // (POPULATION_NR > 0)

  if (!(SystemSize < ODEAllocated))
    {
      ODEAllocated = MemBlocks(SystemSize);
      y       = (double *)Myalloc((void *)y,      (size_t)ODEAllocated, sizeof(double));
      yy1     = (double *)Myalloc((void *)yy1,    (size_t)ODEAllocated, sizeof(double));
      yco     = (double *)Myalloc((void *)yco,    (size_t)ODEAllocated, sizeof(double));
      yerr    = (double *)Myalloc((void *)yerr,   (size_t)ODEAllocated, sizeof(double));
      k1      = (double *)Myalloc((void *)k1,     (size_t)ODEAllocated, sizeof(double));
      k2      = (double *)Myalloc((void *)k2,     (size_t)ODEAllocated, sizeof(double));
      k3      = (double *)Myalloc((void *)k3,     (size_t)ODEAllocated, sizeof(double));
      k4      = (double *)Myalloc((void *)k4,     (size_t)ODEAllocated, sizeof(double));
      k5      = (double *)Myalloc((void *)k5,     (size_t)ODEAllocated, sizeof(double));
      k6      = (double *)Myalloc((void *)k6,     (size_t)ODEAllocated, sizeof(double));
      k7      = (double *)Myalloc((void *)k7,     (size_t)ODEAllocated, sizeof(double));
      if(!(y && yy1 && yco && yerr && k1 && k2 && k3 && k4 && k5 && k6 && k7)) ErrorAbort(MAFO);
    }

  // Copy environment variables
  (void)memcpy((DEF_TYPE *)y, (DEF_TYPE *)env, ENVIRON_DIM*sizeof(double));

#if (POPULATION_NR > 0)
  int                             len;

  len = ENVIRON_DIM;
  for(ii = 0; ii < POPULATION_NR; ii++)
    {
      // Copy all populations
      (void)memcpy((DEF_TYPE *)(y + len), (DEF_TYPE *)pop[ii], (table_size[ii] * COHORT_SIZE) * sizeof(double));
      u_pop[ii]       = (population)(yy1 + len);
      u_ofs[ii]       = (population)(yy1 + len + CohortNo[ii] * COHORT_SIZE);
      c_pop[ii]       = (population)(yco + len);
      c_ofs[ii]       = (population)(yco + len + CohortNo[ii] * COHORT_SIZE);
      u_popgrad1[ii]  = (population)(k1  + len);
      u_ofsgrad1[ii]  = (population)(k1  + len + CohortNo[ii] * COHORT_SIZE);
      u_popgrad2[ii]  = (population)(k2  + len);
      u_ofsgrad2[ii]  = (population)(k2  + len + CohortNo[ii] * COHORT_SIZE);
      u_popgrad3[ii]  = (population)(k3  + len);
      u_ofsgrad3[ii]  = (population)(k3  + len + CohortNo[ii] * COHORT_SIZE);
      u_popgrad4[ii]  = (population)(k4  + len);
      u_ofsgrad4[ii]  = (population)(k4  + len + CohortNo[ii] * COHORT_SIZE);
      u_popgrad5[ii]  = (population)(k5  + len);
      u_ofsgrad5[ii]  = (population)(k5  + len + CohortNo[ii] * COHORT_SIZE);
      u_popgrad6[ii]  = (population)(k6  + len);
      u_ofsgrad6[ii]  = (population)(k6  + len + CohortNo[ii] * COHORT_SIZE);
      u_popgrad7[ii]  = (population)(k7  + len);
      u_ofsgrad7[ii]  = (population)(k7  + len + CohortNo[ii] * COHORT_SIZE);
      len            += (table_size[ii] * COHORT_SIZE);
    }
#endif // (POPULATION_NR > 0)

  return;
}

/*
 *==================================================================================
 * The Butcher tableau constants
 *==================================================================================
 */

/*
 * The original constants presented in the paper:
 *          Ch. Tsitouras, 2011. Runge–Kutta pairs of order 5(4) satisfying only the first 
 *          column simplifying assumption, Computers & Mathematics with Applications 62(2):
 *          770-775, https://doi.org/10.1016/j.camwa.2011.06.002.
 * 
 * are:
 *
 *  static const double a21 =   0.161,
 *                      a31 =   0.327              - (  0.3354806554923570),
 *                      a41 =   0.9                - ( -6.359448489975075  + 4.362295432869581),
 *                      a51 =   0.9800255409045097 - (-11.74888356406283   + 7.495539342889836  + -0.09249506636175525),
 *                      a61 =   1.0                - (-12.92096931784711   + 8.159367898576159  + -0.071584973281401006 + -0.02826905039406838),
 *                      a71 =   1.0                - (  0.01               + 0.4798896504144996 +  1.379008574103742    + -3.290069515436081   + 2.324710524099774), 
 *                      a32 =   0.3354806554923570, 
 *                      a42 =  -6.359448489975075, a43 = 4.362295432869581,
 *                      a52 = -11.74888356406283,  a53 = 7.495539342889836,  a54 = -0.09249506636175525,
 *                      a62 = -12.92096931784711,  a63 = 8.159367898576159,  a64 = -0.071584973281401006, a65 = -0.02826905039406838,
 *                      a72 =   0.01,              a73 = 0.4798896504144996, a74 =  1.379008574103742,    a75 = -3.290069515436081,   a76 = 2.324710524099774;
 * 
 *  static const double b1 = 0.09646076681806523, b2 = 0.01,               b3 = 0.4798896504144996,
 *                      b4 = 1.379008574103742,   b5 = -3.290069515436081, b6 = 2.324710524099774,  b7 = 0;
 * 
 *  static const double bhat1 = 0.001780011052226, bhat2 =  0.000816434459657,  bhat3 = -0.007880878010262,
 *                      bhat4 = 0.144711007173263, bhat5 = -0.582357165452555,  bhat6 =  0.458082105929187, bhat7 = 1.0 / 66.0;
 *
 *
 * However, these constants seems to fail in particular because of the values of bhat.
 * 
 * The following coefficients were obtained from the website of Peter Stone http://www.peterstone.name/Maplepgs/RKcoeff.html
 * (see the file Maple/RKcoeff5n_1.pdf)
 * 
 * The coefficients are the same as used in the Julia package DifferentialEquations.jl as can be inferred from the 
 * StackExchange thread 
 * 
 * https://scicomp.stackexchange.com/questions/30276/how-well-do-explicit-runge-kutta-tableau-methods-compare-to-the-state-of-the-a
 * 
 * in which there is a link to the file 
 * 
 * https://github.com/SciML/DiffEqDevTools.jl/blob/master/src/ode_tableaus.jl
 * 
 * in which the Butcherr tableau for the Tsitouras method is generated in the function constructTsitouras5
 * 
 */
  // static const double c2    = 0.161;
  // static const double c3    = 0.327;
  // static const double c4    = 0.9;
  // static const double c5    = 0.9800255409045096857298102862870245954942137979563024768854764293221195950761080302604;
  // static const double c6    = 1.;
  // static const double c7    = 1.;
    
  static const double a21   =   0.161;
  static const double a31   =  -0.008480655492356988544426874250230774675121177393430391537369234245294192976164141156943;
  static const double a32   =   0.3354806554923569885444268742502307746751211773934303915373692342452941929761641411569;
  static const double a41   =   2.897153057105493432130432594192938764924887287701866490314866693455023795137503079289;
  static const double a42   =  -6.359448489975074843148159912383825625952700647415626703305928850207288721235210244366;
  static const double a43   =   4.362295432869581411017727318190886861027813359713760212991062156752264926097707165077;
  static const double a51   =   5.325864828439256604428877920840511317836476253097040101202360397727981648835607691791;
  static const double a52   = -11.74888356406282787774717033978577296188744178259862899288666928009020615663593781589;
  static const double a53   =   7.495539342889836208304604784564358155658679161518186721010132816213648793440552049753;
  static const double a54   =  -0.09249506636175524925650207933207191611349983406029535244034750452930469056411389539635;
  static const double a61   =   5.861455442946420028659251486982647890394337666164814434818157239052507339770711679748;
  static const double a62   = -12.92096931784710929170611868178335939541780751955743459166312250439928519268343184452;
  static const double a63   =   8.159367898576158643180400794539253485181918321135053305748355423955009222648673734986;
  static const double a64   =  -0.07158497328140099722453054252582973869127213147363544882721139659546372402303777878835;
  static const double a65   =  -0.02826905039406838290900305721271224146717633626879770007617876201276764571291579142206;
  static const double a71   =   0.09646076681806522951816731316512876333711995238157997181903319145764851595234062815396;
  static const double a72   =   0.01;
  static const double a73   =   0.4798896504144995747752495322905965199130404621990332488332634944254542060153074523509;
  static const double a74   =   1.379008574103741893192274821856872770756462643091360525934940067397245698027561293331;
  static const double a75   =  -3.290069515436080679901047585711363850115683290894936158531296799594813811049925401677;
  static const double a76   =   2.324710524099773982415355918398765796109060233222962411944060046314465391054716027841;

  // static const double b1    =   0.09646076681806522951816731316512876333711995238157997181903319145764851595234062815396;
  // static const double b2    =   0.01;
  // static const double b3    =   0.4798896504144995747752495322905965199130404621990332488332634944254542060153074523509;
  // static const double b4    =   1.379008574103741893192274821856872770756462643091360525934940067397245698027561293331;
  // static const double b5    =  -3.290069515436080679901047585711363850115683290894936158531296799594813811049925401677;
  // static const double b6    =   2.324710524099773982415355918398765796109060233222962411944060046314465391054716027841;

  // static const double bhat1 =   0.09468075576583945807478876255758922856117527357724631226139574065785592789071067303271;
  // static const double bhat2 =   0.009183565540343253096776363936645313759813746240984095238905939532922955247253608687270;
  // static const double bhat3 =   0.4877705284247615707855642599631228241516691959761363774365216240304071651579571959813;
  // static const double bhat4 =   1.234297566930478985655109673884237654035539930748192848315425833500484878378061439761;
  // static const double bhat5 =  -2.707712349983525454881109975059321670689605166938197378763992255714444407154902012702;
  // static const double bhat6 =   1.866628418170587035753719399566211498666255505244122593996591602841258328965767580089;
  // static const double bhat7 =   0.01515151515151515151515151515151515151515151515151515151515151515151515151515151515152;

  static const double db1   =  ( 0.09646076681806522951816731316512876333711995238157997181903319145764851595234062815396 -
                                 0.09468075576583945807478876255758922856117527357724631226139574065785592789071067303271);
  static const double db2   =  ( 0.01 -
                                 0.009183565540343253096776363936645313759813746240984095238905939532922955247253608687270);
  static const double db3   =  ( 0.4798896504144995747752495322905965199130404621990332488332634944254542060153074523509 -
                                 0.4877705284247615707855642599631228241516691959761363774365216240304071651579571959813);
  static const double db4   =  ( 1.379008574103741893192274821856872770756462643091360525934940067397245698027561293331 -
                                 1.234297566930478985655109673884237654035539930748192848315425833500484878378061439761);
  static const double db5   =  (-3.290069515436080679901047585711363850115683290894936158531296799594813811049925401677 -
                                -2.707712349983525454881109975059321670689605166938197378763992255714444407154902012702);
  static const double db6   =  ( 2.324710524099773982415355918398765796109060233222962411944060046314465391054716027841 -
                                 1.866628418170587035753719399566211498666255505244122593996591602841258328965767580089);
  static const double db7   =   -0.01515151515151515151515151515151515151515151515151515151515151515151515151515151515152;

static void rktsi5(double dt)

/* 
 * rktsi5 - Routine performs an integration step of the system specified in the
 *          function (*deriv)()using the Tsitouras 5(4) integration method
 */

{
  register int                    ii;

  initState      = y;
  currentState   = yy1;
  currentDers[1] = k1;
  currentDers[2] = k2;
  currentDers[3] = k3;
  currentDers[4] = k4;
  currentDers[5] = k5;
  currentDers[6] = k6;
  currentDers[7] = k7;

  // Set all time derivatives
  k1[0] = k2[0] = k3[0] = k4[0] = k5[0] = k6[0] = k7[0] = 1.0;

  // Don't do first derivative if already present
  if (!recur_no)
    {
      (void)memcpy((DEF_TYPE *)yy1, (DEF_TYPE *)y, SystemSize * sizeof(double));
      rk_level = 1;
      Gradient(yy1, u_pop, u_ofs, k1, u_popgrad1, u_ofsgrad1, bpoints);
#if (EVENT_NR > 0)
      for (ii = 0; ii < EVENT_NR; ii++) oldELvalue[ii] = NO_EVENT;
      EventLocation(yy1, u_pop, u_ofs, bpoints, oldELvalue);
#endif
    }

  for (ii = 0; ii < SystemSize; ii++) yy1[ii] = y[ii] + dt * a21 * k1[ii];

  rk_level = 2;
  Gradient(yy1, u_pop, u_ofs, k2, u_popgrad2, u_ofsgrad2, bpoints);

  for (ii = 0; ii < SystemSize; ii++) yy1[ii] = y[ii] + dt * (a31 * k1[ii] + a32 * k2[ii]);

  rk_level = 3;
  Gradient(yy1, u_pop, u_ofs, k3, u_popgrad3, u_ofsgrad3, bpoints);

  for (ii = 0; ii < SystemSize; ii++) yy1[ii] = y[ii] + dt * (a41 * k1[ii] + a42 * k2[ii] + a43 * k3[ii]);

  rk_level = 4;
  Gradient(yy1, u_pop, u_ofs, k4, u_popgrad4, u_ofsgrad4, bpoints);

  for (ii = 0; ii < SystemSize; ii++) yy1[ii] = y[ii] + dt * (a51 * k1[ii] + a52 * k2[ii] + a53 * k3[ii] + a54 * k4[ii]);

  rk_level = 5;
  Gradient(yy1, u_pop, u_ofs, k5, u_popgrad5, u_ofsgrad5, bpoints);

  for (ii = 0; ii < SystemSize; ii++) yy1[ii] = y[ii] + dt * (a61 * k1[ii] + a62 * k2[ii] + a63 * k3[ii] + a64 * k4[ii] + a65 * k5[ii]);

  rk_level = 6;
  Gradient(yy1, u_pop, u_ofs, k6, u_popgrad6, u_ofsgrad6, bpoints);

  for (ii = 0; ii < SystemSize; ii++) yy1[ii] = y[ii] + dt * (a71 * k1[ii] + a72 * k2[ii] + a73 * k3[ii] + a74 * k4[ii] + a75 * k5[ii] + a76 * k6[ii]);
  rk_level = 7;
  Gradient(yy1, u_pop, u_ofs, k7, u_popgrad7, u_ofsgrad7, bpoints);

  for (ii = 0; ii < SystemSize; ii++) yerr[ii] = dt * (db1 * k1[ii] + db2 * k2[ii] + db3 * k3[ii] + db4 * k4[ii] + db5 * k5[ii] + db6 * k6[ii] + db7 * k7[ii]);

  return;
} // rktsi5

//==================================================================================
#if (EVENT_NR > 0)

static double delvalue(double b, int eventindex, double del_h)

{
  register int                    ii;
  double                          result[EVENT_NR];
  double                          btilde1, btilde2, btilde3, btilde4, btilde5, btilde6, btilde7;

  btilde1  = -1.0530884977290216 * b * (b - 1.3299890189751412);
  btilde1 *= b * b - 1.4364028541716351 * b + 0.7139816917074209;

  btilde2  = 0.1017 * b * b;
  btilde2 *= b * b - 2.1966568338249754 * b + 1.2949852507374631;

  btilde3  = 2.490627285651252793 * b * b;
  btilde3 *= b * b - 2.38535645472061657 * b + 1.57803468208092486;

  btilde4  = -16.54810288924490272 * b * b;
  btilde4 *= (b - 1.21712927295533244) * (b - 0.61620406037800089);

  btilde5  = 47.37952196281928122 * b * b;
  btilde5 *= (b - 1.203071208372362603) * (b - 0.658047292653547382);

  btilde6  = -34.87065786149660974 * b * b;
  btilde6 *= (b - 1.2) * (b - 0.666666666666666667);

  btilde7  = 2.5 * b * b;
  btilde7 *= (b - 1) * (b  - 0.6);

  for (ii = 0; ii < SystemSize; ii++)
    yco[ii] = y[ii] + del_h * (btilde1 * k1[ii] + btilde2 * k2[ii] + btilde3 * k3[ii] +  btilde4 * k4[ii] +
                               btilde5 * k5[ii] + btilde6 * k6[ii] + btilde7 * k7[ii]);

  for (ii=0; ii<EVENT_NR; ii++) result[ii] = NO_EVENT;
  EventLocation(yco, c_pop, c_ofs, bpoints, result);
  
  return result[eventindex];
}

//==================================================================================
#define ITMAX        500
#define EPS        1.0e-16

static double zbrent(double olddel, double newdel, int eventindex, double del_h)

{
  int                             iter;
  double                          a, b, c = 0.0, d = 0.0, e = 0.0, min1, min2;
  double                          fa, fb, fc, p, q, r, s, tol1, xm;

  a  = 0.0;
  fa = olddel;
  b  = 1.0;
  fb = newdel;

  if (fb * fa > 0.0)
    {
      Warning(ZBB);
      return -1.0;
    }

  fc = fb;
  for (iter = 0; iter < ITMAX; iter++)
    {
      if (fb * fc > 0.0)
        {
          c  = a;
          fc = fa;
          e = d = b - a;
        }
      if (fabs(fc) < fabs(fb))
        {
          a  = b;
          fa = fb;
          b  = c;
          fb = fc;
          c  = a;
          fc = fa;
        }
      tol1 = 2.0 * EPS * fabs(b) + 0.5 * identical_zero;
      xm   = 0.5 * (c - b);

      if ((fabs(xm) <= tol1 && fb * fc <= 0.0) || fb == 0.0) return b;

      if (fabs(e) >= tol1 && fabs(fa) > fabs(fb))
        {
          s = fb / fa;
          if (a == c)
            {
              p = 2.0 * xm * s;
              q = 1.0 - s;
            }
          else
            {
              q = fa / fc;
              r = fb / fc;
              p = s * (2.0 * xm * q * (q - r) - (b - a) * (r - 1.0));
              q = (q - 1.0) * (r - 1.0) * (s - 1.0);
            }
          if (p > 0.0) q = -q;
          p    = fabs(p);
          min1 = 3.0 * xm * q - fabs(tol1 * q);
          min2 = fabs(e * q);
          if (2.0 * p < (min1 < min2 ? min1 : min2))
            {
              e = d;
              d = p / q;
            }
          else
            {
              d = xm;
              e = d;
            }
        }
      else
        {
          d = xm;
          e = d;
        }
      a  = b;
      fa = fb;
      if (fabs(d) > tol1)
        b += d;
      else
        b += (xm > 0.0 ? fabs(tol1) : -fabs(tol1));
      fb = delvalue(b, eventindex, del_h);
    }
  Warning(ZBM);

  return -1.0;
}
#undef ITMAX
#undef EPS

//==================================================================================

static void LocateEvent(double prev_dt, int *doloc, int *located)

/* 
 * LocateEvent - This routine is called to locate the events that are triggered by 
 *               the user-defined routine EventLocation(). It is assumed that on 
 *               entrance to this routine, the array doloc[] flags which events to 
 *               locate, i.e. for which event indicator oldELvalue[] and newELvalue[]
 *               have sound values, i.e. both non-zero and their product negative.
 */

{
  register int                    ii;
  int                             index = -1;
  double                          level;
  double                          stepfrac[EVENT_NR], smallest = 2.0;

  for (ii = 0; ii < EVENT_NR; ii++)
    {
      located[ii] = 0;
      if (doloc[ii])
        {
          stepfrac[ii] = zbrent(oldELvalue[ii], newELvalue[ii], ii, prev_dt);
          if ((stepfrac[ii] < 0.0) || (stepfrac[ii] > 1.0))
            {
              if (EBTDEBUG(1))
                {
                  (void)fprintf(dbgfil, "Problem locating event %d at T = %15.8f", ii, yy1[0]);
                  (void)fprintf(dbgfil, "  Start = %12.7E", oldELvalue[ii]);
                  (void)fprintf(dbgfil, "  Stop = %12.7E", newELvalue[ii]);
                  (void)fprintf(dbgfil, "  dt old = %12.7E", prev_dt);
                  (void)fprintf(dbgfil, "  dt new = %12.7E\n", stepfrac[ii] * prev_dt);
                  fflush(dbgfil);
                }
            }
          else if (stepfrac[ii] < smallest)
            {
              smallest = stepfrac[ii];
              index    = ii;
            }
        }
    }

  if (index < 0)
    {
      (void)memcpy((DEF_TYPE *)yy1, (DEF_TYPE *)y, SystemSize * sizeof(double));
      return;
    }

  if (stepfrac[index] == 0.0)
    (void)memcpy((DEF_TYPE *)yy1, (DEF_TYPE *)y, SystemSize * sizeof(double));
  else
    {
#if (1)
      // Use the continuous output state to continue
      double                      b;
      double                      btilde1, btilde2, btilde3, btilde4, btilde5, btilde6, btilde7;

      b = stepfrac[index];
      btilde1  = -1.0530884977290216 * b * (b - 1.3299890189751412);
      btilde1 *= b * b - 1.4364028541716351 * b + 0.7139816917074209;

      btilde2  = 0.1017 * b * b;
      btilde2 *= b * b - 2.1966568338249754 * b + 1.2949852507374631;

      btilde3  = 2.490627285651252793 * b * b;
      btilde3 *= b * b - 2.38535645472061657 * b + 1.57803468208092486;

      btilde4  = -16.54810288924490272 * b * b;
      btilde4 *= (b - 1.21712927295533244) * (b - 0.61620406037800089);

      btilde5  = 47.37952196281928122 * b * b;
      btilde5 *= (b - 1.203071208372362603) * (b - 0.658047292653547382);

      btilde6  = -34.87065786149660974 * b * b;
      btilde6 *= (b - 1.2) * (b - 0.666666666666666667);

      btilde7  = 2.5 * b * b;
      btilde7 *= (b - 1) * (b  - 0.6);

      for (ii = 0; ii < SystemSize; ii++)
        yy1[ii] = y[ii] + prev_dt * (btilde1 * k1[ii] + btilde2 * k2[ii] + btilde3 * k3[ii] +  btilde4 * k4[ii] +
                                     btilde5 * k5[ii] + btilde6 * k6[ii] + btilde7 * k7[ii]);
#else
      rktsi5(stepfrac[index] * prev_dt);
#endif
    }

  located[index] = 1;
  LocatedEvent   = index;

  for (ii = 0; ii < EVENT_NR; ii++) newELvalue[ii] = NO_EVENT;
  EventLocation(yy1, u_pop, u_ofs, bpoints, newELvalue);

  equal2zero = max(identical_zero, pow(10, ceil(log10(fabs(newELvalue[index]) + ABS_ERR))));

  cohort_end = ForceCohortEnd(yy1, u_pop, u_ofs, bpoints);

  // Report performance if required by user
  if (EBTDEBUG(1))
    {
      if (EBTDEBUG(4))
        {
          fprintf(dbgfil, "%-14s%3d: T = %15.8f     dt = %12.7E\n", "Step to event", index, env[0], stepfrac[index] * prev_dt);
          fflush(dbgfil);
        }

      if (EBTDEBUG(2) || (EBTDEBUG(1) && (fabs(newELvalue[index]) >= identical_zero)))
        {
          if (cohort_end)
            (void)fprintf(dbgfil, "%-18s T = %15.8f", "Cohort closed:", yy1[0]);
          else
            (void)fprintf(dbgfil, "%-18s T = %15.8f", "Event located:", yy1[0]);
          (void)fprintf(dbgfil, "  Value = %12.7E", newELvalue[index]);
          if (fabs(newELvalue[index]) >= identical_zero)
            {
              (void)fprintf(dbgfil, " ");
              level = (ceil(log10(fabs(newELvalue[index]) / identical_zero)) - identical_zero);
              for (ii = 0; ii < level; ii++) (void)fprintf(dbgfil, "*");
            }
          (void)fprintf(dbgfil, "\n");
          (void)fflush(dbgfil);
        }
    }

  return;
} // LocateEvent
#endif // EVENT_NR > 0

//==================================================================================

static void      IntermediateState(double b, double del_h)

/* 
 * IntermediateState - Routine computes the state of the system at an intermediate 
 *                     time point by interpolation. Values are stored in basic data 
 *                     copy for further use in output routines.
 */
  
{
  register int                    ii;
  double                          btilde1, btilde2, btilde3, btilde4, btilde5, btilde6, btilde7;

  btilde1  = -1.0530884977290216 * b * (b - 1.3299890189751412);
  btilde1 *= b * b - 1.4364028541716351 * b + 0.7139816917074209;

  btilde2  = 0.1017 * b * b;
  btilde2 *= b * b - 2.1966568338249754 * b + 1.2949852507374631;

  btilde3  = 2.490627285651252793 * b * b;
  btilde3 *= b * b - 2.38535645472061657 * b + 1.57803468208092486;

  btilde4  = -16.54810288924490272 * b * b;
  btilde4 *= (b - 1.21712927295533244) * (b - 0.61620406037800089);

  btilde5  = 47.37952196281928122 * b * b;
  btilde5 *= (b - 1.203071208372362603) * (b - 0.658047292653547382);

  btilde6  = -34.87065786149660974 * b * b;
  btilde6 *= (b - 1.2) * (b - 0.666666666666666667);

  btilde7  = 2.5 * b * b;
  btilde7 *= (b - 1) * (b  - 0.6);

  for (ii = 0; ii < SystemSize; ii++)
    yco[ii] = y[ii] + del_h * (btilde1 * k1[ii] + btilde2 * k2[ii] + btilde3 * k3[ii] +  btilde4 * k4[ii] +
                               btilde5 * k5[ii] + btilde6 * k6[ii] + btilde7 * k7[ii]);

  (void)memcpy((DEF_TYPE *)env, (DEF_TYPE *)yco, ENVIRON_DIM * sizeof(double));

#if (POPULATION_NR > 0)
  register int                    jj, kk;
  int                             len;

  len = ENVIRON_DIM;
  for (ii = 0; ii < POPULATION_NR; ii++)
    {
      (void)memcpy((DEF_TYPE *)pop[ii], (DEF_TYPE *)(yco + len), (table_size[ii] * COHORT_SIZE) * sizeof(double));
      len += (table_size[ii] * COHORT_SIZE);
    }

  for (ii = 0; ii < POPULATION_NR; ii++)
    {
      for (jj = 0; jj < BpointNo[ii]; jj++)
        {
          if (ofs[ii][jj][number] > 0)
            {
              for (kk = 1; kk < COHORT_SIZE; kk++) ofs[ii][jj][kk] /= ofs[ii][jj][number];
            }
          for (kk = 1; kk < COHORT_SIZE; kk++) ofs[ii][jj][kk] += bpoints[ii][jj][kk];
        }
    }
#endif // (POPULATION_NR > 0)

  return;
}

//==================================================================================

double IntegrationStep(double del_tim, double del_max, int recurs)

/* 
 * IntegrationStep - Performs an integration with adaptable step size but maximum 
 *                   "del_max". Step size control and event location as implemented
 *                   by E. Hairer & G. Wanner.
 */

{
  register int                    ii;
  double                          del_h;
  double                          err, err0, sk, sqr, maxsqr;
  double                          fac, fac11, hnew;
  int                             adjust = 1, intermediate = 0, maxerri;
#if (EVENT_NR > 0)
  int dolocation[EVENT_NR], events = 0;
#endif

  // Adjust stepsize to hit cohort end, look ahead two steps to avoid too drastic step changes
  del_h = del_tim;
  if (del_max < (2 * del_tim))
    {
      del_h = 0.5 * del_max;
      if (del_max < del_tim) del_h = del_max;
      adjust = 0;
    }
  if (del_h < SMALLEST_STEP)
    {
#if (POPULATION_NR > 0)
      TransBcohorts();
#endif // (POPULATION_NR > 0)
      ErrorExit(0, SSS);
    }
  recur_no     = recurs;
  LocatedEvent = -1;

  if (EBTDEBUG(4))
    {
      fprintf(dbgfil, "%-18s T = %15.8f     dt = %12.7E recurs = %2d\n", "Starting step:", env[0], del_h, recurs);
      fflush(dbgfil);
    }

  // Do an integration step
  rktsi5(del_h);

   // Error estimation
  err     = 0.0;
  maxsqr  = 0.0;
  maxerri = 0;
  for (ii = 0; ii < SystemSize; ii++)
    {
      sk  = ABS_ERR + accuracy * max(fabs(y[ii]), fabs(yy1[ii]));
      sqr = yerr[ii] / sk;
      if (sqr > maxsqr)
        {
          maxsqr  = sqr;
          maxerri = ii;
        }
      err += sqr * sqr;
    }
  err0 = err;
  err  = sqrt(err / (double)SystemSize);

  if (err > 1.0)                                                                    // Step rejected
    {
      if (EBTDEBUG(3))
        {
          fprintf(dbgfil, "%-18s T = %15.8f     dt = %12.7E recurs = %2d Largest error contribution in ODE #%d (%.3f%%)\n", "Step failed:", env[0],
                  del_h, recur_no, maxerri, 100 * maxsqr * maxsqr / err0);
          fflush(dbgfil);
        }

      // If bigger than accuracy take smaller step and restart
      recur_no++;
      step_failed = 1;
      if (recur_no > 25)
        {
#if (POPULATION_NR > 0)
          TransBcohorts();
#endif // (POPULATION_NR > 0)
          fprintf(stderr, "\n%-18s T = %15.8f     dt = %12.7E recurs = %2d Largest error contribution in ODE #%d (%.3f%%)\n\n", "Step failed:", env[0],
                  del_h, recur_no, maxerri, 100 * maxsqr * maxsqr / err0);
          ErrorExit(0, REC);
        }

      fac11  = pow(err, 0.2 - BETA * 0.75);
      del_h /= min(FACC1, fac11 / SAFETY);

      step_size = del_h;
      del_h     = IntegrationStep(del_h, del_max, recur_no);
    }
  else                                                                              // Step accepted
    {
      if (EBTDEBUG(4))
        {
          fprintf(dbgfil, "%-18s T = %15.8f     dt = %12.7E recurs = %2d\n", "Step OK:", env[0], del_h, recur_no);
          fflush(dbgfil);
        }

      // Step size adjustment using Lund-stabilization: we require fac1 <=  hnew/h <= fac2
      // No increase if failed just before
      if (adjust && !step_failed)
        {
          fac11  = pow(err, 0.2 - BETA * 0.75);
          fac    = fac11 / pow(facold, BETA);
          fac    = max(FACC2, min(FACC1, fac / SAFETY));
          hnew   = del_h / fac;
          facold = max(err, FACOLD);

          step_size = min(hnew, cohort_limit);
          step_size = min(step_size, LARGEST_STEP);
        }
      step_failed = 0;
      accepted_steps++;

      /* 
       * Location of events: If located in previous and no change in ELvalue do not locate.
       * WARNING: The order of these statements seems odd but is OK! What is checked is whether 
       *          the newELvalue from before cohort closure equals the value computed at the 
       *          beginning of this time integration step!
       */
#if (EVENT_NR > 0)
      for (ii = 0; ii < EVENT_NR; ii++)
        {
          dolocation[ii] = 1;
          if (located[ii] && isequal(oldELvalue[ii], newELvalue[ii])) dolocation[ii] = 0;
        }

      for (ii = 0; ii < EVENT_NR; ii++) newELvalue[ii] = NO_EVENT;
      EventLocation(yy1, u_pop, u_ofs, bpoints, newELvalue);

      for (ii = 0, events = 0; ii < EVENT_NR; ii++)
        {
          if (((oldELvalue[ii] < 0.0) && (newELvalue[ii] < 0.0)) || ((oldELvalue[ii] > 0.0) && (newELvalue[ii] > 0.0))) dolocation[ii] = 0;
          if (dolocation[ii]) events++;
        }
#endif

      intermediate = ((next_output < (yy1[0] - identical_zero)) || ((state_out > 0.0) && (next_state_output < (yy1[0] - identical_zero))));

#if (EVENT_NR > 0)
      if (events)
        LocateEvent(del_h, dolocation, located);
      else
        for (ii = 0; ii < EVENT_NR; ii++) located[ii] = 0;
#endif

      // Produce intermediate output and state output if requested
      if (intermediate)
        {
#if (POPULATION_NR > 0)
          for (ii = 0; ii < POPULATION_NR; ii++) CohortNo[ii] += BpointNo[ii];
#endif // (POPULATION_NR > 0)
          while (next_output < (yy1[0] - identical_zero))
            {
              IntermediateState((next_output - y[0]) / del_h, del_h);
              FileOut();
              next_output += delt_out;
            }
#if (POPULATION_NR > 0)
          while ((state_out > 0.0) && (next_state_output < (yy1[0] - identical_zero)))
            {
              IntermediateState((next_state_output - y[0]) / del_h, del_h);
              FileState();
              next_state_output += state_out;
            }
          for (ii = 0; ii < POPULATION_NR; ii++) CohortNo[ii] -= BpointNo[ii];
          for (ii = 0; ii < POPULATION_NR; ii++) cohort_no[ii] = CohortNo[ii];
#endif // (POPULATION_NR > 0)
        }

      // Update the basic data copy, and the local copy 
      (void)memcpy((DEF_TYPE *)env, (DEF_TYPE *)yy1, ENVIRON_DIM * sizeof(double));

#if (POPULATION_NR > 0)
      int len;

      len = ENVIRON_DIM;
      for (ii = 0; ii < POPULATION_NR; ii++)
        {
          (void)memcpy((DEF_TYPE *)pop[ii], (DEF_TYPE *)(yy1 + len), (table_size[ii] * COHORT_SIZE) * sizeof(double));
          len += (table_size[ii] * COHORT_SIZE);
        }
#endif // (POPULATION_NR > 0)

      (void)memcpy((DEF_TYPE *)y, (DEF_TYPE *)yy1, SystemSize * sizeof(double));
    }

  return del_h;
}

//==================================================================================
