/***
   NAME
     ebttmpl.c
   PURPOSE
     A template for the user-defined routines of the Escalator Boxcar Train
     integration program.
   NOTES
     
   HISTORY
     AMdR - Jul 21, 1995 : Created.
     AMdR - Jan 06, 2012 : Revised last.
***/

/*==========================================================================
 *
 *			INCLUDING THE HEADER FILE
 *
 * Include the header file for standard library prototyping and definition
 * of the data-structures.
 *
 *==========================================================================
 */
#include  "escbox.h"



/*
 *==========================================================================
 *
 *		LABELLING ENVIRONMENT AND I-STATE VARIABLES
 *
 * For convenience it is possible to label the environment and i-state
 * variables with a more meaningful name by defining, for instance: 
 *
 *		#define time	env[0]
 *		#define length	i_state(0)
 *
 * Note in this respect again the zero based array indexing in the
 * C-language.
 *
 *==========================================================================
 */

#define	age		i_state(0)
#define	length		i_state(1)



/*
 *==========================================================================
 *
 *		DEFINING AND LABELLING CONSTANTS AND PARAMETERS
 *
 * Define the constant names and values used in the user-specified
 * routines. Most parameters in the problem can be treated as constants for
 * this file. This seems the most easy way to specify the parameters. The
 * parameters that are free to change will be in a vector called
 * "parameter[]", defined elsewhere. As with the i-state variables it is 
 * possible to label these parameters with a more meaningful name, e.g.:
 *
 *		#define	   ALPHA    parameter[0]
 *
 *==========================================================================
 */

#define AMAX		10000
#define MAX_COHORT_SIZE	1.0

#define ALPHA		parameter[0]
#define DELTA		parameter[1]
#define RM		parameter[2]
#define K		parameter[3]



/*
 *==========================================================================
 *
 * USER INITIALIZATION ROUTINE ALLOWS OPERATIONS ON INITIAL POPULATIONS
 *
 *==========================================================================
 */

void	UserInit( int argc, char **argv, double *env,  population *pop)
  
  /*  
   * The function header above should not be changed.
   *
   * In the routine below the user can specify some operations that are to 
   * be performed before the actual integration starts. Initialization of 
   * user defined variables can, for instance, be carried out here.
   *
   *     NB : -	The integer array "cohort_no[]" is globally available and
   *		denotes the number of internal cohorts present in each
   *		structured population. 
   */
  
{
  
  return;
}



/*
 *==========================================================================
 *
 *	SPECIFICATION OF THE NUMBER AND VALUES OF BOUNDARY POINTS
 *
 *==========================================================================
 */

void	SetBpointNo(double *env, population *pop, int *bpoint_no)
  /*
   * This function can also be defined as:

   void	CycleStart(double *env, population *pop, int *bpoint_no)

   * The function names SetBpointNo and CycleStart are synonyms.
   *
   * Apart from this, the function header above should not be changed.
   *
   * Specify below the number of boundary cohorts that should be created at
   * the start of the next cohort integration cycle. Fill the array
   * "bpoint_no[]" with the appropriate integer values. The length of the
   * array is "POPULATION_NR".
   * The state of the environment and the population can be used to adapt
   * the number of the fixed points on the boundary to the current state.
   *
   *     NB : -	The integer array "cohort_no[]" is globally available and
   *		denotes the number of internal cohorts present in each
   *		structured population. 
   */
  
{
  bpoint_no[0]=1;
  
  return;
}



/*==========================================================================*/

void	SetBpoints(double *env, population *pop, population *bpoints)
  
  /*
   * The function header above should not be changed.
   *
   * Define below the fixed points at the boundary of the state space that
   * will characterize the boundary cohorts in the subsequent cohort cycle.
   * These boundary points should be addressed in the same way as the
   * population cohorts and the offspring cohorts, i.e.:
   *
   *		      bpoints[1][4][i_state(2)]
   *
   * to denote the fixed value of the 3rd i_state variable of the 5th
   * boundary point in the 2nd population (Note again the base of 0 of an
   * array in the C-language). 
   * Set the fixed value for all the i_state variables in all the boundary
   * cohorts for all populations. The number of populations is given by
   * "POPULATION_NR", the number of boundary cohorts is given by the
   * globally available vector "bpoint_no[]" of length "POPULATION_NR" and
   * the number of i_state variables is obviously determined by
   * "I_STATE_DIM". 
   * The state of the environment and the population can be used to adapt
   * the values of the fixed points on the boundary to the current state.
   *
   *     NB : -	The integer array "cohort_no[]" is globally available and
   *		denotes the number of internal cohorts present in each
   *		structured population. 
   */
  
{
  bpoints[0][0][age]=0.0; bpoints[0][0][length]=0.0;
  
  return;
}



/*
 *==========================================================================
 *
 *			SPECIFICATION OF DERIVATIVES
 *
 *==========================================================================
 */

void	Gradient(double *env,     population *pop,     population *ofs,
		 double *envgrad, population *popgrad, population *ofsgrad,
		 population *bpoints)
  
  /*
   * The function header above should not be changed.
   *
   * Define the derivatives of the various environment variables,
   * which have to be returned to the main program in the array
   * "envgrad[]". 
   * Define also the derivatives of the various population variables,
   * which have to be returned to the main program in the matrix of cohort
   * variables "popgrad[][][]" for each population.
   * Finally define the derivatives of the various offspring variables,
   * which have to be returned to the main program in the matrix of cohort
   * variables "ofsgrad[][][]" for each population.
   *
   *     NB : -	The integer array "cohort_no[]" is globally available and
   *		denotes the number of internal cohorts present in each
   *		structured population. 
   *	      - The integer array "bpoint_no[]" is globally available and
   *		denotes the number of boundary cohorts present in each
   *		structured population during the current cohort cycle.
   *	      - The offspring number can be zero. If used in a division, check
   *		for this equality to zero!! 
   */

{
  double		functional_response;
  double		total_area, mean_offspring_length;
  register int		i;
  
  functional_response = env[1]/(1+env[1]);	/* Determine S/(1+S)        */
  
						/* Determine the integral   */
  for(i=0, total_area=0; i<cohort_no[0]; i++)	/* of the total area        */
      total_area += pow(pop[0][i][length], 2.0)*pop[0][i][number];

						/* Include the boundary     */
  if(ofs[0][0][number] > 1.0E-10)		/* cohort if non-zero       */
    {						/* Convert length moment    */
      mean_offspring_length = ofs[0][0][length]/ofs[0][0][number];
      total_area += pow(mean_offspring_length, 2.0)*ofs[0][0][number];
    }
						/* The time derivative      */
  envgrad[0] = 1;				/* The food derivative      */
  envgrad[1] = RM*env[1]*(1-env[1]/K) - functional_response*total_area;
  
  for(i=0; i<cohort_no[0]; i++)			/* The derivatives for all  */
    {						/* internal cohorts	    */
      popgrad[0][i][number] = -DELTA*pop[0][i][number];
      popgrad[0][i][age]    = 1;
      popgrad[0][i][length] = functional_response - pop[0][i][length];
    }
						/* The derivatives for the  */
						/* boundary cohort	    */
  ofsgrad[0][0][number] = -DELTA*ofs[0][0][number]
      + ALPHA*functional_response*total_area;
  ofsgrad[0][0][age]    = -DELTA*ofs[0][0][age]
      + ofs[0][0][number];
  ofsgrad[0][0][length] = -DELTA*ofs[0][0][length]   
      + functional_response*ofs[0][0][number]
      - ofs[0][0][length];
  
  return;
}



/*
 *==========================================================================
 *
 *	SPECIFICATION OF EVENT LOCATION AND DYNAMIC COHORT CLOSURE
 *
 *==========================================================================
 */

void	EventLocation(double *env, population *pop, population *ofs,
		      population *bpoints, double *events)
  
  /*
   * The function header above should not be changed. 
   *
   * When the value of EVENT_NR is larger than 0, the current routine is called
   * to check whether a discontinuity in the time derivatives or an event has
   * occurred. Such a discontinuity or event is indicated by a zero (0) value
   * of one of the elements of the array events[]. Hence, every time the
   * value of one of these array elements changes from positive to negative or
   * vice versa within a time integration step, this is interpreted as an event
   * occurrence. If an event occurs, the integration will not step over the
   * time point of the event, but will localize the moment of its occurrence by
   * means of a root finding procedure. This localization will stop when it has
   * found a time, for which the value of the particular array element that
   * changed sign is within the tolerance bounds, distinguishing values from
   * zero, as set in the CVF file. After the event time has been localized,
   * integration will be carried out exactly up to the moment of event
   * occurrence. Subsequently the routine ForceCohortEnd() (see below) is
   * called and the integration is started from the moment of the event
   * location onwards. If no event location is required, this routine should
   * not change the value of the elements in events[] or the macro EVENT_NR
   * should be set to 0. 
   *
   *     NB : -	The integer array "cohort_no[]" is globally available and
   *		denotes the number of internal cohorts present in each
   *		structured population. 
   *	      - The integer array "bpoint_no[]" is globally available and
   *		denotes the number of boundary cohorts present in each
   *		structured population during the current cohort cycle.
   *	      - The offspring number can be zero. If used in a division, check
   *		for this equality to zero!! 
   */
  
{
  register int		i;

  events[0] = ofs[0][0][number] - MAX_COHORT_SIZE;
  for (i=1; i<bpoint_no[0]; i++)
      events[0] = max(events[0], ofs[0][i][number] - MAX_COHORT_SIZE);
  
  return;
}




/*==============================================================================*/

int	ForceCohortEnd(double *env, population *pop, population *ofs,
		       population *bpoints)
  
  /*
   * The function header above should not be changed. 
   *
   * When the value of EVENT_NR is larger than 0, the current routine is called
   * after an event has been localized on the basis of the EventLocation()
   * routine defined above. This routine should then return either COHORT_END
   * (=1) or NO_COHORT_END (=0), indicating whether a cohort closure is
   * currently required or not. If the macro DYNAMIC_COHORTS in the program
   * header file is set to 0 (its default value) closing off boundary cohorts
   * and transforming them into internal cohorts, takes place both at constant
   * time intervals, as specified by the cohort cycle limit variable in the
   * .CVF file and whenever the current routine returns COHORT_END. This allows
   * for dynamic closure of boundary cohorts, for instance, to ensure a limited
   * number of individuals per cohort. If DYNAMIC_COHORTS is set to 1 in the
   * program header file, closing off of boundary cohorts is exclusively
   * performed when the current routine returns COHORT_END.
   *
   *     NB : -	The integer array "cohort_no[]" is globally available and
   *		denotes the number of internal cohorts present in each
   *		structured population. 
   *	      - The integer array "bpoint_no[]" is globally available and
   *		denotes the number of boundary cohorts present in each
   *		structured population during the current cohort cycle.
   *	      - The offspring number can be zero. If used in a division, check
   *		for this equality to zero!! 
   */
  
{
  static int		iter = 0;
  
  iter++;
  if ((iter%3) == 0) return COHORT_END;
  else return NO_COHORT_END;

  /* return COHORT_END;    */
  /* return NO_COHORT_END; */
}



/*
 *==========================================================================
 *
 *		SPECIFICATION OF BETWEEN COHORT CYCLE DYNAMICS
 *
 *==========================================================================
 */

void	InstantDynamics(double *env, population *pop, population *ofs)
  
  /*
   * This function can also be defined as:

   void	CycleEnd(double *env, population *pop, population *ofs)

   * The function names SetBpointNo and CycleStart are synonyms.
   *
   * Apart from this, the function header above should not be changed.
   *
   * This routine is called at the end of each cohort cycle. The routine
   * can be used to implement any type of instantaneous dynamics, occurring
   * between two subsequent cohort cycles. It can, for instance, be used
   * for a pulsed, instantaneous reproduction process or to set the number
   * of individuals in cohorts that have reached their maximum lifespan to
   * 0. 
   * Note that the transformation of boundary cohorts into internal cohorts
   * has already been performed and that the boundary cohorts are hence
   * characterized by the number of individuals and their transformed
   * moments (usually denoted by the symbol mu).
   * In an instantaneous reproduction process the i_state of the offspring
   * can therefore be simply specified in terms of the mean i_state!
   *
   *     NB : -	The integer array "cohort_no[]" is globally available and
   *		denotes the number of internal cohorts present in each
   *		structured population. 
   *	      - The integer array "bpoint_no[]" is globally available and
   *		denotes the number of boundary cohorts present in each
   *		structured population during the current cohort cycle.
   *	      - The offspring number can be zero. If used in a division, check
   *		for this equality to zero!! 
   */

{
  register int		i;
  
  for(i=0; i<cohort_no[0]; i++)			/* Empty cohorts that have  */
    {						/* reached maximum age      */
      if(pop[0][i][age] > AMAX) pop[0][i][number] = 0;
    }
  
  return;
}



/*
 *==========================================================================
 *
 *			SPECIFICATION OF OUTPUT VARIABLES
 *
 *==========================================================================
 */

void	DefineOutput(double *env, population *pop, double *output)
  
  /*
   * The function header above should not be changed.
   *
   * Define below the values of the output variables in terms of the
   * population and environment statistics. These values have to be
   * returned to the main program in the array "output[]".
   *
   *     NB : -	The integer array "cohort_no[]" is globally available and
   *		denotes the number of internal cohorts present in each
   *		structured population. 
   */

{						/* OUTPUT VARIABLES:        */
  register int		i;			/* 1: Food density	    */
						/* 2: Total number	    */
						/* 3: Total biomass	    */
  output[0] = env[1];
  for(i=0; i<cohort_no[0]; i++)
    {
      output[1] += pop[0][i][number];
      output[2] += pow(pop[0][i][length], 3.0)*pop[0][i][number];
    }
  
  return;
}



/*==========================================================================*/
