/***
  NAME
    csb2txt

    Program converts a binary CSB file to text. Either the entire file is
    converted, or just the single state indicated by the time value that is
    passed to the program as command-line

  Last modification: AMdR - Aug 13, 2010
***/
#ifndef CSB2TXT
#define CSB2TXT
#endif

#include "stdio.h"
#include "stdlib.h"
#include "string.h"
#include "math.h"
#include "float.h"
#include <dlfcn.h>
#include <sys/stat.h>
#include <sys/param.h>
#include <unistd.h>

#define MAXCOLUMNS                100
#define MAXPARS                   1024
#define MAXSTRINGLEN              1024
#define MAXPOP                    20
#define USAGE                     "csb2txt [-isf] [-f \"printf() format string\"] [-c DefineOutput() file] <CSB filename> [time value]"

#if defined(__APPLE__)
#define SOEXT    ".dylib"
#define SOEXTLEN  6
#else
#define SOEXT    ".so"
#define SOEXTLEN  3
#endif // defined(__APPLE__)

// List of magic keys of all recognized CSB file types

#define CSB_MAGIC_KEY1            20030509

// Macros to round to memory pages

#define PAGESIZE                  (getpagesize())
#define ROUND_PAGE(s)             ((int)((s & (PAGESIZE - 1)) ? (s & ~(PAGESIZE - 1)) + PAGESIZE : s))

#if defined(DBL_MAX)                                                                // Stub value for missing data point
#define   MISSING_VALUE           (DBL_MAX)
#elif defined(MAXDOUBLE)
#define   MISSING_VALUE           (MAXDOUBLE)
#else
#define   MISSING_VALUE           (1.23456789e+307)
#endif

// Type definition for environment and population

#include "ebtcsbdefs.h"

void                              *mem_base = NULL;
long                              MemAllocated = 0L;
int                               (*defout)(double *, double **, double *, double *, int *);
char                              buf[MAXPATHLEN];
char                              printfmt[128];
double                            **pop;
int                               *Cohort_No;
double                            output[MAXCOLUMNS];
double                            csb_pars[MAXPARS];

/*==================================================================================================================================*/

int isfile(char *f)                                                                 // Returns 1 if file is regular, 0 otherwise
{
  struct stat st;

  if (stat(f, &st) != 0)
    return 0;

  return ((st.st_mode & S_IFMT) == S_IFREG);
}

/*==================================================================================================================================*/

static void write_txt_file(int isf)

{
  register int  i, j, k;
  double        *cdbl, *base_pnt, output;
  Envdim        *cenv;
  Popdim        *cpop;

  cenv = (Envdim *)mem_base;
  cdbl = ((double *)mem_base + cenv->data_offset + cenv->columns);
  cpop = (Popdim *)cdbl;
  base_pnt = ((double *)mem_base + cenv->data_offset);

  // Write environment state
  for (i = 0; i < cenv->columns; i++)
    {
      if ((i == 0) && isf)
        output = 0.0;
      else if (i == 0)
        output = base_pnt[i];
      else
        {
          (void)printf("\t");
          output = base_pnt[i];
        }
      if (strlen(printfmt))
        (void)printf(printfmt, output);
      else if (((fabs(output) <= 1.0E4) && (fabs(output) >= 1.0E-4)) || (output == 0))
        (void)printf("%.10f", output);
      else
        (void)printf("%.6E", output);
    }
  (void)printf("\n\n");

  // Write population state
  while (1)
    {
      base_pnt = ((double *)cpop + cpop->data_offset);
      for (j = 0; j < cpop->cohorts; j++, (void)printf("\n"))
        for (k = 0; k < cpop->columns; k++)
          {
            if (k > 0)
              (void)printf("\t");
            output = *(base_pnt + k * cpop->cohorts + j);
            if (strlen(printfmt))
              (void)printf(printfmt, output);
            else if (((fabs(output) <= 1.0E4) && (fabs(output) >= 1.0E-4)) ||
                     (output == 0))
              (void)printf("%.10f", output);
            else
              (void)printf("%.6E", output);
          }
      (void)printf("\n");
      if (cpop->lastpopdim)
        break;

      cdbl = (((double *)cpop) + cpop->data_offset + (cpop->cohorts * cpop->columns));
      cpop = (Popdim *)cdbl;
    }
  (void)printf("\n");

  return;
}

/*==================================================================================================================================*/

static void load_defout_proc(char *fn)

{
  char          *ch;
  const char    *error;
  char          libnam[MAXPATHLEN];
  static void   *dlhandle;

  if (!strcmp((fn + strlen(fn) - 2), ".c"))
    {
      strcpy(libnam, fn);
      ch = libnam + strlen(libnam) - 2;
      (void)strcpy(ch, SOEXT);
      if (isfile(libnam))
        (void)unlink(libnam);

      (void)sprintf(buf, "%s %s", "CompDefOut", fn);
      if (system(buf) || !isfile(libnam))
        {
          (void)fprintf(stderr, "\n\n%s %s %s\n\n", "Compilation of", fn, "unsuccessful.\nCheck the \'ErrLog\' file.");
          exit(1);
        }
      load_defout_proc(libnam);
      return;
    }
  else if (!strcmp((fn + strlen(fn) - SOEXTLEN), SOEXT))
    {
      if (!isfile(fn))
        {
          (void)fprintf(stderr, "\n\nLibrary file %s not found!\n\n", fn);
          exit(1);
        }
      if (dlhandle)
        {
          dlclose(dlhandle);
          dlhandle = (void *)NULL;
          defout = NULL;
        }
      if (fn[0] != '/')
        {
          if (!getcwd(libnam, MAXPATHLEN))
            {
              (void)fprintf(stderr, "\n\nFailed to get current working directory: %s!\n\n", libnam);
              exit(1);
            }
          strcat(libnam, "/");
          strcat(libnam, fn);
        }
      else
        strcpy(libnam, fn);
      dlhandle = dlopen(libnam, RTLD_NOW);
      if (!dlhandle)
        {
          (void)fprintf(stderr, "\n\nLoading of %s unsuccessful.\n%s\n\n",
                        fn, dlerror());
          exit(1);
        }
      defout = dlsym(dlhandle, "DefOutWrapper");
      if ((error = (const char *)dlerror()) != NULL)
        {
          (void)fprintf(stderr,
                        "\n\nLoading of DefineOutput() unsuccessful.\n%s\n\n",
                        dlerror());
          exit(1);
        }
    }

  return;
}

/*==================================================================================================================================*/

void compute_from_state()

{
  register int  i, maxpop;
  int           ncols;
  double        *env, *cdbl;
  Envdim        *cenv;
  Popdim        *cpop;

  if (!defout) return;

  cenv = (Envdim *)mem_base;
  cdbl = ((double *)mem_base + cenv->data_offset + cenv->columns);
  cpop = (Popdim *)cdbl;
  env = ((double *)mem_base + cenv->data_offset);

  for (maxpop = 0; maxpop < MAXPOP; maxpop++)
    {
      cpop = (Popdim *)cdbl;
      cdbl += cpop->data_offset + cpop->cohorts * cpop->columns;
      if (cpop->lastpopdim) break;
    }
  maxpop++;
  pop = (double **)calloc((size_t)maxpop, sizeof(double *));
  Cohort_No = (int *)calloc((size_t)maxpop, sizeof(int));

  cdbl = ((double *)mem_base + cenv->data_offset + cenv->columns);
  for (i = 0; i < maxpop; i++)
    {
      cpop = (Popdim *)cdbl;
      Cohort_No[i] = cpop->cohorts;
      pop[i] = cdbl + cpop->data_offset;
      cdbl += cpop->data_offset + cpop->cohorts * cpop->columns;
      if (cpop->lastpopdim) break;
    }

  for (i = 0; i < MAXCOLUMNS; i++)
    output[i] = 0.0;
  ncols = (*defout)(env, pop, output, csb_pars, Cohort_No);

  for (i = 0; i < ncols; i++)
    {
      if (i)
        (void)printf("\t");
      if (strlen(printfmt))
        (void)printf(printfmt, output);
      else if (((fabs(output[i]) <= 1.0E4) && (fabs(output[i]) >= 1.0E-4)) || (output[i] == 0))
        (void)printf("%.10f", output[i]);
      else
        (void)printf("%.6E", output[i]);
    }
  (void)printf("\n");

  free(pop);
  free(Cohort_No);

  return;
}

/*==================================================================================================================================*/

int main(int argc, char **argv)

{
  char                            **argpnt1 = NULL, **argpnt2 = NULL, **my_argv = NULL;
  int                             my_argc;
  int                             single = 0, isf = 0, cmpval = 0;
  double                          timeval = -1.0, closest_time = HUGE_VAL;
  char                            *fn;
  FILE                            *fp = NULL;
  struct stat                     stattmp;
  uint32_t                        filepos = 0, closest_fpos = -1, magic;
  Envdim                          cur_env;
  int                             pars, file_ok = 1, closest_fpos_set = 0;

  my_argv = (char **)malloc((size_t)argc * sizeof(char *));
  argpnt1 = argv;
  argpnt2 = my_argv;
  my_argc = 0;
  strcpy(printfmt, "");
  while (*argpnt1)
    {
      if (!strcmp(*argpnt1, "-isf")) isf = 1;
      else if (!strcmp(*argpnt1, "-c"))
        {
          argpnt1++;
          if (!*argpnt1)
            {
              fprintf(stderr, "\nNo DefineOutput() file specified!\n");
              fprintf(stderr, "\nUsage: %s\n\n", USAGE);
              exit(1);
            }
          load_defout_proc(*argpnt1);
          cmpval = 1;
        }
      else if (!strcmp(*argpnt1, "-f"))
        {
          argpnt1++;
          if (!*argpnt1)
            {
              fprintf(stderr, "\nNo print format specified!\n");
              fprintf(stderr, "\nUsage: %s\n\n", USAGE);
              exit(1);
            }
          strcpy(printfmt, *argpnt1);
        }
      else
        {
          *argpnt2 = *argpnt1;
          my_argc++;
          argpnt2++;
        }
      argpnt1++;
    }

  switch (my_argc)
    {
    case 3:
      timeval = atof(my_argv[2]);
      single = 1;
    case 2:
      fn = my_argv[1];
      if (!strcmp(fn + strlen(fn) - 1, "."))
        strcat(fn, "csb");
      else if (strcmp(fn + strlen(fn) - 4, ".csb"))
        strcat(fn, ".csb");

      if (!stat(fn, &stattmp)) /* couldn't stat            */
        {
          if ((!strcmp(fn + strlen(fn) - 3, "csb")) && (stattmp.st_mode & S_IFREG))
            fp = fopen(fn, "r"); /* open if regular file     */
          else
            {
              fprintf(stderr, "File %s is not a regular CSB file.\n", fn);
              return 1;
            }
        }
      if (fp == NULL)
        {
          fprintf(stderr, "Can't open file %s\n", fn);
          return 1;
        }
      break;
    default:
      fprintf(stderr, "\nUsage: %s\n\n", USAGE);
      return 1;
    }

  if (fread(&magic, sizeof(uint32_t), 1, fp) != 1)
    {
      fprintf(stderr, "\nCSB file does not contain any data!\n\n");
      return 1;
    }

  switch (magic)
    {
    case CSB_MAGIC_KEY1:
      //CSB file version dd. May 09, 2003 with array of parameter values at the start of the file
      file_ok = (fread(&pars, sizeof(int), 1, fp) == 1);
      if (file_ok)
        file_ok = (fread(csb_pars, sizeof(double), pars, fp) == pars);
      if (!file_ok)
        {
          fprintf(stderr, "\nCSB file seems to be invalid or corrupt!\n\n");
          return 1;
        }
      filepos = ftell(fp);
      break;
    default:
      fseek(fp, filepos, SEEK_SET);
      break;
    }

  while (fread(&cur_env, sizeof(Envdim), 1, fp) == 1)
    {
      fseek(fp, filepos, SEEK_SET);
      if (cur_env.memory_used > MemAllocated)
        {
          MemAllocated = ROUND_PAGE(cur_env.memory_used);
          mem_base = (void *)realloc(mem_base, (size_t)MemAllocated);
          if (!mem_base)
            {
              fprintf(stderr, "\nCould not allocate memory to read in data!\n");
              return 1;
            }
        }

      if (fread(mem_base, 1, (size_t)cur_env.memory_used, fp) != (size_t)cur_env.memory_used)
        {
          (void)fprintf(stderr, "\nFailed to read values from file: %s at line %d!\n\n", fn, __LINE__);
          return 1;
        }

      if (single)
        {
          if (fabs(cur_env.timeval - timeval) < fabs(closest_time - timeval))
            {
              closest_time = cur_env.timeval;
              closest_fpos = filepos;
              closest_fpos_set = 1;
            }
        }
      else
        {
          if (cmpval)
            compute_from_state();
          else
            write_txt_file(isf);
        }

      fseek(fp, filepos, SEEK_SET);
      fseek(fp, cur_env.memory_used, SEEK_CUR);
      filepos = ftell(fp);
    }

  if (single && closest_fpos_set)
    {
      fseek(fp, closest_fpos, SEEK_SET);
      if (fread(&cur_env, sizeof(Envdim), 1, fp) != 1)
        {
          (void)fprintf(stderr, "\nFailed to read values from file: %s at line %d!\n\n", fn, __LINE__);
          return 1;
        }

      fseek(fp, closest_fpos, SEEK_SET);
      if (fread(mem_base, 1, (size_t)cur_env.memory_used, fp) != (size_t)cur_env.memory_used)
        {
          (void)fprintf(stderr, "\nFailed to read values from file: %s at line %d!\n\n", fn, __LINE__);
          return 1;
        }
      if (cmpval)
        compute_from_state();
      else
        write_txt_file(isf);
      return 1;
    }

  return 0;
}
