/***
   NAME
     csbappend
   DESCRIPTION
     Program appends a binary CSB file to another binary CSB file. 

   Last modification: AMdR - Aug 13, 2010
***/
#ifndef CSBAPPEND
#define CSBAPPEND
#endif

#include "stdio.h"
#include "stdlib.h"
#include "string.h"
#include <sys/stat.h>
#include <unistd.h>

#define MAXPARS 256
#define USAGE "csbappend <CSB filename #1> <CSB filename to be appended> "

// List of magic keys of all recognized CSB file types

#define CSB_MAGIC_KEY1 20030509

// Macros to round to memory pages

#define PAGESIZE (getpagesize())
#define ROUND_PAGE(s) ((int)((s & (PAGESIZE - 1)) ? (s & ~(PAGESIZE - 1)) + PAGESIZE : s))

// Type definition for environment and population

#include "ebtcsbdefs.h"

void*                             mem_base     = NULL;
long                              MemAllocated = 0L;
double                            csb_pars[MAXPARS];


/*==============================================================================*/

int main(int argc, char **argv)

{
  char *      fn;
  FILE *      fp1 = NULL;
  FILE *      fp2 = NULL;
  struct stat stattmp;
  uint32_t    filepos = 0, magic1, magic2;
  Envdim      cur_env;
  int         pars, file_ok = 1;

  if (argc != 3)
    {
      fprintf(stderr, "\nUsage: %s\n\n", USAGE);
      exit(1);
    }
  // First CSB file
  fn = argv[1];
  if (!strcmp(fn + strlen(fn) - 1, "."))
    strcat(fn, "csb");
  else if (strcmp(fn + strlen(fn) - 4, ".csb"))
    strcat(fn, ".csb");
  if (!stat(fn, &stattmp))                                                          // couldn't stat
    {
      if ((!strcmp(fn + strlen(fn) - 3, "csb")) && (stattmp.st_mode & S_IFREG))
        fp1 = fopen(fn, "r+b");                                                     // if regular file open as binary for appending
      else
        {
          fprintf(stderr, "File %s is not a regular CSB file.\n", fn);
          return 1;
        }
    }
  if (fp1 == NULL)
    {
      fprintf(stderr, "Can't open file %s\n", fn);
      return 1;
    }

  // Second CSB file
  fn = argv[2];
  if (!strcmp(fn + strlen(fn) - 1, "."))
    strcat(fn, "csb");
  else if (strcmp(fn + strlen(fn) - 4, ".csb"))
    strcat(fn, ".csb");
  if (!stat(fn, &stattmp))                                                          // couldn't stat
    {
      if ((!strcmp(fn + strlen(fn) - 3, "csb")) && (stattmp.st_mode & S_IFREG))
        fp2 = fopen(fn, "rb");                                                       // if regular file open for reading
      else
        {
          fprintf(stderr, "File %s is not a regular CSB file.\n", fn);
          return 1;
        }
    }
  if (fp2 == NULL)
    {
      fprintf(stderr, "Can't open file %s\n", fn);
      return 1;
    }

  if (fread(&magic1, sizeof(uint32_t), 1, fp1) != 1)
    {
      fprintf(stderr, "\nFirst CSB file does not contain any data!\n\n");
      return 1;
    }

  if (fread(&magic2, sizeof(uint32_t), 1, fp2) != 1)
    {
      fprintf(stderr, "\nSecond CSB file does not contain any data!\n\n");
      return 1;
    }

  if ((magic1 == CSB_MAGIC_KEY1) && (magic2 == CSB_MAGIC_KEY1))
    {
      // CSB file version dd. May 09, 2003 with array of parameter values at
      // the start of the file
      file_ok              = (fread(&pars, sizeof(int), 1, fp1) == 1);
      if (file_ok) file_ok = (fread(csb_pars, sizeof(double), pars, fp1) == pars);
      if (!file_ok)
        {
          fprintf(stderr, "\nFirst CSB file seems to be invalid or corrupt!\n\n");
          return 1;
        }
      fseek(fp1, 0L, SEEK_END);

      file_ok              = (fread(&pars, sizeof(int), 1, fp2) == 1);
      if (file_ok) file_ok = (fread(csb_pars, sizeof(double), pars, fp2) == pars);
      if (!file_ok)
        {
          fprintf(stderr, "\nSecond CSB file seems to be invalid or corrupt!\n\n");
          return 1;
        }
      filepos = ftell(fp2);
    }
  else
    {
      fprintf(stderr, "\nAppending not implemented for this type of CSB file!\n\n");
      return 1;
    }

  while (fread(&cur_env, sizeof(Envdim), 1, fp2) == 1)
    {
      fseek(fp2, filepos, SEEK_SET);
      if (cur_env.memory_used > MemAllocated)
        {
          MemAllocated = ROUND_PAGE(cur_env.memory_used);
          mem_base     = (void *)realloc(mem_base, (size_t)MemAllocated);
          if (!mem_base)
            {
              fprintf(stderr, "\nCould not allocate memory to read in data!\n");
              return 1;
            }
        }

      if (fread(mem_base, 1, (size_t)cur_env.memory_used, fp2) != (size_t)cur_env.memory_used)
        {
          (void)fprintf(stderr, "\nFailed to read values from file: %s at line %d!\n\n", argv[1], __LINE__);
          return 1;
        }

      // Now append to first CSB file
      fseek(fp1, 0L, SEEK_END);
      if (fwrite(mem_base, 1, (size_t)cur_env.memory_used, fp1) != (size_t)cur_env.memory_used)
        {
          (void)fprintf(stderr, "\nFailed to write values to file: %s at line %d!\n\n", argv[2], __LINE__);
          return 1;
        }

      fseek(fp2, filepos, SEEK_SET);
      fseek(fp2, cur_env.memory_used, SEEK_CUR);
      filepos = ftell(fp2);
    }

  return 0;
}
